/*
** Module   : SessionUtils.java
** Abstract : Defines APIs used to access the SESSION attributes and methods.
**
** Copyright (c) 2013-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description---------------------------------------
** 001 CA  20130112 Created initial version.
** 002 CS  20130214 Added support for SESSION: EXPORT, SERVER-OPERATING-MODE, REMOTE attributes.
** 003 CA  20130221 getType was renamed to getResourceType.
** 004 CA  20130222 Added stubs for more SESSION-related attributes and methods: REMOTE,
**                  PARAMETER. The implementations for EXPORT, SERVER-OPERATING-MODE and REMOVE 
**                  attrs call their ServerImpl counterparts.
** 005 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 006 CA  20130529 Added implementations for FIRST-SERVER, LAST-SERVER, REMOTE, EXPORT and
**                  SERVER-OPERATING-MODE.
** 007 OM  20130510 Added support for date-related SESSION attributes so they can be queried from 
**                  client side.
** 008 CA  20130822 Allow access to the DATE-FORMAT attribute (set/getDateOrder) outside of a P2J
**                  server.
** 009 CA  20130927 Static proxies need the unknown() API. Resource type is determined from 
**                  LegacyResource constants. Implemented the following attributes: FIRST-BUFFER,
**                  FIRST-QUERY, FIRST-/LAST-SOCKET, FIRST-/LAST-SERVER-SOCKET.
** 010 CA  20131002 Implemented the PARAMETER attribute.
** 011 CA  20140701 Added PIXELS-PER-ROW and PIXELS-PER-COLUMN.
** 012 CA  20140702 Implemented PIXELS-PER-ROW and PIXELS-PER-COLUMN.
** 013 OM  20141110 Refactored get/setDateOrder() to get/setDateFormat().
** 014 EVL 20141216 Adding SESSION:TOOLTIPS attribute handling, server side part.
** 015 GES 20141228 Completed PARAMETER attribute implementation by storing it in a context-local
**                  and allowing the client command line/bootstrap to override any default value
**                  found in the directory.
** 016 GES 20150227 Added FIRST-CHILD and LAST-CHILD attribute support.
** 017 HC  20150508 Improvements to GUI window scrolling support.
** 018 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
** 019 OM  20150915 Added stub for SESSION:DEBUG-ALERT, DATA-ENTRY-RETURN and APPL-ALERT-BOXES
**                  attributes and method. Fixed set3D method.  Added LOAD-ICON.
** 020 HC  20160222 Fixed error reporting when assigning session read-only attributes.
** 021 IAS 20160302 Support for the APPL-ALERT-BOXES, DATA-ENTRY-RETURN, DEBUG-ALERT and
**                  IMMEDIATE-DISPLAY SESSION attributes.
** 022 IAS 20160328 Client log support.
** 023 OM  20160401 Fixed method name collision for static proxy.
** 024 HC  20170811 Added conversion support for SESSION attributes PRINTER-NAME, PRINTER-PORT,
**                  PRINTER-CONTROL-HANDLE, PRINTER-HDC and the method GET-PRINTERS.
**     HC  20170817 Added conversion support and runtime stubs for SESSION:STARTUP-PARAMETERS and
**                  SESSION:SUPPRESS-WARNINGS.
** 025 HC  20171003 Implemented SESSION:SUPPRESS-WARNINGS runtime support and the ability to
**                  handle (display or output depending on the runtime state) legacy warnings.
** 026 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 027 EVL 20180220 Now the current printer name is getting from client OS base.
** 028 CA  20180502 Fixed issues with time-related attributes used by SessionExports. 
** 029 CA  20180503 Added MULTITASKING-INTERVAL, SYSTEM-ALERT-BOXES and SERVER-CONNECTION-ID. 
**                  Renamed/moved some SESSION attribute getters/setters to fix issues with using
**                  the SESSION via a handle.
** 030 CA  20181213 Added SESSION:FIRST-OBJECT and LAST-OBJECT attributes (stubs).
**                  Added loadIcon(character).
** 031 CA  20190122 Implemented FIRST-OBJECT and LAST-OBJECT.
** 032 CA  20190416 Added LOCAL-VERSION-INFO and CURRENT-REQUEST-INFO stubs.
**     OM  20190416 Added FIRST-DATASET and FIRST-DATA-SOURCE methods.
** 033 CA  20190423 Added these attributes: CURRENT-RESPONSE-INFO, ICFPARAMETER, CONTEXT-HELP-FILE,
**                  SERVER-CONNECTION-BOUND-REQUEST, stubs only.
** 034 EVL 20190513 Added I18nOps for SESSION CP attributes getters moved from EnvironmentOps.
** 035 CA  20190520 Fixed the name of some OO classes.
** 036 CA  20190628 Added SERVER-CONNECTION-BOUND and SERVER-CONNECTION-CONTEXT stubs.
** 037 CA  20190710 LOCAL-VERSION-INFO and LOCAL-REQUEST-INFO return proper instances.
**     CA  20190717 Added ICFPARAMETER runtime support.
** 038 RFB 20200304 Added SESSION:EXIT-CODE handlers.
** 039 IAS 20200410 Reworked support of the date-related SESSION attributes.
**     CA  20200427 Added support for SERVER-CONNECTION-BOUND, SERVER-CONNECTION-CONTEXT, 
**                  CURRENT-REQUEST-INFO, CURRENT-RESPONSE-INFO, CONTEXT-HELP-FILE, 
**                  SERVER-CONNECTION-BOUND-REQUEST.
** 040 OM  20201030 Invalid attribute API support for getters/setters.
**     OM  20201203 Fixed handling of READ/ONLY attributes.
**     CA  20210213 The requestInfo and responseInfo must be vars.
**     OM  20210226 Added DATA-SOURCE sibling navigation support. Fixed Dataset handles order.
**     CA  20210920 Send MULTITASKING-INTERVAL to the client-side (which has runtime support).
**     CA  20220114 Added SESSION:CLIENT-DISCONNECTED and TERMINATION-HOOK FWD extension attributes.
**     CA  20220128 TERMINATION-HOOK can not be set for remote sessions (appserver).
**     CA  20220208 Added ERROR-STACK-TRACE and -errorstack runtime.
**     CA  20220930 Cache the proxy associated with the SESSION system handle.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.oo.lang.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.proxy.*;

/**
 * This class provides implementation for SESSION attributes and methods.
 * Mainly, the APIs implemented here are defined in the {@link CommonSession}
 * interface (or one of its super interfaces) and are not statically 
 * implemented by the classes used by the {@link #asHandle()}.
 */
public class SessionUtils
{
   /** Context local proxy that allows SESSION attributes to be pushed from server to client. */
   private static final ContextLocal<WorkArea> work = new ContextLocal<WorkArea>()
   {
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       * <p>
       * This will obtain a unique instance of <code>SessionExports</code> depending on which
       * side (server or client) it is called the returned value can be a local or a network
       * proxy instance.
       *
       * @return   The newly instantiated work area.
       */
      @Override
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         if (!ErrorManager.isHeadless() && !SessionManager.get().isLeaf())
         {
            wa.exports = (SessionExports) RemoteObject.obtainInstance(SessionExports.class, true);
         }
         
         ClientParameters cp = StandardServer.getClientParameters();
         if (cp != null)
         {
            wa.isDebugAlert = cp.debugAlert;
            wa.isClientLog  = cp.clientlog != null;
            wa.errorStackTrace = cp.errorStack;
            wa.param = cp.param;
         }
         if (wa.param == null)
         {
             Directory dir = DirectoryManager.getInstance();
             wa.param = dir.getString(Directory.ID_RELATIVE, "startup_parameter", "");
            
         }
         
         return wa;
      }
   };
   
   /**
    * Get a the instance for the SESSION system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link CommonSession} interface and its methods implemented by these
    * classes: {@link SessionUtils}, {@link EnvironmentOps}, {@link date},
    * {@link LogicalTerminal}, {@link ProcedureManager}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = work.get();

      if (wa.proxy != null)
      {
         return new handle(wa.proxy);
      }
      
      Class[] classes = 
      {
         SessionUtils.class,
         EnvironmentOps.class,
         date.class,
         I18nOps.class,
         LogicalTerminal.class,
         ProcedureManager.class,
         NumberType.class
      };
      CommonSession proxy = StaticProxy.obtain(CommonSession.class, classes);
      wa.proxy = proxy;
      
      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }

      return new handle(proxy);
   }
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = work.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      // the following attribute names are reported in their shortest forms
      if (attribute.toUpperCase().startsWith("DISPLAY-TY"))
      {
         attribute = "DISPLAY-T";
      }
      else if (attribute.toUpperCase().startsWith("HEIGHT-C"))
      {
         attribute = "HEIGHT";
      }
      else if (attribute.toUpperCase().startsWith("WIDTH-C"))
      {
         attribute = "WIDTH";
      }
      
      // the error message for session read-only attributes is different from the one emitted by
      // handle.readOnlyError() - settable is spelled with two 't' and there is no word 'widget' at the end of
      // the message.
      ErrorManager.recordOrShowError(4052, attribute.toUpperCase(), "settable", LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Emitted and to be used only in cases the attribute is read-only. It displays appropriate error message.
    * <br>
    * TODO: decide whether the following message can be raised here:<br>
    * {@code **Unable to assign UNKNOWN value to attribute <attribute> on <widget id>.}
    *
    * @param    attribute
    *           The read-only attribute.
    * @param    expr
    *           The value which is attempted to be assigned to the read-only attribute. 
    */
   public static void readOnlyError(String attribute, Object expr)
   {
      readOnlyError(attribute);
   }   
   
   /**
    * Indicates if the session is for the context of an appserver.
    * 
    * @return   <code>true</code> if the session is for the context of an appserver.
    */
   public  static logical isRemote()
   {
      return new logical(_isRemote());
   }

   /**
    * Indicates if the session is for the context of an appserver.
    * 
    * @return   <code>true</code> if the session is for the context of an appserver.
    */
   public static boolean _isRemote()
   {
      return AppServerManager.isRemote();
   }

   /**
    * Get the CLIENT-DISCONNECTED attribute of this session.
    * 
    * @return   <code>true</code> if the FWD client has disconnected (due to an abend, socket close, etc).
    */
   public static logical isClientDisconnected()
   {
      if (work.get().clientDisconnected)
      {
         return new logical(true);
      }
      
      // check if the client socket is gone
      Session session = SessionManager.get().getSession();
      return new logical(!session.isRunning());
   }
   
   /**
    * Get the TERMINATION-HOOK attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "TERMINATION-HOOK")
   public static character getTerminationHook()
   {
      return new character(work.get().terminationHook);
   }
   
   /**
    * Set the TERMINATION-HOOK attribute of this session.
    * <p>
    * This attribute is valid only for non-remote sessions - for a remote session, a message will be logged.
    * 
    * @param   hook
    *          The hook, a character expression with an external program name or a static method reference
    *          following the <code>com.goldencode.Foo:someStaticMethod</code> syntax.
    */
   @LegacyAttribute(name = "TERMINATION-HOOK", setter = true)
   public static void setTerminationHook(character hook)
   {
      if (_isRemote())
      {
         ErrorManager.recordOrShowError(-1001, "TERMINATION-HOOK can not be set for remote sessions.", 
                                        false, false);
         return;
      }
      work.get().terminationHook = hook.isUnknown() ? null : hook.toStringMessage();
   }
   
   /**
    * Set the TERMINATION-HOOK attribute of this session.
    * 
    * @param   hook
    *          The hook, a character expression with an external program name or a static method reference
    *          following the <code>com.goldencode.Foo:someStaticMethod</code> syntax.
    */
   public static void setTerminationHook(String hook)
   {
      work.get().terminationHook = hook;
   }

   /**
    * Method implementation for getting the SERVER-OPERATING-MODE attribute of the session handle.
    * <p>
    * Possible values:
    * <ul>
    *    <li>State-reset
    *    <li>State-aware
    *    <li>Stateless
    *    <li>State-free
    * </ul>
    * 
    * @return   The corresponding SERVER-OPERATING-MODE attribute of the session handle.
    */
   public static character getServerOperatingMode()
   {
      return AppServerManager.getServerOperatingMode();
   }
   
   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and  name-patterns separated by comma.
    *         
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(String procList)
   {
      return AppServerManager.export(procList);
   }
   
   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and  name-patterns separated by comma.
    *         
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(character procList)
   {
      return AppServerManager.export(procList);
   }

   /**
    * Get the value of the PARAMETER attribute, which is set to the value of the 4GL 
    * <code>-param</code> command line option specified for the current session.
    * <p>
    * The following is the precedence order (highest to lowest) for obtaining the
    * value of this parameter:
    * <p>
    * <ul>
    *    <li> The client command line.
    *    <li> The client bootstrap configuration.
    *    <li> The directory using relative lookup and the "startup_parameter" node.
    * </ul>
    *
    * @return   The <code>SESSION:PARAMETER</code> value for this context.
    */
   public static character getParameter()
   {
      return new character(work.get().param);
   }

   /**
    * Returns the handle for the first dynamic buffer in the first table containing a dynamic 
    * buffer.
    * 
    * @return   See above.
    */
   public static handle firstBuffer()
   {
      // TODO: this should be OK, but more testing is needed to check if 4GL doesn't sort them in
      // a different order (i.e. using the table name)
      return HandleChain.firstResource(LegacyResource.BUFFER);
   }

   /**
    * Returns the handle for the first window in the session.
    * 
    * @return   The first window in the session's list of windows.
    */
   public static handle firstChild()
   {
      return HandleChain.firstResource(LegacyResource.WINDOW);
   }

   /**
    * Returns the handle for the last window in the session.
    * 
    * @return   The last window in the session's list of windows.
    */
   public static handle lastChild()
   {
      return HandleChain.lastResource(LegacyResource.WINDOW);
   }

   /**
    * Returns the handle for the first created dynamic query.
    * <p>
    * TODO: this should call a worker method from the persist package.
    * 
    * @return   See above.
    */
   public static handle firstQuery()
   {
      return HandleChain.firstResource(LegacyResource.QUERY);
   }
   
   /**
    * Get the FIRST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstServer()
   {
      return HandleChain.firstResource(LegacyResource.SERVER);
   }
   
   /**
    * Get the FIRST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstServerSocket()
   {
      return HandleChain.firstResource(LegacyResource.SERVER_SOCKET);
   }
   
   /**
    * Get the FIRST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstSocket()
   {
      return HandleChain.firstResource(LegacyResource.SOCKET);
   }
   
   /**
    * Get the LAST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastServer()
   {
      return HandleChain.lastResource(LegacyResource.SERVER);
   }
   
   /**
    * Get the LAST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastServerSocket()
   {
      return HandleChain.lastResource(LegacyResource.SERVER_SOCKET);
   }
   
   /**
    * Get the LAST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastSocket()
   {
      return HandleChain.lastResource(LegacyResource.SOCKET);
   }
   
   /**
    * Obtain a handle to the first dynamic {@code DataSet} object created in the current session.
    * <p>
    * The list of Datasets is reversed (it is a stack instead of a queue).
    *
    * @return  the value of {@code SESSION:FIRST-DATASET}, as described above.
    */
   public static handle getFirstDataSet()
   {
      return HandleChain.lastResource(LegacyResource.DATASET);
   }
   
   /**
    * Obtain a handle to the first dynamic {@code DataSource} object created in the current session.
    * <p>
    * The list of DataSources is reversed (it is a stack instead of a queue).
    *
    * @return  the value of {@code SESSION:FIRST-DATA-SOURCE}, as described above.
    */
   public static handle getFirstDataSource()
   {
      return HandleChain.lastResource(LegacyResource.DATA_SOURCE);
   }
   
   /** 
    * Sets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using "M", "D" and "Y" once 
    * each.  The order from index 0 to index 2 represents the left to right ordering of the 
    * components. So the leftmost date component will be defined by the character at index 
    * position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   public static void setDateFormat(String dateFormat)
   {
      setDateFormat(new character(dateFormat));
   }
   
   /** 
    * Sets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using "M", "D" and "Y" once 
    * each.  The order from index 0 to index 2 represents the left to right ordering of the 
    * components. So the leftmost date component will be defined by the character at index 
    * position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   public static void setDateFormat(character dateFormat)
   {
      SessionDateUtils.INSTANCE.setDateFormat(dateFormat);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setDateFormat(dateFormat);
      }
   }
   
   /** 
    * Gets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components is a 3 character string using
    * "M", "D" and "Y" once each.  The order from index 0 to index 2
    * represents the left to right ordering of the components. So the leftmost
    * date component will be defined by the character at index position 0.
    * <p>
    * The value accessed is specific to the current user's context.
    *
    * @return   Defines the order of date components for conversion to/from strings.
    */
   public static character getDateFormat()
   {
      return SessionDateUtils.INSTANCE.getDateFormat();
   }
   
   /**
    * Get the state of SESSION:DEBUG-ALERT attribute. 
    * 
    * @return  the state of SESSION:DEBUG-ALERT attribute.
    */
   public static logical isDebugAlert()
   {
      return new logical(work.get().isDebugAlert);
   }
   
   /**
    * Set the state of SESSION:DEBUG-ALERT attribute.
    * 
    * @param   on
    *          The new state for SESSION:DEBUG-ALERT attribute.
    */
   public static void setDebugAlert(logical on)
   {
      setDebugAlert(on == null || on.isUnknown() ? false : on.booleanValue());
   }
   
   /**
    * Set the state of SESSION:DEBUG-ALERT attribute.
    * 
    * @param   on
    *          The new state for SESSION:DEBUG-ALERT attribute.
    */
   public static void setDebugAlert(boolean on)
   {
      work.get().isDebugAlert = on;
   }
   
   /**
    * Get the state of SESSION:DATA-ENTRY-RETURN attribute. 
    * 
    * @return  the state of SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static logical isDataEntryReturn()
   {
      return LogicalTerminal._isDataEntryReturn();
   }
   
   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static void setDataEntryReturn(logical on)
   {
      if (on == null || on.isUnknown())
      {
         return;
      }
      setDataEntryReturn(on.booleanValue());
   }
   
   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static void setDataEntryReturn(boolean on)
   {
      LogicalTerminal.setDataEntryReturn(on);
   }
   
   /**
    * Set the state of SESSION:LOAD-ICON attribute.
    *
    * @param   iconName
    *          The file name of the icon to be loaded
    *  @return <code>true</code> if successful
    */
   public static logical loadIcon(String iconName)
   {
      UnimplementedFeature.missing("LOAD-ICON attribute is not implemented");
      return new logical();
   }
   
   /**
    * Set the state of SESSION:LOAD-ICON attribute.
    *
    * @param   iconName
    *          The file name of the icon to be loaded
    *  @return <code>true</code> if successful
    */
   public static logical loadIcon(character iconName)
   {
      UnimplementedFeature.missing("LOAD-ICON attribute is not implemented");
      return new logical();
   }
   
   /**
    * Sets the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @param    offset
    *           The new timezone offset which must be between -840 and 840
    *           inclusive.  The value may also be <code>unknown</code> which
    *           is the equivalent of resetting this attribute to the default.
    */
   public static void setSessionTimeZone(NumberType offset)
   {
      SessionDateUtils.INSTANCE.setSessionTimeZone(offset);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionTimeZone(offset);
      }
   }
   
   /**
    * Accesses the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @return   The timezone offset. If this has never been set, then
    *           <code>unknown</code> will be returned.
    */
   public static integer getSessionTimeZone()
   {
      return SessionDateUtils.INSTANCE.getSessionTimeZone();
   }
   
   /**
    * Obtains the currently configured TIME-SOURCE attribute for the current
    * user's context.  This attribute is can be read and written but it is
    * not honored for date/time processing in P2J.  The intention of the
    * attribute is to define the system from which time values will be obtained.
    * In P2J, time values are always obtained from the application server. 
    * <p>
    * The default value is "local" which in Progress means that the client
    * machine is the source  (this is the same as the empty string "").  It is
    * common to use a database name in the 4GL, which makes the application
    * use the database server as the time source.
    *
    * @return   The time source attribute.
    */
   public static character getSessionTimeSource()
   {
      return SessionDateUtils.INSTANCE.getSessionTimeSource();
   }
   
   /**
    * Sets the <code>TIME-SOURCE</code> attribute for the user's context.
    * This attribute is can be read and written but it is not honored for 
    * date/time processing in P2J.  The intention of the attribute is to define
    * the system from which time values will be obtained. In P2J, time values
    * are always obtained from the application server.
    *
    * @param    timeSource
    *           The new value to set. If <code>null</code>, then the empty
    *           string will be used, which in the 4GL has the same affect as
    *           "local".
    */
   public static void setSessionTimeSource(character timeSource)
   {
      SessionDateUtils.INSTANCE.setSessionTimeSource(timeSource);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionTimeSource(timeSource);
      }
   }
   
   /**
    * Obtains the currently configured DISPLAY-TIMEZONE attribute for the current
    * user's context. This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @return   The display-timezone attribute.
    */
   public static integer getSessionDisplayTimeZone()
   {
      return SessionDateUtils.INSTANCE.getSessionDisplayTimeZone();
   }
   
   /**
    * Sets the <code>DISPLAY-TIMEZONE</code> attribute for the user's context.
    * This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @param    displayTimezone
    *           The new value to set.
    */
   public static void setSessionDisplayTimeZone(NumberType displayTimezone)
   {
      SessionDateUtils.INSTANCE.setSessionDisplayTimeZone(displayTimezone);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionDisplayTimeZone(displayTimezone);
      }
   }
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute getter. Returns the 
    * currently configured start year for the two-digit Y2K windowing range.
    * This is the equivalent to the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @return  The start year for the two-digit windowing range.
    */
   public static integer getYearOffset()
   {
      return SessionDateUtils.INSTANCE.getYearOffset();
   }
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute setter. Sets the 
    * configured start year for the two-digit Y2K windowing range. This is
    * the equivalent of the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @param   year
    *          The new start year for the two-digit windowing range.
    */
   public static void setYearOffset(NumberType year)
   {
      SessionDateUtils.INSTANCE.setYearOffset(year);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setYearOffset(year);
      }
   }
   
   /**
    * Get the PIXELS-PER-ROW attribute of this session.
    * 
    * @return   See above.
    */
   public static integer getPixelsPerRow()
   {
      return new integer(LogicalTerminal.getClient().getPixelsPerRow());
   }
   
   /**
    * Get the PIXELS-PER-COLUMN attribute of this session.
    * 
    * @return   See above.
    */
   public static integer getPixelsPerColumn()
   {
      return new integer(LogicalTerminal.getClient().getPixelsPerColumn());
   }

   /**
    * Sets the SESSION:TOOLTIPS attribute for this session.
    *
    * @param    tooltips
    *           The new value for session related TOOLTIPS attribute.
    */
   public static void setSessionTooltips(logical tooltips)
   {
      WorkArea wa = work.get();

      // change the local copy
      if (tooltips != null && !tooltips.isUnknown())
      {
         wa.tooltips = tooltips.booleanValue();
      }
      else
      {
         wa.tooltips = false;
      }

      // and push new value to the client
      LogicalTerminal.getClient().setSessionTooltips(wa.tooltips);
   }
   
   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   <code>true</code> if tooltips are available for this session
    *           <code>false</code> otherwise.
    */
   public static logical getSessionTooltips()
   {
      WorkArea wa = null;
      try
      {
         wa = work.get();
      }
      catch (Throwable e)
      {
         // NOTE: this is solely for debug: when the client stops with a strange NPE in
         //       ThinClient.setSessionTooltips(). This is the fist remote call so if things go
         //       wrong with initialization of the context on server-side here is the error
         //       accessible./visible.
         e.printStackTrace();
         // rethrow
         throw e;
      }
      return new logical(wa.tooltips);
   }
   
   /**
    * Get THREE-D attribute value for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static logical is3D()
   {
      return LogicalTerminal.is3D();
   }
   
   /**
    * Set THREE-D attribute value for SESSION handle.
    * 
    * @param   value
    *          The value to set.
    */
   public static void set3D(logical value)
   {
      LogicalTerminal.set3D(value);
   }
   
   /**
    * Check if client log is active
    * 
    * @return <code>true</code> if client log is active
    */
   public static boolean isClientlog()
   {
      return work.get().isClientLog;
   }

   /**
    * Get the state of SESSION:PRINTER-NAME attribute.
    *
    * @return  the state of SESSION:PRINTER-NAME attribute.
    */
   public static character getPrinterName()
   {
      return EnvironmentOps.getCurrentPrinterName();
   }

   /**
    * Set the state of SESSION:PRINTER-NAME attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-NAME attribute.
    */
   public static void setPrinterName(character value)
   {
      // currently only the built-in PDF printer is supported, so this method is a no-op
   }

   /**
    * Get the state of SESSION:PRINTER-PORT attribute.
    *
    * @return  the state of SESSION:PRINTER-PORT attribute.
    */
   public static character getPrinterPort()
   {
      // currently only the built-in PDF printer is supported, so this method always returns an unknown
      return new character();
   }

   /**
    * Set the state of SESSION:PRINTER-PORT attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-PORT attribute.
    */
   public static void setPrinterPort(character value)
   {
      readOnlyError("PRINTER-PORT");
   }

   /**
    * Get the state of SESSION:PRINTER-CONTROL-HANDLE attribute.
    *
    * @return  the state of SESSION:PRINTER-CONTROL-HANDLE attribute.
    */
   public static integer getPrinterControlHandle()
   {
      return new integer(PrintingService.getControlHandle());
   }

   /**
    * Set the state of SESSION:PRINTER-CONTROL-HANDLE attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-CONTROL-HANDLE attribute.
    */
   public static void setPrinterControlHandle(integer value)
   {
      PrintingService.resetControlHandle();
   }

   /**
    * Get the state of SESSION:PRINTER-HDC attribute.
    *
    * @return  the state of SESSION:PRINTER-HDC attribute.
    */
   public static integer getPrinterHDC()
   {
      return new integer(PrintingService.getHDC());
   }

   /**
    * The SESSION:GET-PRINTERS method.
    *
    * @return  the result of SESSION:GET-PRINTERS method.
    */
   public static character getPrinters()
   {
      // currently only the built-in PDF printer is supported
      return new character(PrintingService.PDF_PRINTER_DESCR);
   }

   /**
    * Get the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @return  the state of SESSION:SUPPRESS-WARNINGS attribute.
    */
   public static logical isSuppressWarnings()
   {
      return new logical(ErrorManager.isSuppressWarnings());
   }

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   public static void setSuppressWarnings(logical value)
   {
      if (value.isUnknown())
      {
         // UNKNOWN is silently ignored
         return;
      }

      setSuppressWarnings(value.booleanValue());
   }

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   @LegacyAttribute(name = "SUPPRESS-WARNINGS", setter = true)
   public static void setSuppressWarnings(boolean value)
   {
      ErrorManager.setSuppressWarnings(value);
   }

   /**
    * Get the state of SESSION:STARTUP-PARAMETERS attribute.
    *
    * @return  the state of SESSION:STARTUP-PARAMETERS attribute.
    */
   public static character getStartupParameters()
   {
      return new character(LegacyCLI.getInstance().getStartupParameters());
   }

   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static void setSystemAlertBoxes(boolean on)
   {
      ErrorManager.setSystemAlertBoxes(on);
   }
   
   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static void setSystemAlertBoxes(logical on)
   {
      setSystemAlertBoxes(on == null || on.isUnknown() ? false : on.booleanValue());
   }

   /**
    * Get the state of SESSION:SYSTEM-ALERT-BOXES attribute. 
    * 
    * @return  the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static logical isSystemAlertBoxes()
   {
      return new logical(ErrorManager.isSystemAlertBoxes());
   }
   
   /**
    * Set the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @param   interval 
    *          The new value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   public static void setMultitaskingInterval(NumberType interval)
   {
      long val = interval.isUnknown() ? 0 : interval.longValue();
      work.get().multitaskingInterval = val;
      LogicalTerminal.getClient().setMultitaskingInterval(val);
   }
   
   /**
    * Get the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @return  the value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   public static integer getMultitaskingInterval()
   {
      return new integer(work.get().multitaskingInterval);
   }
   
   /**
    * Set the value of SESSION:EXIT-CODE attribute. 
    * 
    * @param   value 
    *          The new value of SESSION:EXIT-CODE attribute.
    */
   public static void setExitCode(long value)
   {
      // TODO: The exit code needs to get out of this call, and be returned from the JVM to the OS
      work.get().exitCode = value;      
   }
   
   /**
    * Get the value of SESSION:EXIT-CODE attribute. 
    * 
    * @return  the value of SESSION:EXIT-CODE attribute.
    */
   public static long getExitCode()
   {
      return work.get().exitCode;
   }
   
   /**
    * Get the SERVER-CONNECTION-BOUND attribute of this session.
    * 
    * @return   See above.
    */
   public static logical isServerConnectionBound()
   {
      return AppServerManager.isServerConnectionBound();
   }
   
   /**
    * Get the SERVER-CONNECTION-CONTEXT attribute of this session.
    * 
    * @return   See above.
    */
   public static character getServerConnectionContext()
   {
      return new character(AppServerManager.getServerConnectionContext());
   }
   
   /**
    * Set the SERVER-CONNECTION-CONTEXT attribute of this session.
    * 
    * @param    val
    *           The attribute's value.
    */
   public static void setServerConnectionContext(String val)
   {
      AppServerManager.setServerConnectionContext(val);
   }
   
   /**
    * Set the SERVER-CONNECTION-CONTEXT attribute of this session.
    * 
    * @param    val
    *           The attribute's value.
    */
   public static void setServerConnectionContext(character val)
   {
      AppServerManager.setServerConnectionContext(val.getValue());
   }
   
   /**
    * Get the value of SESSION:SERVER-CONNECTION-ID attribute. 
    * 
    * @return  the value of SESSION:SERVER-CONNECTION-ID attribute.
    */
   public static character getServerConnectionId()
   {
      String id = AppServerManager.getServerConnectionId();
      return new character(id == null ? "" : id);
   }
   
   
   /**
    * Get the FIRST-OBJECT attribute of this session.
    * 
    * @return   See above.
    */
   public static object<?> firstObject()
   {
      handle obj = HandleChain.firstResource(LegacyResource.OBJECT);
      return new object<>((ObjectResource) obj.get());
   }
   
   /**
    * Get the LAST-OBJECT attribute of this session.
    * 
    * @return   See above.
    */
   public static object<?> lastObject()
   {
      handle obj = HandleChain.lastResource(LegacyResource.OBJECT);
      return new object<>((ObjectResource) obj.get());
   }
   
   /**
    * Get the LOCAL-VERSION-INFO attribute of this session.
    * 
    * @return   See above.
    */
   public static object<? extends OeversionInfo> localVersionInfo()
   {
      return ObjectOps.newInstance(OeversionInfo.class);
   }
   
   /**
    * Get the CURRENT-REQUEST-INFO attribute of this session.
    * 
    * @return   See above.
    */
   public static object<? extends OerequestInfo> currentRequestInfo()
   {
      // TODO: unknown only if the client is connected to an appserver...
      WorkArea wa = work.get();
      
      if (wa.currentRequestInfo == null)
      {
         wa.currentRequestInfo = new ObjectVar<>(OerequestInfo.class);
         wa.currentRequestInfo.assign(ObjectOps.newInstance(OerequestInfo.class));
      }
      
      return new object<>(wa.currentRequestInfo);
   }

   /**
    * Get the CURRENT-RESPONSE-INFO attribute of this session.
    * 
    * @return   See above.
    */
   public static object<? extends OerequestInfo> currentResponseInfo()
   {
      // TODO: unknown only if the client is connected to an appserver...
      WorkArea wa = work.get();

      if (wa.currentResponseInfo == null)
      {
         wa.currentResponseInfo = new ObjectVar<>(OerequestInfo.class);
         wa.currentResponseInfo.assign(ObjectOps.newInstance(OerequestInfo.class));
      }
      
      return new object<>(wa.currentResponseInfo);
   }
   
   /**
    * Get the ICFPARAMETER attribute of this session.
    * 
    * @return   See above.
    */
   public static character getICFParameter()
   {
      return new character(work.get().icfparam);
   }
   
   /**
    * Get the CONTEXT-HELP-FILE attribute of this session.
    * 
    * @return   See above.
    */
   public static character getContextHelpFile()
   {
      return new character(work.get().contextHelpFile);
   }
   
   /**
    * Set the CONTEXT-HELP-FILE attribute of this session.
    * 
    * @param   file
    *          The file name.
    */
   public static void setContextHelpFile(String file)
   {
      work.get().contextHelpFile = file;
   }
   
   /**
    * Set the CONTEXT-HELP-FILE attribute of this session.
    * 
    * @param   file
    *          The file name.
    */
   public static void setContextHelpFile(character file)
   {
      work.get().contextHelpFile = file.getValue();
   }
   
   /**
    * Get the SERVER-CONNECTION-BOUND-REQUEST attribute of this session.
    * 
    * @return   See above.
    */
   public static logical getServerConnectionBoundRequest()
   {
      return AppServerManager.isServerConnectionBoundRequest();
   }
   
   /**
    * Set the SERVER-CONNECTION-BOUND-REQUEST attribute of this session.
    * 
    * @param   state
    *          The bound state.
    */
   public static void setServerConnectionBoundRequest(boolean state)
   {
      AppServerManager.setServerConnectionBoundRequest(state);
   }
   
   /**
    * Set the SERVER-CONNECTION-BOUND-REQUEST attribute of this session.
    * 
    * @param   state
    *          The bound state.
    */
   public static void setServerConnectionBoundRequest(logical state)
   {
      AppServerManager.setServerConnectionBoundRequest(state.booleanValue());
   }

   /**
    * Set the ERROR-STACK-TRACE attribute of this session.
    * 
    * @param   state
    *          The attribute's state.
    */
   public static void setErrorStackTrace(logical state)
   {
      work.get().errorStackTrace = state.booleanValue();
   }
   
   /**
    * Set the ERROR-STACK-TRACE attribute of this session.
    * 
    * @param   state
    *          The attribute's state.
    */
   public static void setErrorStackTrace(boolean state)
   {
      work.get().errorStackTrace = state;
   }
   
   /**
    * Get the ERROR-STACK-TRACE attribute of this session.
    * 
    * @return   See above.
    */
   public static logical isErrorStackTrace()
   {
      return new logical(work.get().errorStackTrace);
   }

   /**
    * Mark this context as the client has disconnected (due to an abend, socket disconnect, etc).
    */
   static void clientDisconnected()
   {
      work.get().clientDisconnected = true;
   }
   
   /**
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** Cached instance of the SESSION resource. */
      public CommonSession proxy = null;

      /** The resource's ID. */
      private Long id = null;

      /** Remote client interface. */
      private SessionExports exports = null;
      
      /** Command line <code>-param</code> value (SESSION:PARAM attribute). */
      private String param = null;
      
      /** The state of SESSION:DEBUG-ALERT attribute */
      private boolean isDebugAlert = false;

      /** The flag indicating whether the widgets have the tooltip feature. */
      private boolean tooltips = Utils.getDirectoryNodeBoolean(null, "session-tooltips", true);
      
      /** The flag indicating whether the clieet log is active */ 
      private boolean isClientLog = false;
      
      /** This is just a holder for SESSION:MULTITASKING-INTERVAL. */
      private long multitaskingInterval = 0;
      
      /** The resultant value the 4GL may want to return to the OS (SESSION:EXIT-CODE) */
      /*
       * TODO: The return of this value needs to be implemented somewhere in the ConversionDriver
       *       so it can make its way back to the OS.
       */
      private long exitCode = 0;
      
      /** The value of the <code>-icfparam</code>. */
      private String icfparam = LegacyCLI.getInstance().getICFParameter();

      /** The value of the CONTEXT-HELP-FILE attribute. */
      private String contextHelpFile = null;
      
      /** The value of the CURRENT-REQUEST-INFO attribute. */
      private object<OerequestInfo> currentRequestInfo = null;
      
      /** The value of the CURRENT-RESPONSE-INFO attribute. */
      private object<OerequestInfo> currentResponseInfo = null;
      
      /**
       * Flag indicating if the FWD client has disconnected (due to an abend, socket close, etc). 
       * Will be set to <code>true</code> by the FWD when {@link TransactionManager#abnormalEnd} is being
       * called for an unrecoverable exception.
       */
      private boolean clientDisconnected = false;
      
      /**
       * A hook which will be executed before the root top-level block terminates.  Can specify either an
       * external program or a static method reference following the 
       * <code>com.goldencode.Foo:someStaticMethod</code> syntax.
       */
      private String terminationHook = null;
      
      /**
       * Flag indicating if the {@link ProError#getCallStack()} should be populated with the stacktrace when
       * an legacy OO error is thrown.
       */
      private boolean errorStackTrace = false; 
   }
}
