/*
** Module   : BufferImpl.java
** Abstract : Implementation of Buffer interface.
**
** Copyright (c) 2013-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------------Description----------------------------------------
** 001 ECF 20130118 Created initial version. Implementation of the Buffer
**                  interface; provides services for buffer-related, Progress
**                  methods, attributes, and some language statements.
** 002 VMN 20130207 Added support for handle-based methods
**                  BUFFER-COPY (KW_BUF_COPY)
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-FIELD (KW_BUF_FLD)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  and attributes
**                  NUM-FIELDS (KW_NUM_FLD)
**                  COLUMN-LABEL (KW_COL_LAB)
**                  INITIAL (KW_INIT)
**                  AMBIGUOUS (KW_AMBIG)
** 003 ECF 20130214 Added bufferCreate and variants of bufferCopy.
** 004 SVL 20130213 Added disable*Triggers functions.
** 005 VMN 20130217 Added support for handle-based attributes:
**                  CURRENT-ITERATION (KW_CUR_ITER)
**                  ERROR (KW_ERROR)
**                  DBNAME (KW_DBNAME)
**                  TABLE (KW_TABLE)
** 006 VMN 20130219 Added support for attributes
**                  TABLE-HANDLE (KW_TAB_HAND)
** 007 CA  20130221 Added support for ADM-DATA, UNIQUE-ID, NAMESPACE-URI. Implemented Deletable.
**                  Added missing BUFFER-FIELD stubs.
** 008 GES 20130224 Added support for a DELETE statement with a validation expression.
** 009 ECF 20130224 Added limited support for embedded SQL.
** 010 OM  20130222 Added bufferField variants and dereference from Dereferenceable.
** 011 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 012 VMN 20130306 Corrected support for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 013 VMN 20130311 Corrected support and runtime for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  BUFFER-COMPARE (KW_BUF_COMP)
** 014 ECF 20130309 Changed API for bufferCopy and bufferCompare; only wrapped data types are
**                  allowed for literal parameters, to prevent the explosion of the API. 
** 015 VMN 20130314 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP) also noLobs parameter type changed
**                  BUFFER-COPY (KW_BUF_COPY) (simple signatures)
**                  First parameter type changed from String to character for
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 016 CA  20130323 Changed the APIs for KW_BUF_DEL and KW_EMPTY_TT to return logical.
** 017 ECF 20130417 Moved initial() method to FieldReference.
** 018 VMN 20130426 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-COPY (KW_BUF_COPY)
** 019 CA  20130603 resourceDelete must return a boolean value.
** 020 VMN 20130711 Implemented verifyTransaction() and added to handle-based methods:
**                  BUFFER-CREATE (KW_BUF_CREA)
**                  BUFFER-DELETE (KW_BUF_DELETE)
**                  FIND-BY-ROWID (KW_FIND_BR)
** 021 VMN 20130717 Added aggregated functions support for embedded SQL.
** 022 SVL 20130822 Added DYNAMIC attribute, implemented bufferField.
** 023 CA  20130927 Removed unknown() API, as by default the resource is known. 
** 024 CA  20131021 Do not delete static buffers.
** 025 ECF 20131028 Import change.
** 026 SVL 20131028 Handled bufferField() for static tables, added doSetName.
** 027 CA  20131013 Static buffers are deleted only when the instantiating procedure gets deleted.
** 028 OM  20131128 Stub implementation for getUniqueID.
** 029 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
** 030 VMN 20140110 Implemented bufferField(character).
** 031 OM  20140107 Implemented disableDumpTriggers() and disableLoadTriggers() methods and 
**                  and removed unneeded openReadOnlyScope().
** 032 OM  20140116 Implemented attributes and methods (NUM-FIELDS, ERROR, DBNAME, TABLE,
**                  NAMESPACE-URI, dereference(::), BUFFER-FIELD(int), intermediary step for
**                  findFirst, findLast, findUnique).
** 033 OM  20140122 Added primary index selection for find-by-rowid dynamic finds. Added 
**                  implementation for dynamic finds methods using DynamicQueryHelper.
** 034 ECF 20140305 Added defaultBufferHandle implementation, reworked bufferField(String) to
**                  handle invalid buffer name.
** 035 VMN 20140330 Added LockType errors processing.
** 036 VIG 20140314 Changed currentIteration() method signature.
** 037 OM  20141103 Added doGetName() method.
** 038 ECF 20141117 Added stub methods for all raw-transfer variants.
** 039 ECF 20150114 Reimplemented valid() to always return true. A buffer whose scope is not open
**                  is still valid for use.
** 040 OM  20150205 Queries do not need to be unregistered from DynamicQueryHelper.
** 041 OM  20150302 Fixed bufferField() behavior on invalid field name.
** 042 ECF 20150315 Replaced FieldReference with BufferFieldImpl for BUFFER-FIELD use cases.
** 043 OM  20150303 Fixed hasNext/PrevSibling(). Implemented special resource chaining for
**                  BUFFERs (TEMP-TABLEs in front of permanent TABLE buffers).
** 044 OM  20150506 Updated javadoc for deleteAll() and delete() methods.
** 045 CA  20151219 Implicit buffers for a defined global shared var can not be deleted.  Also,
**                  deleted buffers must be tracked explicitly.
** 046 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 047 ECF 20160219 Added checkExtentRange and outOfBoundsError, methods which are called from
**                  DMO proxies. Reimplemented _numFields.
** 048 EVL 20160325 Javadoc fixes.
** 049 ECF 20160404 Javadoc fixes.
** 050 OM  20160302 Dropped errors generated inside _find() methods.
**         20160507 Partially reverted dropping of _find() errors to fix regressions.
** 051 SVL 20160510 Support for table-level delete validation.
** 052 ECF 20160517 Modified findByRowID to deal with null primary index more efficiently.
** 053 GES 20160530 Validation processing now uses lambdas instead of BufferValidator instances.
** 054 OM  20160623 Small change in error message.
** 055 SVL 20160705 Added indexInformation method.
** 056 IAS 20160714 Use legacy name instead of DMO alias.
** 057 ECF 20160829 Performance enhancement: use local BufferManager to check transaction status
**                  instead of TransactionManager.
** 058 IAS 20161907 Resolve P2JQuery.Parameter in delete(String, Object...)
** 059 ECF 20171016 Implemented readXml and writeXml methods.
** 060 GES 20171206 Switch to ErrorManager.silent().
**     ECF 20171212 Remove redundant query substitution parameter preprocessing for bulk delete.
** 061 OM  20171216 Added delete() variant with support for external buffers for nested selects.
**     CA  20171220 Cleanup the SharedVariableManager.bufferCreator on buffer delete.
** 062 GES 20180419 Added the RAW-TRANSFER() legacy method.
** 063 OM  20181114 Added NamedSerializable interface implementation.
** 064 ECF 20181208 Added stubs for recordLength, findCurrent, readJson, writeJson methods.
** 065 OM  20181206 Added QUERY readonly attribute for BUFFERs.
** 066 OM  20190117 Added CURRENT-CHANGED, FIND-CURRENT. Fixed locking in _find.
**     ECF 20190118 Added basic runtime implementation of readJson, writeJson.
**     OM  20190206 Added implementation of recordLength().
**     OM  20190213 Added error messages in dereference().
**     ECF 20190212 Added no-lobs implementation to bufferCopy, bufferCompare methods.
**     OM  20190228 Removed deleted buffers from dynamic queries.
** 067 OM  20190327 Renamed DataSource to avoid conflicts with DataSet source.
** 068 CA  20190423 Added dereference(Class, String), used for chained :: operators.
**     OM  20190423 Implemented copyTempTable() methods.
** 069 CA  20190511 Added API for BUFFER-FIELD(POLY).
**                  Added BUFFER-VALIDATE stub.
** 070 SVL 20190614 Reflected changes in TableMapper.
** 071 OM  20190426 Extracted hasField() and getHibernateField() methods to avoid code duplication.
**     OM  20190510 Clarified generation of methods and attributes for variables with BUFFER /
**                  TEMP-TABLE ABL namespaces specified. Implemented several attributes.
**     HC  20190707 Implemented read(dataSet) and read(tempTable) methods of JsonObject legacy
**                  class.
** 072 ECF 20190813 Change to RecordBuffer.release API.
** 073 OM  20190720 Added BATCH FILL support. FILL returns FALSE and stops if a child FILL fails.
**                  Fixed CamelCase Sensitivity. Fixed FILL with DATA-SOURCE with provided query.
**     CA  20190728 If a callback raises an ERROR condition, invokeCallback must return false.
**                  Also, any FILL operation must be aborted.
**     CA  20190812 Allow this class to be mutable, to allow switching between resources, in cases
**                  of BY-REFERENCE or REFERENCE-ONLY.  WARNING: it is not allowed to access
**                  instance fields, if they are in another instance than 'this' reference!
**                  Use the 'ref()' API to get the runtime instance or add the field to an 
**                  accessor method.
**     OM  20190906 Added support for MERGE and REPLACE fill modes. Added ERROR-STRING and
**                  ERROR-FLAG attributes support.
** 075 CA  20191119 Fixed some issue with SAVE-ROW-CHANGES: any ERROR message prior to invocation
**                  must not affect the result; fixed the 'source was not changed in the database'
**                  check, where the source and before buffer must be checked. 
** 076 AIL 20191128 Moved the validation phase in RecordBuffer.
** 077 AIL 20200107 Added name override and restore for temporarily changing name with proxy.
**     AIL 20200108 Prevented direct access to record buffer members
**     CA  20200110 Fixes for static buffer resource delete.
** 078 IAS 20200221 Reworked RAW-TRANSFER method support.
** 079 IAS 20200414 bufferValidate() method implementation.
**     CA  20200427 Allow bulk assign support for BUFFER-VALUE and ::, in case of extent fields.
**     CA  20200428 Added LAST-BATCH runtime stub.
**     CA  20200503 Added BUFFER:IS-MULTI-TENANT runtime stub.
** 080 ECF 20200419 Removed Hibernate dependencies.
**     CA  20200714 The rowState is available even for temp-tables not part of a dataset.
** 081 GES 20200807 Reduced ErrorManager context-local usage.
**     CA  20200910 Added BUFFER:SERIALIZE-ROW() runtime stub.
**         20200911 attachDataSource allows empty include/excludeFields values.
**         20200914 attachDataSource allows ROWID function for source field.
**     OM  20201016 Added implementation for LAST-BATCH, DATA-SOURCE-MODIFIED.
**     OM  20201030 Invalid attribute API support for getters/setters.
**     OM  20201202 Fixed persistence support for ACCEPT-CHANGES and REJECT-CHANGES methods.
**     OM  20201205 Added AUTO-DELETE attribute implementation.
**     OM  20201218 Fixed implementation for error and rejected attributes/hidden fields.
**     OM  20210127 Replaced TableMapper.getLegacyName() triple map lookup with direct access to local
**                  dmoMeta.legacyTable. Fixed implementation of validate() method/statement.
**     OM  20210223 Invalidate index-based cache in the event of low-level SQL operations.
**     OM  20210309 Do not use DmoMeta as key in TableMapper because temp-tables might share the same
**                  DmoMeta instance.
**     ECF 20210506 Use RecordBuffer.getDmoInfo() getter instead of direct field access to prevent
**                  NPE in proxy case.
**     CA  20210509 Fixed 'defaultTypeMapping' implicit value for readXml - it must be the unknown value, not
**                  the "?" string.
**     AIL 20210511 Ensure field string names are normalized before searching.
**     OM  20210521 DELETE and CREATE return the proper logical value.
**     CA  20210626 Do not delete the master shared temp-table until the last of its associated buffers (i.e.
**                  child shared temp-tables) is deleted.
**     CA  20210629 Added dereference operator for indexed extent fields (rvalue and lvalue).
**     OM  20210809 Completed implementation of SERIALIZE-ROW|WRITE-JSON}(JsonObject|JsonArray).
**     ECF 20210820 Retrofit for query substitution parameter preprocessing fixes.
**     OM  20210921 Added missing implementations for NAMESPACE-PREFIX and XML-NODE-NAME attribute.
**     ECF 20210924 Minor refactoring.
**     CA  20210929 Fixed ATTACH-DATA-SOURCE when rowid function is used in the pairs.
**     OM  20211023 Fixed FILL for EXTENT fields.
**     AL2 20211105 Character values provided by dereference are case insensitive.
**     ECF 20211203 Fixed error checking in _find method.
**     OM  20211206 Improve implementation of DATA-SOURCE-COMPLETE-MAP attribute. Fixed FILL assignments.
**     OM  20211020 Added _DATASOURCE_ROWID property.
**     CA  20211228 BUFFER:IS-MULTI-TENANT attribute must return false, even if is not implemented - unknown 
**                  is not an acceptable returned value for this attribute.
**                  Replaced the UnimplementedFeature.missing from tenant-related APIs with a TODO comment.
**     CA  20211204 Fixed BUFFER-FIELD behavior, they need to behave like real resources, with their lifetime 
**                  linked with the parent BUFFER resource.
**     OM  20220218 When saving DS changes, allow Session.PK to be resolved instead of dereferencing it.
**                  Javadoc update.
**     OM  20220224 Small fixed of the error message for invalid WHERE clause.
**     OM  20220422 Allow Xml/JsonImport to load records into buffer without exposing package private methods.
**     OM  20220620 Added missing features in ATTACH-DATA-SOURCE method.
**     OM  20220707 Shared meta-knowledge about fields and properties of a record was collected in
**                  DmoMeta to avoid duplicating their collection and storage.
**     CA  20220707 Create the BUFFER-FIELD resources only when they are accessed.
**     OM  20220801 Allow unqualified field names in attachDataSource().
**     OM  20220810 Fixed check for changes in saveRowChangesImpl2() to include extent fields.
**     CA  20220930 Refactored the callback invocation to be performed via a call-site and InvokeConfig, to
**                  allow caching of the resolved target.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import java.util.function.*;
import java.lang.reflect.*;
import java.util.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.persist.orm.*;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.persist.serial.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.BlockManager.TransactionType;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.UniqueIdGenerator.*;
import com.goldencode.p2j.persist.TempTable.BeforeType;
import com.goldencode.util.*;
import com.google.common.collect.*;

import static com.goldencode.p2j.util.BlockManager.*;

/**
 * This class implements services representing Progress builtin functions,
 * methods, attributes, and language statements, which operate on record
 * buffers.
 * <p>
 * Converted business logic invokes these methods on DMO proxies via a
 * DMO-specific interface in the form <code>&lt;DMO interface&gt;.Buf</code>.
 * Each such interface extends both a custom interface of DMO-specific getter
 * and setter methods, as well as the {@link Buffer} interface, which defines
 * the API for these standard services.
 * <p>
 * Dynamic proxy classes for each DMO type extend this class, so that business
 * logic can invoke these standard services directly as method calls on DMO
 * proxy instances.
 * <p>
 * This class is intentionally package private, since it is not meant to be
 * extended outside of this package (and within this package only by dynamic
 * proxy class), nor instantiated directly.
 * <p>
 * This class implements {@link BufferReference} in order to grant its
 * instance methods access to the {@link RecordBuffer} object associated with
 * each instance of the DMO proxy class which extends this class. This access
 * is necessary for these methods to do their work. Even though {@link
 * #buffer()} is a public method, external packages will not be able to invoke
 * this method (except through reflection). This is by design, as business
 * logic is not meant to have access to <code>RecordBuffer</code> instances.
 * The default implementation of this method here throws an exception, as it
 * is never intended to be called, but it needs to be present in order for
 * this class to compile, since it is referenced by other methods. Each DMO
 * proxy class overrides this method to provide a working implementation.
 */
public abstract class BufferImpl
extends HandleChain
implements Buffer
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(BufferImpl.class.getName());
   
   /** The call-site for the {@link Buffer#BEFORE_FILL_EVENT} event callback. */
   private static final InvokeConfig BEFORE_FILL_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#AFTER_FILL_EVENT} event callback. */
   private static final InvokeConfig AFTER_FILL_CALL_SITE = new InvokeConfig();
   
   /** The call-site for the {@link Buffer#BEFORE_ROW_FILL_EVENT} event callback. */
   private static final InvokeConfig BEFORE_ROW_FILL_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#AFTER_ROW_FILL_EVENT} event callback. */
   private static final InvokeConfig AFTER_ROW_FILL_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#ROW_CREATE_EVENT} event callback. */
   private static final InvokeConfig ROW_CREATE_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#ROW_DELETE_EVENT} event callback. */
   private static final InvokeConfig ROW_DELETE_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#ROW_UPDATE_EVENT} event callback. */
   private static final InvokeConfig ROW_UPDATE_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#FIND_FAILED_EVENT} event callback. */
   private static final InvokeConfig FIND_FAILED_CALL_SITE = new InvokeConfig();

   /** The call-site for the {@link Buffer#SYNCHRONIZE_EVENT} event callback. */
   private static final InvokeConfig SYNCHRONIZE_CALL_SITE = new InvokeConfig();

   /**
    * Special 'pointer' used in management of BUFFER resources. This helps delimit the TEMP-tables
    * from PERMANENT ones in the list of BUFFERs. The TEMP-TABLE buffers are kept always in front
    * of the list.
    * To be sure there are no conflicts with legacy resource types declared in 
    * {@link LegacyResource}, this type starts with an underscore.
    */
   private static final String PERMANENT_TABLE = "_PERMANENT-TABLE";
   
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID = new integer(UniqueIdGenerator.getUniqueId(IdKind.BUFFER));
   
   /** Flag indicating this buffer was deleted. */
   private boolean deleted;
   
   /** Flag indicating that the deletion of this buffer is in progress. */
   private boolean deleting;
   
   /** The last query this buffer was added/set to. */
   private handle lastQuery = null;
   
   /** The current dataset. May be {@code null}. */
   private DataSet dataSet = null;
   
   /** The attached {@code DataSource}, if any. */
   private DataSource dataSource = null;
   
   /** The current value of BATCH-SIZE attribute. Valid only when {@code dataSet} is valid. */
   private int batchSize = 0;
   
   /** Current value of the {@code DATA-SOURCE-MODIFIED} attribute. */
   private boolean dataSourceModified = false;
   
   /** The current value of LAST-BATCH attribute. Valid only when {@code dataSet} is valid. */
   private boolean lastBatch = false;
   
   /**
    * The current value of the {@code FILL-MODE} attribute. The default value is {@code MERGE}.  For buffers
    * not part of a dataset, {@link #fillMode()} will return the unknown value. 
    */
   private String fillMode = FM_MERGE;
   
   /** The current value of AUTO-DELETE attribute. */
   private boolean autoDelete = true;
   
   /**
    * Flag used for bracketing internal management of BEFORE-BUFFER. ABL programmer's direct
    * actions (CREATE/DELETE/ALTER buffer) are outside this bracket and will lead to errors. 
    */
   private boolean setUpBeforeBuffer = false;
   
   /** Is this an AFTER-BUFFER or BEFORE-BUFFER? Cached after value is computed. */
   private BeforeType beforeBufferType = BeforeType.UNKNOWN;
   
   /** The current value of the {@code ATTACHED-PAIRLIST} attribute. */
   private String attachedPairlist = null;
   
   /** The current value of the {@code DATA-SOURCE-COMPLETE-MAP} attribute. */
   private String dataSrcMappingStr = null;
   
   /** The mapping of fields in data source. */
   private Map<CaseInsensitiveString, DataSource.FieldReference> dataSrcMapping = null;
   
   /** A map with all known callbacks. The keys MUST be in UPPERCASE. */
   private final Map<String, CallbackData> callbacks = new HashMap<>();
   
   /** This is not null, when a Buffer proxy overrides the name and a backup is required */
   private character oldName;
   
   /** The list of buffer fields. */
   private BufferFieldImpl[] fields = null;
   
   /** A mapping of fields by their property name. */
   private Map<String, BufferFieldImpl> fieldsByName = new HashMap<>();
   
   /**
    * The cached peer buffer. For AFTER-BUFFER it is the BEFORE-BUFFER and vice-versa. The SIMPLE
    * buffers do not have a peer buffer.
    */
   protected BufferImpl peerBuffer = null;
   
   /** The mode for finding (first, last, unique) records in a dynamic call to a buffer handle. */
   enum FindMode
   {
      FIRST,   /* apply first() to a query */
      LAST,    /* apply last() to a query */
      UNIQUE   /* apply unique() to a query */
   }
   
   /**
    * The default constructor. Calls the super with appropriate parameter values.
    */
   protected BufferImpl()
   {
      super(false, false);
   }
   
   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findFirst() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   private static logical wrapFirst(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.first(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }
   
   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findLast() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   private static logical wrapLast(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.last(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }
   
   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findUnique() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   private static logical wrapUnique(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.unique(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }
   
   /**
    * Checks if transaction is active otherwise records or shows error message.
    * 
    * @param   buffer
    *          Record buffer.
    *
    * @return  <code>true</code> if a transaction is active else <code>false</code>.
    */
   private static boolean verifyTransaction(RecordBuffer buffer)
   {
      if (buffer.getBufferManager().isTransaction())
      {
         return true;
      }
      else
      {
         String msg = "Unable to CREATE/DELETE unless a TRANSACTION is running";
         ErrorManager.recordOrShowError(7345, msg, false);
         
         return false;
      }
   }
   
   /**
   * Get the {@link RecordBuffer} instance associated with this object.
   * <p>
   * This implementation exists only to satisfy the Java compiler. This
   * method is only meaningful when a proxy DMO class is created by {@link
   * RecordBuffer}, which will implement this method to return an instance
   * of <code>RecordBuffer</code> associated with the proxy DMO instance.
   *
   * @return  Record buffer associated with this object.
   */
   public abstract RecordBuffer buffer();
   
   /**
    * Resolve the actual resource used by the runtime.
    * 
    * @return   The instance.
    */
   public BufferImpl ref()
   {
      return this;
   }
   
   /**
    * Get the BUFFER:LAST-BATCH attribute.
    * 
    * @return  the value of BUFFER:LAST-BATCH attribute.
    */
   @Override
   public logical lastBatch()
   {
      return (dataSet == null) ? new logical() : new logical(lastBatch && batchSize > 0);
   }
   
   /**
    * Set the BUFFER:LAST-BATCH attribute.
    * 
    * @param    lastBatch
    *           The flag state.
    */
   @Override
   public void lastBatch(boolean lastBatch)
   {
      if (dataSet == null)
      {
         ErrorManager.recordOrShowError(12360, null, null);
         return;
      }
      
      // all checked, can set value now:
      this.lastBatch = lastBatch; 
   }
   
   /**
    * Set the BUFFER:LAST-BATCH attribute.
    * 
    * @param    lastBatch
    *           The flag state.
    */
   @Override
   public void lastBatch(logical lastBatch)
   {
      if (lastBatch == null || lastBatch.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, "LAST-BATCH", "BUFFER widget");
         return;
      }
      
      lastBatch(lastBatch.booleanValue()); 
   }
   
   /**
    * Get the BUFFER:AUTO-DELETE attribute.
    *
    * @return  the value of BUFFER:AUTO-DELETE attribute.
    */
   @Override
   public logical autoDelete()
   {
      if (!_dynamic() || dataSet == null)
      {
         return new logical();
      }
      
      return new logical(this.autoDelete);
   }
   
   /**
    * Set the BUFFER:AUTO-DELETE attribute.
    *
    * @param    autoDelete
    *           The flag state.
    */
   @Override
   public void autoDelete(boolean autoDelete)
   {
      this.autoDelete = autoDelete;
   }
   
   /**
    * Set the BUFFER:AUTO-DELETE attribute.
    *
    * @param    autoDelete
    *           The flag state.
    */
   @Override
   public void autoDelete(logical autoDelete)
   {
      if (autoDelete == null || autoDelete.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, "AUTO-DELETE", LegacyResource.BUFFER + " widget");
         // **Unable to assign UNKNOWN value to attribute <attribute> on <widget id>.
         return;
      }
      
      autoDelete(autoDelete.booleanValue());
   }
   
   /**
    * Checks whether the data in the data source has been modified.
    *
    * @return  {@code true} when the data has been modified.
    */
   @Override
   public logical isDataSourceModified()
   {
      return new logical(dataSourceModified);
   }
   
   /**
    * Manually mark the data in the target structure as modified.
    *
    * @param   mod
    *          {@code true} when the data is marked as modified and {@code false} otherwise.
    */
   @Override
   public void setDataSourceModified(logical mod)
   {
      if (mod == null || mod.isUnknown())
      {
         unableToAssignUnknown("DATA-SOURCE-MODIFIED", "DATASET widget");
         return;
      }
      
      setDataSourceModified(mod.getValue());
   }
   
   /**
    * Manually mark the data in the target structure as modified.
    *
    * @param   mod
    *          {@code true} when the data is marked as modified and {@code false} otherwise.
    */
   @Override
   public void setDataSourceModified(boolean mod)
   {
      this.dataSourceModified = mod;
   }
   
   /**
    * Reports if this object is valid for use.
    *
    * @return  the state of the {@link #deleted} flag.  If a record buffer has been defined, it
    *          is valid for use until is deleted.
    */
   public boolean valid()
   {
      return !deleted;
   }
   
   /**
    * Check whether the delete operation for this buffer is in progress. 
    * 
    * @return  {@code true} when the delete operation for this buffer is under way. 
    */
   public boolean isDeleting()
   {
      return deleting;
   }
   
   /**
    * Obtain the handle to the buffer that was created by default for the temporary table backing
    * this buffer. Note that this implementation presumes that this is a temp-table buffer. It
    * should not be invoked on a permanent table buffer.
    * 
    * @return  Default buffer for the temporary table backing this buffer, or an unknown handle if
    *          that resource is not valid.
    */
   public handle defaultBufferHandle()
   {
      handle tempTableHandle = buffer().tableHandle();
      if (!tempTableHandle._isValid())
      {
         return new handle();
      }
      
      TempTable tempTable = (TempTable) tempTableHandle.getResource();
      
      return tempTable.defaultBufferHandle();
   }
   
   /**
    * Get the record ID of the current buffer's current record, if any.  P2J
    * implementation of the RECID function.
    * <p>
    * This method will return a 32-bit recid which represents the internal
    * row ID of the given record.  If the actual row ID (which is implemented
    * as a wider integer) cannot fit due to overflow, an error condition will
    * be raised.
    *
    * @return  Record ID of the underlying data record. If no record currently
    *          backs the buffer associated with <code>dmo</code>, a recid
    *          initialized as unknown value is returned.
    * 
    * @throws  ErrorConditionException
    *          if the actual record ID is too wide to fit within a 32-bit
    *          integer.
    */
   public recid recordID()
   {
      return buffer().recordID();
   }
   
   /**
    * Get the row ID of the current buffer's current record, if any.  P2J
    * implementation of the ROWID function.
    *
    * @return  Row ID of the underlying data record.  If no record currently
    *          backs the buffer associated with <code>dmo</code>, a
    *          <code>rowid</code> initialized as unknown value is returned.
    */
   public rowid rowID()
   {
      return buffer().rowID();
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available()
   {
      return (new logical(_available()));
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available()
   {
      return buffer().isAvailable();
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   @LegacyAttribute(name = "CURRENT-CHANGED")
   public logical currentChanged()
   {
      return new logical(_currentChanged());
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged()
   {
      return buffer().isRecordChanged();
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated()
   {
      return (new logical(_newlyCreated()));
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated()
   {
      return buffer().isNewlyCreated();
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked()
   {
      return (new logical(_locked()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked()
   {
      return buffer().wasLocked();
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous()
   {
      return (new logical(_ambiguous()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous()
   {
      return buffer().wasAmbiguous();
   }
   
   /**
    * Getter for the IS-MULTI-TENANT attribute.
    * 
    * @return   See above.
    */
   @Override
   public logical isMultiTenant()
   {
      // TODO: multi-tenant is not supported at this time
      return new logical(false);
   }
   
   /**
    * Setter for the IS-MULTI-TENANT attribute.
    * 
    * @param    l
    *           The flag value.
    */
   @Override
   public void setMultiTenant(logical l)
   {
      // TODO: multi-tenant is not supported at this time
   }
   
   /**
    * Setter for the IS-MULTI-TENANT attribute.
    * 
    * @param    l
    *           The flag value.
    */
   @Override
   public void setMultiTenant(boolean l)
   {
      // TODO: multi-tenant is not supported at this time
   }
   
   /**
    * Perform a RAW-TRANSFER operation from this buffer to the given raw variable. Serializes the
    * record currently in the buffer into the specified field.
    * 
    * @param   r
    *          A <code>raw</code> field to which to copy the record in this buffer.
    */
   public void rawCopyTo(raw r)
   {
      // TODO: this is just a basic implementation, it needs improvements to handle boundary
      //       conditions, add proper error handling is missing and the serialization is 
      //       not binary compatible with the 4GL version
      
      if (r == null)
      {
         return;
      }
      
      boolean error = false;
      
      // obtain the signature for the buffer
      List<TableMapper.LegacyFieldInfo> fields = TableMapper.getLegacyOrderedList(this, true);
      TableSignature sig = new TableSignature(fields);
      
      // collect all serialized output in an in-memory byte[]
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream    oos = null;
      
      try
      {
         // collect all serialized output in an in-memory byte[]
         oos = new ObjectOutputStream(baos);
         
         // serialize the signature
         sig.writeExternal(oos);
         
         RecordBuffer         buffer  = buffer();
         BufferReference      proxy   = buffer.getDMOProxy();
         Map<String, Method>  getters = buffer.getDmoInfo().getLegacyGetterMap();
         Map<String, Integer> extents = buffer.getDmoInfo().getExtentMap();
         
         // iterate through the field descriptors (in POSITION sequence)
         for (TableMapper.LegacyFieldInfo lfi : fields)
         {
            String prop = lfi.getJavaName();
            
            // extent or not?
            Integer extent = extents.get(prop);
            
            // obtain the getter methods for the current property
            Method getter = getters.get(prop);
            
            Externalizable result;
            
            // read the value
            if (extent == null)
            {
               // scalar case
               result = (Externalizable) getter.invoke(proxy);
               
               // serialize
               result.writeExternal(oos);
            }
            else
            {
               // array case
               for (int i = 0; i < extent; i++)
               {
                  result = (Externalizable) getter.invoke(proxy, i);
               
                  // serialize
                  result.writeExternal(oos);
               }
            }
         }
      }
      
      catch (IOException            | 
             SecurityException      |
             IllegalAccessException |
             InvocationTargetException ex)
      {
         error = true;
         ErrorManager.recordOrThrowError(new NumberedException("RAW-TRANSFER failed!", -1, ex));
      }
      
      finally
      {
         // the baos does not need to be closed and can be accessed even after you call close()
         if (oos != null)
         {
            try
            {
               // this one should be closed
               oos.close();
            }
            
            catch (IOException ex)
            {
               // don't care
            }
         }
      }
      
      // save the result in the raw variable
      if (!error)
      {
         r.assign(baos.toByteArray());
      }
   }
   
   /**
    * Perform a RAW-TRANSFER operation from this buffer to the given field, which should be of
    * type <code>raw</code>. Serializes the record currently in the buffer into the specified
    * field.
    * 
    * @param   field
    *          A <code>raw</code> field to which to copy the record in this buffer.
    */
   public void rawCopyTo(FieldReference field)
   {
      if (field == null)
      {
         return;
      }
      
      raw r = new raw();
      
      rawCopyTo(r);
      field.set(r);
   }
   
   /**
    * Perform a RAW-TRANSFER operation from this buffer to the given buffer.
    * 
    * @param   buffer
    *          A buffer into which to copy the record in this buffer.
    * @return <code>true</code> if successful
    */
   public boolean rawCopyTo(Buffer buffer)
   {
      if (buffer == null)
      {
         return false;
      }
      
      raw r = new raw();
      
      rawCopyTo(r);
      return buffer.rawCopyFrom(r);
   }
   
   /**
    * Perform a RAW-TRANSFER operation from the given raw field into this buffer. Deserializes the
    * data in the specified field into this buffer.
    * 
    * @param   field
    *          A <code>raw</code> field from which to copy the record into this buffer.
    */
   public void rawCopyFrom(FieldReference field)
   {
      if (field == null)
      {
         return;
      }
      
      rawCopyFrom((raw) field.get());
   }
   
   /**
    * Perform a RAW-TRANSFER operation from the given raw field into this buffer. Deserializes the
    * data in the specified field into this buffer.
    * 
    * @param   field
    *          A <code>raw</code> field from which to copy the record into this buffer.
    * @return <code>true</code> if successful
    */
   public boolean rawCopyFrom(raw field)
   {
      // TODO: this is just a basic implementation, it needs improvements to handle boundary
      //       conditions, add proper error handling is missing and the deserialization is 
      //       not binary compatible with the 4GL version
      
      if (field == null || field.isUnknown())
      {
         ErrorManager.recordOrShowError(
               4952,
               "Couldn't extract raw value from source field of RAW-TRANSFER",
               false, false, true
         );
         return true;
      }
      
      byte[] data = field.getByteArray();
      
      if (data == null || data.length == 0)
      {
         ErrorManager.recordOrShowError(
               4952,
               "Couldn't extract raw value from source field of RAW-TRANSFER",
               false, false, true
         );
         return true;
      }
      
      // this should create a new record (causing any existing record release and firing
      // triggers as needed
      create();
      
      // obtain the signature for the buffer
      List<TableMapper.LegacyFieldInfo> fields = TableMapper.getLegacyOrderedList(this, true);
      TableSignature tarSig = new TableSignature(fields);
      
      // collect all serialized output in an in-memory byte[]
      ByteArrayInputStream bais = new ByteArrayInputStream(data);
      try (ObjectInputStream ois = new ObjectInputStream(bais))
      {
         // read all serialized input from an in-memory byte[]
         
         TableSignature srcSig = new TableSignature();
         
         // deserialize the signature
         srcSig.readExternal(ois);
         
         if (!srcSig.equals(tarSig))
         {
            ErrorManager.recordOrShowError(
                  4955,
                  "Table signatures do not match in RAW-TRANSFER operation",
                  false, false, true
            );
            return true;
         }
         
         RecordBuffer         buffer  = buffer();
         Map<String, Integer> extents = buffer.getDmoInfo().getExtentMap();
         Map<String, BaseDataType[]> values = readValues(fields, extents, ois);
         if (values == null)
         {
            return true;
         }
         
         BufferReference      proxy   = buffer.getDMOProxy();
         Map<String, Method>  setters = buffer.getDmoInfo().getLegacySetterMap();
         
         boolean batchError = true;
         RecordBuffer.startBatch(true);
         
         try
         {
            // iterate through the field descriptors (in POSITION sequence)
            for (TableMapper.LegacyFieldInfo lfi : fields)
            {
               String prop = lfi.getJavaName();
               
               // extent or not?
               Integer extent = extents.get(prop);
               
               // obtain the setter methods for the current property
               Method setter = setters.get(prop);
               
               Class<? extends BaseDataType> cls = lfi.getDataType();
               BaseDataType value = null;
               
               // process the value
               if (extent == null)
               {
                  // call the default constructor
                  value = cls.newInstance();
                  
                  // deserialize
                  value.readExternal(ois);
                  
                  // scalar case
                  setter.invoke(proxy, value);
               }
               else
               {
                  // array case
                  for (int i = 0; i < extent; i++)
                  {
                     // call the default constructor
                     value = cls.newInstance();
                     
                     // deserialize
                     value.readExternal(ois);
                     
                     // scalar case
                     setter.invoke(proxy, i, value);
                  }
               }
            }
            
            batchError = false;
         }
         finally
         {
            RecordBuffer.endBatch(batchError);
         }
         return true;
      }
      catch (StreamCorruptedException sce)
      {
         ErrorManager.recordOrShowError(
               4952,
               "Couldn't extract raw value from source field of RAW-TRANSFER",
               false, false, true
         );
         return true;
      }
      catch (IOException               |
             ClassNotFoundException    |
             InstantiationException    |
             IllegalAccessException    |
             InvocationTargetException |
             IllegalArgumentException e)
      
      {
         // TODO: if something happened during the copy but after the create, do we need to put
         //       the original record back in?
         
         ErrorManager.recordOrThrowError(new NumberedException("RAW-TRANSFER failed!", -1, e));
      }
      return false;
   }

   /**
    * RAW-TRANSFER() method implementation that maps its processing to the same methods used
    * for implementation of the RAW-TRANSFER statement.  This will serialize data from this
    * buffer into a RAW field or buffer OR it will deserialize data from a RAW field or buffer
    * into this buffer.
    * 
    * @param   toMode
    *          {@code true} if the copy direction is into the handle's resource and {@code false}
    *          if the data is being copied into the buffer.
    * @param   h
    *          A buffer or raw buffer field.
    * 
    * @return  <code>true</code> if RAW-TRANSFER() succeeds, else <code>false</code>.
    */
   @LegacyMethod(name = "RAW-TRANSFER")
   public logical rawTransfer(boolean toMode, handle h)
   {
      if (h == null || h.isUnknown())
      {
         ErrorManager.recordOrShowError(
               4065,
               "**The RAW-TRANSFER attribute on the BUFFER widget has invalid arguments",
               false, false, false
         );
         return new logical(false);
      }

      WrappedResource wr = (WrappedResource) h.get();
      
      if (wr instanceof Buffer)
      {
         Buffer buf = (Buffer) wr;
         
         if (toMode)
         {
            if (!available().booleanValue())
            {
               ErrorManager.recordOrShowError(
                     4951,
                     "Source element of a RAW-TRANSFER statement has no record",
                     false, false, true
               );
               return new logical(false);
            }
            return new logical(rawCopyTo(buf));
         }
         else
         {
            if (!buf.available().booleanValue())
            {
               ErrorManager.recordOrShowError(
                     4951,
                     "Source element of a RAW-TRANSFER statement has no record",
                     false, false, true
               );
               return new logical(false);
            }
            new logical(buf.rawCopyTo(this));
         }
      }
      else if (wr instanceof BufferField)
      {
         BufferField fld = (BufferField) wr;
         String ftype = h.unwrapCommonField().getDataType().getValue();
         if (!"raw".equals(ftype))
         {
            String fn = h.unwrap().name().getValue();
            ErrorManager.recordOrShowError(
                  4824,
                  String.format("Field '%s' referenced in RAW-TRANSFER must be of type RAW", fn),
                  false, false, false
            );
            return new logical(false);
         }
         if (toMode)
         {
            if (!available().booleanValue())
            {
               ErrorManager.recordOrShowError(
                     4951,
                     "Source element of a RAW-TRANSFER statement has no record",
                     false, false, true
               );
               return new logical(false);
            }
            raw r = new raw();
            rawCopyTo(r);
            
            fld.changeValue(r);
         }
         else
         {
            if (!h.unwrapBufferField().available().booleanValue())
            {
               ErrorManager.recordOrShowError(
                     4951,
                     "Source element of a RAW-TRANSFER statement has no record",
                     false, false, true
               );
               return new logical(false);
            }
            BaseDataType bdt = fld.value();
            
//            if (!(bdt instanceof raw))
//            {
//               // TODO: error?
//               return new logical(false);
//            }
            
            return new logical(rawCopyFrom((raw) bdt));
         }
      }
      else
      {
         ErrorManager.recordOrShowError(
               4067,
               "The second argument for RAW-TRANSFER on the BUFFER widget should" +
               " not be of type WIDGET-HANDLE",
               false, true, false
         );
         return new logical(false);
      }
      
      // TODO: we probably need to know if there was a failure in the backing worker methods
      //       but the current signatures don't return a value
      return new logical(true);
   }
   
   /**
    * RAW-TRANSFER() method implementation that maps its processing to the same methods used
    * for implementation of the RAW-TRANSFER statement.  This will serialize data from this
    * buffer into a RAW field or buffer OR it will deserialize data from a RAW field or buffer
    * into this buffer.
    * 
    * @param   toMode
    *          {@code true} if the copy direction is into the handle's resource and {@code false}
    *          if the data is being copied into the buffer.
    * @param   h
    *          A buffer or raw buffer field.
    * 
    * @return  <code>true</code> if RAW-TRANSFER() succeeds, else <code>false</code>.
    */
   @LegacyMethod(name = "RAW-TRANSFER")
   public logical rawTransfer(logical toMode, handle h)
   {
      if (toMode == null || toMode.isUnknown())
      {
         // TODO: check this behavior
         return new logical(false);
      }
      
      return rawTransfer(toMode.booleanValue(), h);
   }
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create()
   {
      buffer().create();
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @return  {@code true} if the delete was possible.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public logical deleteRecord()
   throws ErrorConditionException
   {
      if (isBeforeBuffer() && !setUpBeforeBuffer)
      {
         // the BEFORE-BUFFERS can only be queried, not altered in any way (REMOVE/CHANGE)
         ErrorManager.recordOrThrowError(12374, doGetName(), "");
         // You may not delete BEFORE-TABLE <name> record.  Use ACCEPT-ROW-CHANGES instead.
         return new logical(false);
      }
      
      // no validation is taken in consideration
      return deleteRecord(() -> new logical(true), () -> new character("")); 
   }
   
   /**
    * Delete the record currently in the buffer, IF and ONLY IF the given validation expression
    * evaluates <code>true</code>.
    *
    * @param    valexp
    *           The validation expression to evaluate.
    * @param    valmsg
    *           The error message to display when validation fails.
    *
    * @return   <code>true</code> if validation succeeds, <code>false</code> if it fails and no
    *           exception is raised.
    *
    * @throws   ErrorConditionException
    *           If no record currently is loaded in the buffer, or if there is an error deleting
    *           the current record from the database or if the validation expression fails.
    */
   public logical deleteRecord(Supplier<logical> valexp, String valmsg)
   throws ErrorConditionException
   {
      return deleteRecord(valexp, () -> new character(valmsg));
   }
   
   /**
    * Delete the record currently in the buffer, IF and ONLY IF the given validation expression
    * evaluates <code>true</code>.
    *
    * @param    valexp
    *           The validation expression to evaluate.
    * @param    valmsg
    *           The error message to display when validation fails.
    *
    * @return   <code>true</code> if validation succeeds, <code>false</code> if it fails and no
    *           exception is raised.
    *
    * @throws   ErrorConditionException
    *           If no record currently is loaded in the buffer, or if there is an error deleting
    *           the current record from the database or if the validation expression fails.
    */
   public logical deleteRecord(Supplier<logical> valexp, Supplier<character> valmsg)
   throws ErrorConditionException
   {
      RecordBuffer buffer = buffer();
      
      if (!buffer.isTemporary())
      {
         if (!verifyTransaction(buffer))
         {
            return new logical(false);
         }
         
         // the validation is done at compile type, thus dynamic buffers with validation
         // can't be deleted - also no delete trigger is fired
         if (buffer.isDynamic())
         {
            Class<? extends DataModelObject> iface = buffer.getDMOInterface();
            if (TableMapper.hasValidation(iface))
            {
               String msg = "Table " + buffer.getDmoInfo().legacyTable + " is defined with schema delete " +
                            "validation and may only be deleted by the DELETE statement";
               ErrorManager.recordOrShowError(7347, msg, false, false);
               
               return new logical(false);
            }
         }
      }
      
      if (dataSet != null) // implies buffer.isTemporary()
      {
         // checks whether tracking is active
         if (buffer.getParentTable().isTrackingChanges().toJavaType())
         {
            if (!invokeCallback(ROW_DELETE_EVENT))
            {
               returnError(); // veto on record delete
            }
         }
      }
      
      try
      {
         return new logical(buffer.delete(valexp, valmsg));
      }
      catch (PersistenceException e)
      {
         return new logical(false);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria specified by the
    * given where clause.  Note that this where clause must not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO which is backed by a
    * permanent table;  only temp-table bulk delete is supported at this time.
    * <p>
    * When this method returns:
    * <ul>
    *    <li><strong>temp-table</strong>buffers became empty (Ie. {@code available} returns 
    *          {@code false}). This is because these buffers are used in a loop and the last
    *          record is also deleted. P2J will detect these loops and convert it to the much
    *          faster bulk delete. P4GL does not allow bulk delete using SQL 
    *          {@code DELETE FROM ... WHERE} syntax.
    *    <li>P2J does not support yet bulk deletion of <strong>permanent table</strong> buffers.
    *          In P4GL they will be empty if the predicate matches the currently stored record and
    *          remains unaffected otherwise.
    * </ul>
    *
    * @param   suppDMOs
    *          The DMOs for the external (additional) buffers that are accessed in inner
    *          subselect, or {@code null} in case of a simple {@code where} predicate.
    * @param   where
    *          An HQL where clause snippet which defines the restriction criteria to apply to the
    *          delete.  All references to properties in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(DataModelObject[] suppDMOs, String where, Object... args)
   {
      AbstractQuery.preprocessSubstitutionArguments(BufferManager.get(), () -> false, false, args);
      try
      {
         buffer().delete(suppDMOs, where, args);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria specified by the
    * given where clause.  Note that this where clause must not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO which is backed by a
    * permanent table;  only temp-table bulk delete is supported at this time.
    * <p>
    * When this method returns:
    * <ul>
    *    <li><strong>temp-table</strong>buffers became empty (Ie. {@code available} returns 
    *          {@code false}). This is because these buffers are used in a loop and the last
    *          record is also deleted. P2J will detect these loops and convert it to the much
    *          faster bulk delete. P4GL does not allow bulk delete using SQL 
    *          {@code DELETE FROM ... WHERE} syntax.
    *    <li>P2J does not support yet bulk deletion of <strong>permanent table</strong> buffers.
    *          In P4GL they will be empty if the predicate matches the currently stored record and
    *          remains unaffected otherwise.
    * </ul>
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction criteria to apply to the
    *          delete.  All references to properties in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args)
   {
      delete(null, where, args);
   }
   
   /**
    * Bulk delete all records associated with the buffer.
    * <p>
    * When this method returns, the buffer is empty (Ie. {@link #available} returns {@code false})
    * for both temp and permanent tables.
    * TODO: when is this invoked for permanent tables ?????
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   @Override
   public logical deleteAll()
   {
      if (!(this instanceof Temporary))
      {
         ErrorManager.recordOrShowError(7352);
         // You may only invoke EMPTY-TEMP-TABLE on valid temp-tables.
         return new logical(false);
      }
      
      // need to update the ORIGIN-HANDLE attribute. Will defer execution through default buffer
      return TempTableBuilder.asTempTable((Temporary) this).deleteAll();
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num)
   {
      average(column, num, null);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where)
   {
      average(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where, Object...args)
   {
      buffer().average(column, num, where, args);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num)
   {
      count(column, num, null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where)
   {
      count(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where, Object...args)
   {
      buffer().count(column, num, where, args);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num)
   {
      minimum(column, num, null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where)
   {
      minimum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where, Object...args)
   {
      buffer().minimum(column, num, where, args);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num)
   {
      maximum(column, num, null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where)
   {
      maximum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where, Object...args)
   {
      buffer().maximum(column, num, where, args);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num)
   {
      sum(column, num, null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where)
   {
      sum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where, Object...args)
   {
      buffer().sum(column, num, where, args);
   }
   
   /**
    * Release the record currently in the buffer, if any. No validation, no write trigger is fired. This
    * method is invoked as result of an internal error condition when the content of the buffer failed
    * internal validation and the buffer must be released.
    */
   public void drop()
   {
      buffer().release(true, false);
   }
   
   /**
    * Validate and release the record currently in the buffer, if any. May cause a write trigger to fire.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   @Override
   public void release()
   {
      buffer().release(true);
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   @Override
   public void validate()
   {
      try
      {
         buffer().validate(true);
      }
      catch (ValidationException e)
      {
         // rethrow as 4GL error condition
         ErrorManager.recordOrThrowError(e.getNumber(), e.getMessage(), false, false);
      }
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   @Override
   public logical bufferValidate()
   {
      RecordBuffer recBuf = buffer();
      
      try
      {
         recBuf.validate(true);
         return new logical(true);
      }
      catch (ValidationException e)
      {
         if (recBuf.isNewlyCreated())
         {
            ErrorManager.recordOrShowError(e.getNumber(), e.getMessage(), false, false, false);
         }
      }
      
      return new logical(false);
   }
   
   /**
    * Conversion of the RECORD-LENGTH built-in function.
    * 
    * @return  Length of the record currently in the buffer, in bytes.
    */
   @Override
   public integer recordLength()
   {
      if (!_available())
      {
         return new integer(); // quick out
      }
      
      RecordBuffer         buffer  = buffer();
      BufferReference      proxy   = buffer.getDMOProxy();
      Map<String, Method>  getters = buffer.getDmoInfo().getLegacyGetterMap();
      Map<String, Integer> extents = buffer.getDmoInfo().getExtentMap();
      
      int counter = 15; // hidden 2 fields and any header
      int fieldCounter = getters.size() + 2;
      if (fieldCounter > 15)
      {
         counter += 3;
      }
      counter += 2 * ((fieldCounter + 1) / 16);
         
      for (Map.Entry<String, Method> entry : getters.entrySet())
      {
         String prop = entry.getKey();
         Method method = entry.getValue();
         Integer ext = extents.get(prop);
         if (ext == null)
         {
            try
            {
               BaseDataType bdt = (BaseDataType) method.invoke(proxy);
               counter += bdt.getSize();
            }
            catch (IllegalAccessException | InvocationTargetException e)
            {
               // log it?
            }
         }
         else 
         {
            counter += 4;  // probably one 32bit int = the size of the extent 
            if (ext > 15)
            {
               counter += 3;
            }
            counter += 2 * ((ext + 1) / 16);
            
            for (int k = 0; k < ext; k++) 
            {
               BaseDataType bdt = null;
               try
               {
                  bdt = (BaseDataType) method.invoke(proxy, k);
                  counter += bdt.getSize();
               }
               catch (IllegalAccessException | InvocationTargetException e)
               {
                  // log it ?
               }
            }
         }
      }
      
      return new integer(counter);
   }
   
   /**
    * Resolve the BUFFER-FIELD with the given converted DMO property name.
    * 
    * @param     propertyName
    *            The property name.
    *            
    * @return    See above.
    */
   public BufferFieldImpl bufferFieldForProperty(String propertyName)
   {
      BufferFieldImpl field = fieldsByName.get(propertyName);
      
      if (field == null)
      {
         RecordBuffer recBuf = buffer();
         int fieldId = TableMapper.getLegacyFieldId(recBuf, propertyName).intValue();

         field = new BufferFieldImpl((DataModelObject) this, propertyName);
         
         if (fields == null)
         {
            int numFields = _numFields();
            fields = new BufferFieldImpl[numFields + 1];
         }

         fields[fieldId] = field;
         fieldsByName.put(propertyName, field);
      }
      
      return field;
   }
   
   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or DataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(int fieldNumber)
   {
      int fieldCount = _numFields();
      RecordBuffer recBuf = buffer();
      String legacyTableName = recBuf.getDmoInfo().legacyTable;
      if (fieldNumber <= 0 || fieldNumber > fieldCount)
      {
         String msg = "BUFFER-FIELD sequence number must be from 1 to " +
                      fieldCount + " for buffer " + legacyTableName;
         ErrorManager.recordOrShowError(7350, msg, false, false, false);
         
         return new handle();
      }

      BufferFieldImpl field = fields == null ? null : fields[fieldNumber];
      if (field == null)
      {
         String hibernateName;
         TempTable parentTempTable = recBuf.getParentTable();
         if (parentTempTable != null && parentTempTable._dynamic())
         {
            // dynamic table
            hibernateName = DynamicTablesHelper.getHibernateFieldName(recBuf.getDMOInterface(), fieldNumber);
         }
         else
         {
            // static temporary or permanent table
            if (recBuf.isTemporary())
            {
               hibernateName = TableMapper.getPropertyName(parentTempTable, fieldNumber);
            }
            else
            {
               // static temporary or permanent table
               hibernateName = TableMapper.getPropertyName(recBuf.getDMOInterface(), fieldNumber);
            }
         }
         
         field = new BufferFieldImpl((DataModelObject) this, hibernateName);
         
         if (fields == null)
         {
            int numFields = _numFields();
            fields = new BufferFieldImpl[numFields + 1];
         }
         fields[fieldNumber] = field;
         fieldsByName.put(field.getProperty(), field);
      }
      
      return new handle(field);
   }
   
   /**
    * Implementation of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular field within a buffer.
    *
    * @param   fieldName
    *          The legacy field name. A CHARACTER string expression representing the name
    *          of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(String fieldName)
   {
      Property property = buffer().getDmoInfo().byLegacyName(fieldName);
      if (property != null)
      {
         BufferFieldImpl field = bufferFieldForProperty(property.name);
         if (field != null)
         {
            return new handle(field);
         }
      }

      String msg = "BUFFER-FIELD " + fieldName + " was not found in buffer " + doGetName();
      ErrorManager.recordOrShowError(7351, msg, false, false, false);
      
      // return unknown handle
      return new handle();
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer)
   {
      return bufferCopy(buffer, null, null);
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    *
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except)
   {
      return bufferCopy(buffer, except, null);
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs)
   {
      return bufferCopy(buffer, except, pairs, new logical(false));
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs, logical noLobs)
   {
      return RecordBuffer.copy(buffer(), ((BufferImpl) buffer).buffer(), except, pairs, noLobs);
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle)
   {
      return bufferCopy((Buffer) bufHandle.get());
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except)
   {
      return bufferCopy((Buffer) bufHandle.get(), except);
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except, character pairs)
   {
      return bufferCopy((Buffer) bufHandle.get(), except, pairs);
   }
   
   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle,
                             character except,
                             character pairs,
                             logical noLobs)
   {
      return bufferCopy((Buffer) bufHandle.get(), except, pairs, noLobs);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer)
   {
      return bufferCompare(buffer, null, null, null);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode)
   {
      return bufferCompare(buffer, mode, null, null);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode, character except)
   {
      return bufferCompare(buffer, mode, except, null);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   srcBuffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer srcBuffer,
                                character mode,
                                character except,
                                character pairs)
   {
      return bufferCompare(srcBuffer, mode, except, pairs, null);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   srcBuffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If {@code true}, BLOB and CLOB fields are ignored during the comparison. May be
    *          {@code false} or {@code null} to include such fields.
    *
    * @return  {@code true} if BUFFER-COMPARE() succeeds, else {@code false}.
    */
   public logical bufferCompare(Buffer    srcBuffer,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      return RecordBuffer.compare(((BufferImpl) srcBuffer).buffer(),
                                  buffer(),
                                  BufferCompare.Mode.fromString(mode),
                                  except,
                                  pairs,
                                  noLobs);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle)
   {
      return bufferCompare((Buffer) bufHandle.get());
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode, character except)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode, character except, character pairs)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except, pairs);
   }
   
   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle    bufHandle,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except, pairs, noLobs);
   }
   
   /**
    * Conversion of BUFFER-CREATE() method (KW_BUF_CREA).
    *
    * Creates a record, sets fields to their default values, and moves a copy of the record into the buffer.
    *
    * @return  {@code true} if successful
    */
   public logical bufferCreate()
   {
      if (isBeforeBuffer() && !setUpBeforeBuffer)
      {
         // the BEFORE-BUFFERS can only be queried, not altered in any way (REMOVE/CHANGE)
         ErrorManager.recordOrThrowError(12373, doGetName(), "");
         // You may not delete BEFORE-TABLE <name> record.  Use ACCEPT-ROW-CHANGES instead.
         return new logical(false);
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() && !verifyTransaction(buffer))
      {
         return new logical(false);
      }
      
      return new logical(buffer.create());
   }
   
   /**
    * Conversion of BUFFER-RELEASE() method (KW_BUF_REL).
    *
    * Releases a record from a buffer object. The BUFFER-RELEASE method
    * corresponds to the RELEASE statement.
    *
    * @return  <code>true</code> if BUFFER-RELEASE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferRelease()
   {
      boolean success = true;
      try
      {
         buffer().release(true);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      
      return new logical(success);
   }
   
   /**
    * Dereference a named member from this collection.
    *
    * @param   memberName
    *          The legacy name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public BaseDataType dereference(String memberName)
   {
      Property property = buffer().getDmoInfo().byLegacyName(memberName);
      
      if (property == null)
      {
         ErrorManager.recordOrShowError(7351, memberName, doGetName());
         // BUFFER-FIELD <field-name> was not found in buffer <buffer-name>.
         return new unknown();
      }
      
      if (!buffer().isAvailable())
      {
         String msg1 = "No " + doGetName() + " record is available";
         ErrorManager.displayError(91, msg1, true);
         
         String msg2 = "Unable to extract BUFFER-VALUE for field " + memberName;
         ErrorManager.displayError(7366, msg2, false);
         
         return new unknown();
      }
      
      BaseDataType dbt = bufferFieldForProperty(property.name).value();
      if (dbt instanceof character)
      {
         // quirk in 4GL: the character fields retrieved using dereference operator are always
         // case-insensitive (no matter the initial field property case sensitivity)
         ((character) dbt).setCaseSensitive(false);
      }
      return dbt;
   }
   
   /**
    * Dereference a named member from this collection. If this collection is a ProDataSet then 
    * the returned value is a {@link handle} to a named buffer, if the collection is a buffer
    * the returned value is the field value with the respective name. In the case that this 
    * collection does not have a named member with the name specified <code>memberName</code>
    * the unknown value is returned.
    *  
    * @param   index
    *          The subscript to be applied to an extent field.
    * @param   memberName
    *          The name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public BaseDataType dereference(NumberType index, String memberName)
   {
      Property property = buffer().getDmoInfo().byLegacyName(memberName);
      
      if (property == null)
      {
         ErrorManager.recordOrShowError(7351, memberName, doGetName());
         // BUFFER-FIELD <field-name> was not found in buffer <buffer-name>.
         return new unknown();
      }
      
      if (!buffer().isAvailable())
      {
         String msg1 = "No " + doGetName() + " record is available";
         ErrorManager.displayError(91, msg1, true);
         
         String msg2 = "Unable to extract BUFFER-VALUE for field " + memberName;
         ErrorManager.displayError(7366, msg2, false);
         
         return new unknown();
      }
      
      if (property.extent == 0)
      {
         return dereference(memberName);
      }

      if (index.isUnknown())
      {
         return new unknown();
      }

      long idx = index.longValue();
      if (idx == 0)
      {
         // no error is shown/raised
         return new unknown();
      }
      
      if (idx < 1 || idx > property.extent)
      {
         String msg1 = "Array subscript " + index + " is out of range";
         ErrorManager.displayError(26, msg1, true);
         
         String msg2 = "Unable to extract BUFFER-VALUE for field " + memberName;
         ErrorManager.displayError(7366, msg2, false);
         
         return new unknown();
      }
      
      return bufferFieldForProperty(property.name).value((int) idx);
   }
   
   /**
    * Dereference a named member from this collection. If this collection is a ProDataSet then 
    * the returned value is a {@link handle} to a named buffer, if the collection is a buffer
    * the returned value is the field value with the respective name. In the case that this 
    * collection does not have a named member with the name specified <code>memberName</code>
    * the unknown value is returned.
    *  
    * @param   index
    *          The subscript to be applied to an extent field.
    * @param   memberName
    *          The name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public BaseDataType dereference(long index, String memberName)
   {
      return dereference(new int64(index), memberName);
   }
   
   /**
    * Dereference a named member from this collection. If this collection is a DataSet then
    * the returned value is a {@link handle} to a named buffer, if the collection is a buffer
    * the returned value is the field value with the respective name. In the case that this
    * collection does not have a named member with the name specified {@code memberName}
    * the unknown value is returned.
    * 
    * @param   type
    *          The expected type of the returned value.
    * @param   memberName
    *          The name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public <T extends BaseDataType> T dereference(Class<T> type, String memberName)
   {
      BaseDataType val = dereference(memberName);
      if (type.isAssignableFrom(val.getClass()))
      {
         return (T) val;
      }
      
      String errMsg2 =
          "Lead attributes in a chained-attribute expression (a:b:c) must be type " +
          "HANDLE or a user-defined type and valid (not UNKNOWN)";
      ErrorManager.recordOrThrowError(10068, errMsg2, false, false);
      
      return (T) BaseDataType.generateUnknown(type);
   }
   
   /**
    * This overloaded method is used when the dereference is the left-side of an assignment.
    * The implementation will extract the member from the collection and assign it the new value.
    *
    * @param   memberName
    *          The name of the member to be extracted.
    * @param   value
    *          The new value that will be assigned to the extracted field.
    */
   @Override
   public void dereference(String memberName, Object value)
   {
      if (isBeforeBuffer() && !setUpBeforeBuffer)
      {
         ErrorManager.recordOrShowError(12375, doGetName());
         // You may not update BEFORE-TABLE <name> record.
         ErrorManager.recordOrShowError(142, doGetName());
         // ** Unable to update <filename> Field.
         ErrorManager.recordOrShowError(3131, "?", "", "BUFFER-FIELD");
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>
         ErrorManager.recordOrThrowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         return;
      }
      
      Property property = buffer().getDmoInfo().byLegacyName(memberName);
      if (property == null)
      {
         String errMsg  = "BUFFER-FIELD " + memberName + " was not found in buffer " + doGetName();
         ErrorManager.recordOrThrowError(7351, errMsg);
         
         return;
      }
      
      boolean bulk = value.getClass().isArray();
      if (bulk)
      {
         BufferFieldImpl field = new BufferFieldImpl((DataModelObject) this, property.name, bulk);
         try
         {
            field.changeValue(value);
         }
         finally
         {
            // force the delete, this instance is required for bulk assign of an extent field, and is not a 
            // real BUFFER-FIELD resource
            field.forceDelete();
         }
      }
      else
      {
         BufferFieldImpl field = bufferFieldForProperty(property.name);
         field.changeValue(value);
      }
   }
   
   /**
    * This overloaded method is used when the dereference is the left-side of an assignment.
    * The implementation will extract the member from the collection and assign it the new value.
    *
    * @param   index
    *          The subscript to be applied to an extent field.
    * @param   memberName
    *          The name of the member to be extracted.
    * @param   value
    *          The new value that will be assigned to the extracted field.
    */
   @Override
   public void dereference(NumberType index, String memberName, Object value)
   {
      if (isBeforeBuffer() && !setUpBeforeBuffer)
      {
         ErrorManager.recordOrShowError(12375, doGetName());
         // You may not update BEFORE-TABLE <name> record.
         ErrorManager.recordOrShowError(142, doGetName());
         // ** Unable to update <filename> Field.
         ErrorManager.recordOrShowError(3131, "?", "", "BUFFER-FIELD");
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>
         ErrorManager.recordOrThrowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         return;
      }
      
      Property property = buffer().getDmoInfo().byLegacyName(memberName);
      if (property == null)
      {
         String errMsg  = "BUFFER-FIELD " + memberName + " was not found in buffer " + doGetName();
         ErrorManager.recordOrThrowError(7351, errMsg);
         
         return;
      }
      
      if (index.isUnknown())
      {
         ErrorManager.recordOrShowError(7365, "f1", "" + property.extent);
         // <field> BUFFER-VALUE argument must be valid array index, less than or equal to field extent <extent>
         ErrorManager.recordOrShowError(3131, "?", "", "BUFFER-FIELD");
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>
         ErrorManager.recordOrThrowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         return;
      }
      
      long idx = index.longValue();
      if (property.extent == 0 && idx == 0)
      {
         dereference(memberName, value);
         return;
      }
      
      if (idx == 0)
      {
         ErrorManager.recordOrShowError(26, "" + idx);
         // ** Array subscript <index> is out of range. (26)
         ErrorManager.recordOrShowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         ErrorManager.recordOrShowError(3131, "?", "", "BUFFER-FIELD");
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>
         ErrorManager.recordOrThrowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         return;
      }
      
      if (idx < 1 || idx > property.extent)
      {
         ErrorManager.recordOrShowError(7365, memberName, "" + property.extent, "BUFFER-FIELD");
         // <field> BUFFER-VALUE argument must be valid array index, less than or equal to field extent <extent>
         ErrorManager.recordOrShowError(3131, "?", "", "BUFFER-FIELD");
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>
         ErrorManager.recordOrThrowError(142, doGetName(), "");
         // ** Unable to update <filename> Field.
         return;
      }
      
      bufferFieldForProperty(property.name).changeValue((int) idx, value);
   }
   
   /**
    * This overloaded method is used when the dereference is the left-side of an assignment.
    * The implementation will extract the member from the collection and assign it the new value.
    *
    * @param   index
    *          The subscript to be applied to an extent field.
    * @param   memberName
    *          The name of the member to be extracted.
    * @param   value
    *          The new value that will be assigned to the extracted field.
    */
   @Override
   public void dereference(long index, String memberName, Object value)
   {
      dereference(new int64(index), memberName, value);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id)
   {
      return findByRowID(id, LockType.LT_SHARE);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lockType
    *          Lock type
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    */
   public logical findByRowID(rowid id, LockType lockType)
   {
      final RecordBuffer buffer = buffer();
      final String dmoAlias = buffer.getDMOAlias();
      
      if (id.isUnknown())
      {
         String msg = "Invalid rowid for FIND-BY-ROWID method of BUFFER object " + buffer.getLegacyName();
         ErrorManager.recordOrShowError(7341, msg, false);
         
         return new logical(false);
      }
      
      try
      {
         boolean err = ErrorManager.silent(() ->
         {
            buffer.initialize();
            
            // manufacture a primary key sort clause; this will be discarded from the final FQL, but it is
            // needed by RAQ.initialize
            String sort = dmoAlias + "." + Session.PK + " asc";
            
            FindQuery findQuery = new FindQuery(buffer,
                                                dmoAlias + "." + Session.PK + " = ?",
                                                null,
                                                sort,
                                                new Object[] { id },
                                                lockType);
            findQuery.unique();
         });
         
         return new logical(!err);
      }
      catch (IllegalArgumentException | ErrorConditionException | QueryOffEndException exc)
      {
         // all ignored
      }
      
      return new logical(false);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  {@code true} if it finds the record, else {@code false}.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findByRowID(rowid id, integer lock, int nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst()
   {
      return _find(null, FindMode.FIRST, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, LockType.fromInt((int) lock));
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate, NumberType lock)
   {
      LockType lockType = lock.isUnknown() ? LockType.SHARE : LockType.fromInt(lock.intValue());
      
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findFirst(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findFirst(character predicate, NumberType lock, long nowait)
   {
      long javaLock = lock.isUnknown() ? LockType.LT_SHARE : lock.longValue();
      
      LockType lockType = fromLockNowait(javaLock, nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findFirst(character predicate, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
        ? new logical(false)
        : _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findFirst(character predicate, NumberType lock, NumberType nowait)
   {
      long javaLock = lock.isUnknown() ? LockType.LT_SHARE : lock.longValue();
      
      LockType lockType = fromLockNowait(javaLock, nowait.longValue());
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets the last record.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast()
   {
      return _find(null, FindMode.LAST, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.fromInt((int) lock));
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, NumberType lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.fromInt(lock.intValue()));
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findLast(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findLast(character predicate, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findLast(character predicate, long lock, NumberType nowait)
   {
      // TODO: add support for ? nowait mode
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }
   
   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findLast(character predicate, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This make sense only when there is only one record in the table.
    * Otherwise the error 3166 "More than one &lt;tablename&gt; records found by a unique FIND."
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique()
   {
      return _find(null, FindMode.UNIQUE, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.SHARE);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.fromInt((int) lock));
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, NumberType lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.fromInt(lock.intValue()));
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findUnique(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findUnique(character predicate, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findUnique(character predicate, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      
      // toJavaType() returns null if predicate is unknown
      return lockType == null
         ? new logical(false)
         : _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }
   
   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   public logical findUnique(character predicate, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent()
   {
      return _current(LockType.SHARE);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lockType
    *          Lock type.
    * 
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(LockType lockType)
   {
      if (lockType == null)
      {
         return new logical(false);
      }
      
      return _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lockType
    *          Lock type.
    * @param   noWait
    *          Zero to lock without waiting; non-zero to wait to acquire the lock.
    * 
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(LockType lockType, long noWait)
   {
      if (lockType == null)
      {
         return new logical(false);
      }
      lockType = fromLockNowait(LockType.toInt(lockType), noWait);
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    * @param   noWait
    *          Zero to lock without waiting; non-zero to wait to acquire the lock.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(NumberType lock, long noWait)
   {
      if (lock == null)
      {
         return new logical(false);
      }
      LockType lockType = fromLockNowait(lock.longValue(), noWait);
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    * @param   noWait
    *          Zero to lock without waiting; non-zero to wait to acquire the lock.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(long lock, long noWait)
   {
      LockType lockType = fromLockNowait(lock, noWait);
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lockType
    *          Lock type.
    * @param   noWait
    *          Zero or unknown value to lock without waiting; non-zero to wait to acquire the
    *          lock.
    * 
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(LockType lockType, NumberType noWait)
   {
      if (lockType == null || noWait == null)
      {
         return new logical(false);
      }
      lockType = fromLockNowait(LockType.toInt(lockType), noWait.longValue());
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    * @param   noWait
    *          Zero or unknown value to lock without waiting; non-zero to wait to acquire the
    *          lock.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(NumberType lock, NumberType noWait)
   {
      if (lock == null || noWait == null)
      {
         return new logical(false);
      }
      LockType lockType = fromLockNowait(lock.longValue(), noWait.longValue());
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of the FIND-CURRENT() method.
    * 
    * Changes the lock type for the record currently in the buffer. Resets the CURRENT-CHANGED
    * attribute.
    * 
    * @param   lock
    *          Lock type.
    * @param   noWait
    *          Zero or unknown value to lock without waiting; non-zero to wait to acquire the
    *          lock.
    *          
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else {@code
    *          false}.
    */
   @LegacyMethod(name = "FIND-CURRENT")
   public logical findCurrent(long lock, NumberType noWait)
   {
      if (noWait == null)
      {
         return new logical(false);
      }
      LockType lockType = fromLockNowait(lock, noWait.longValue());
      
      return (lockType == null) ? new logical(false) : _current(lockType);
   }
   
   /**
    * Conversion of NUM-FIELDS attribute (KW_NUM_FLD).
    *
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   public integer numFields()
   {
      return new integer(_numFields());
   }
   
   /**
    * Disable the triggers associated with the FIND event for the buffer's table.
    */
   public void disableDumpTriggers()
   {
      final Class<?> dmoBuf = buffer().getDMOBufInterface();
      if (Buffer.class.isAssignableFrom(dmoBuf))
      {
         DatabaseTriggerManager.disableDumpTriggers((Class<? extends Buffer>) dmoBuf);
      }
   }
   
   /**
    * Disable the triggers associated with the CREATE, WRITE, DELETE and ASSIGN events for the
    * buffer's table.
    *
    * @param   allowReplication
    *          If <code>true</code> then only CREATE, DELETE, ASSIGN and WRITE triggers are
    *          disabled, and not REPLICATION-CREATE, REPLICATION-DELETE and REPLICATION-WRITE
    *          triggers.
    */
   public void disableLoadTriggers(boolean allowReplication)
   {
      final Class<?> dmoBuf = buffer().getDMOBufInterface();
      if (Buffer.class.isAssignableFrom(dmoBuf))
      {
         DatabaseTriggerManager.disableLoadTriggers((Class<? extends Buffer>) dmoBuf, allowReplication);
      }
   }
   
   /**
    * Conversion of CURRENT-ITERATION attribute (KW_CUR_ITER).
    *
    * Getter of CURRENT-ITERATION attribute.
    *
    * @return  Current value of CURRENT-ITERATION attribute.
    */
   public handle currentIteration()
   {
      // TODO: implement, returning unknown values for the moment
      return new handle();
   }
   
   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param errStr The new value for the ERROR-STRING attribute.
    */
   @Override
   public void changeErrorString(character errStr)
   {
      RecordBuffer buf = buffer();
      if (!buf.isTemporary())
      {
         // permanent table case
         ErrorManager.recordOrShowError(11917, doGetName());
         // Cannot set ERROR if there is no BEFORE-TABLE for <name>.
         ErrorManager.recordOrShowError(3131, "ERROR-STRING", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
         return;
      }
      
      if (!_available())
      {
         // temp-tables case
         ErrorManager.recordOrShowError(11897, doGetName());
         // <name> record is not available to set ERROR in.
         ErrorManager.recordOrShowError(3131, "ERROR-STRING", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
      }
      
      if (!isBeforeBuffer())
      {
         if (!isAfterBuffer() || // SIMPLE BUFFER
             (buf.peerRowid() == null || buf.peerRowid().isUnknown())) // no AFTER-IMAGE
         {
            ErrorManager.recordOrShowError(11918, doGetName(), "");
            // Cannot set ERROR if there is no BEFORE-TABLE row for <name>.
            return;
         }
      }
      
      buffer().errorString(errStr);
   }
   
   /**
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    *
    * @return the current ERROR-STRING of the temp-table
    */
   @Override
   public character errorString()
   {
      RecordBuffer recBuf = buffer();
      if (!recBuf.isAvailable() || !recBuf.isTemporary())
      {
         return new character();
      }
      
      DataSet ds = _dataSet();
      if (ds == null)
      {
         return new character();
      }
      
      return recBuf.errorString();
   }
   
   /**
    * Returns the state of the ERROR attribute
    *
    * @return  Current state of the ERROR attribute.
    */
   @Override
   public logical error()
   {
      RecordBuffer recBuf = buffer();
      if (!recBuf.isAvailable() || !recBuf.isTemporary())
      {
         return new logical(false);
      }
      
      DataSet ds = _dataSet();
      if (ds == null)
      {
         return new logical(false);
      }
      
      // NOTE: internally, the P4GL database the __error-flag__ is an integer storing multiple flags.
      //       We need to extract the right bit (ERROR_ERROR) and convert it to a logical value.
      Integer flags = recBuf.errorFlags();
      return new logical(flags != null && (flags & Buffer.ERROR_ERROR) != 0);
   }
   
   /**
    * Setter for {@code ERROR} attribute.<br>
    * The {@code ERROR} attribute is writable for a {@code Buffer} object. The attribute can be set
    * programmatically in order to manage the application response to errors.<br>
    * If set to {@code FALSE}, the application will ignore all errors that might be returned for a
    * statement using the {@code NO-ERROR} option.<br>
    * However, manually setting this attribute won't raise the {@code ERROR} condition.
    *
    * @param    value
    *           The new value of {@code ERROR} attribute.
    */
   @Override
   public void error(boolean value)
   {
      RecordBuffer recBuf = buffer();
      if (!_available())
      {
         // temp-tables case
         ErrorManager.recordOrShowError(11916, doGetName());
         // <name> record is not available to set ERROR in.
         ErrorManager.recordOrShowError(3131, "ERROR", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
      }
      
      if (!recBuf.isTemporary())
      {
         // permanent table case
         ErrorManager.recordOrShowError(11917, doGetName());
         // Cannot set ERROR if there is no BEFORE-TABLE for <name>.
         ErrorManager.recordOrShowError(3131, "ERROR", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
         return;
      }
      
      if (!isBeforeBuffer())
      {
         if (!isAfterBuffer() || // SIMPLE BUFFER
             (recBuf.peerRowid() == null || recBuf.peerRowid().isUnknown())) // no AFTER-IMAGE
         {
            ErrorManager.recordOrShowError(11918, doGetName(), "");
            // Cannot set ERROR if there is no BEFORE-TABLE row for <name>.
            return;
         }
      }
      
      // NOTE: internally, the P4GL database the __error-flag__ is an integer storing multiple flags.
      //       We need to update the right bit (ERROR_ERROR) to new state (value).
      recBuf.updateErrorFlags(Buffer.ERROR_ERROR, value);
   }
   
   /**
    * Setter for ERROR attribute.
    *
    * @param   value
    *          The new value of ERROR attribute.
    */
   @Override
   public void error(logical value)
   {
      error(value.booleanValue());
   }
   
   /**
    * Returns the state of the DBNAME attribute
    *
    * @return  Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      final RecordBuffer buf = buffer();
      if (buf.getParentTable() != null)
      {
         // in the case of temp-tables (defined statically or dynamically)
         return new character("PROGRESS");
      }
      
      // permanent tables, it works as expected (the logical name of the database of the buffer)
      return new character(buf.getLogicalDatabase());
   }
   
   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   @Override
   public character getTable()
   {
      return new character(TableMapper.getLegacyName(this.buffer()));
   }
   
   /**
    * Conversion of TABLE-HANDLE attribute (KW_TAB_HAND).
    *
    * Gets the handle of a temp-table object, if any, associated with the buffer object.
    *
    * @return  the handle of a temp-table object.
    */
   public handle tableHandle()
   {
      return buffer().tableHandle();
   }
   
   /**
    * Conversion of INDEX-INFORMATION method. Returns information about the first (according to
    * the definition order) table index, as a string.
    *
    * @return Information about the first (according to definition order) index, as a string. The
    *         returned comma-separated list consists of the following in the specified order:
    *         <ul>
    *         <li>the index name;
    *         <li>three integers of value 0 (false) or 1 (true) depending on whether the index is
    *             unique, primary or a word index;
    *         <li>the names of the index fields, each followed by a 0 (ascending) or 1
    *             (descending).
    *         </ul>
    *         An example: "index-one,1,1,0,field-one,0".
    */
   public character indexInformation()
   {
      return indexInformation(new integer(1));
   }
   
   /**
    * Conversion of INDEX-INFORMATION method. Returns information about the specified table index,
    * as a string.
    *
    * @param  indexNum
    *         1-based number of the index in the definition order.
    *
    * @return Information about the specified index, as a string. The returned comma-separated
    *         list consists of the following in the specified order:
    *         <ul>
    *         <li>the index name;
    *         <li>three integers of value 0 (false) or 1 (true) depending on whether the index is
    *             unique, primary or a word index;
    *         <li>the names of the index fields, each followed by a 0 (ascending) or 1
    *             (descending).
    *         </ul>
    *         An example: "index-one,1,1,0,field-one,0".
    */
   public character indexInformation(int indexNum)
   {
      return indexInformation(new integer(indexNum));
   }
   
   /**
    * Conversion of INDEX-INFORMATION method. Returns information about the specified table index,
    * as a string.
    *
    * @param  indexNum
    *         1-based number of the index in the definition order.
    *
    * @return Information about the specified index, as a string. The returned comma-separated
    *         list consists of the following in the specified order:
    *         <ul>
    *         <li>the index name;
    *         <li>three integers of value 0 (false) or 1 (true) depending on whether the index is
    *             unique, primary or a word index;
    *         <li>the names of the index fields, each followed by a 0 (ascending) or 1
    *             (descending).
    *         </ul>
    *         An example: "index-one,1,1,0,field-one,0".
    */
   public character indexInformation(NumberType indexNum)
   {
      if (indexNum.isUnknown())
      {
         return new character();
      }
      
      return new character(TableMapper.getLegacyIndexInfoString(this.buffer(), indexNum.intValue() - 1));
   }
   
   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return <code>true</code> if the buffer is dynamic.
    */
   public boolean _dynamic()
   {
      return buffer().isDynamic();
   }
   
   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return <code>true</code> if the buffer is dynamic.
    */
   public logical dynamic()
   {
      return new logical(_dynamic());
   }
   
   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }
   
   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }
   
   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param   value
    *          The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }
   
   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return  An integer representing an unique id between all buffers ever created in the 
    *          current process.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }
   
   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return  See above.
    */
   @Override
   public character namespaceURI()
   {
      return buffer().getParentTable().namespaceURI();
   }
   
   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param   uri
    *          The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void namespaceURI(String uri)
   {
      buffer().getParentTable().namespaceURI(uri);
   }
   
   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param   uri
    *          The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void namespaceURI(character uri)
   {
      buffer().getParentTable().namespaceURI(uri);
   }
   
   /**
    * Returns the value of the the NAMESPACE-PREFIX attribute. Only for ELEMENT or ATTRIBUTE,
    * for other attributes return unknown value.
    *
    * @return  See above.
    */
   public character namespacePrefix()
   {
      return buffer().getParentTable().namespacePrefix();
   }
   
   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param   prefix
    *          The new value of the NAMESPACE-PREFIX attribute.
    */
   public void namespacePrefix(String prefix)
   {
      buffer().getParentTable().namespacePrefix(prefix);
   }
   
   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param   prefix
    *          The new value of the NAMESPACE-PREFIX attribute.
    */
   public void namespacePrefix(character prefix)
   {
      buffer().getParentTable().namespacePrefix(prefix);
   }
   
   /**
    * Implementation of the read access of the {@code SERIALIZE-NAME} attribute.
    *
    * @return  The current value of the {@code SERIALIZE-NAME} attribute.
    */
   @Override
   public character getSerializeName()
   {
      return buffer().getParentTable().getSerializeName();
   }
   
   /**
    * Implementation of the write access of the {@code SERIALIZE-NAME} attribute.
    *
    * @param   sName
    *          the new value for the {@code SERIALIZE-NAME} attribute.
    */
   @Override
   public void setSerializeName(Text sName)
   {
      buffer().getParentTable().setSerializeName(sName);
   }
   
   /**
    * Implementation of the write access of the {@code SERIALIZE-NAME} attribute.
    *
    * @param  sName
    *         The new value for the {@code SERIALIZE-NAME} attribute.
    */
   @Override
   public void setSerializeName(String sName)
   {
      buffer().getParentTable().setSerializeName(sName);
   }
   
   /**
    * Obtain the name of the XML element or attribute representing the target object (either a
    * {@code DataSet}, a {@code TempTable}, a {@code TemporaryBuffer}, or a temp-table
    * {@code BufferField}) name in an XML document.
    *
    * @return  the name of the XML element for this object.
    */
   public character getXmlNodeName()
   {
      return buffer().getParentTable().getXmlNodeName();
   }
   
   /**
    * Sets the name of the XML element or attribute representing the target object (either a
    * {@code DataSet}, a {@code TempTable}, a {@code TemporaryBuffer}, or a temp-table
    * {@code BufferField}) name in an XML document.
    *
    * @param   name
    *          The new name of the XML element for this object.
    */
   public void setXmlNodeName(String name)
   {
      buffer().getParentTable().setXmlNodeName(name);
   }
   
   /**
    * Sets the name of the XML element or attribute representing the target object (either a
    * {@code DataSet}, a {@code TempTable}, a {@code TemporaryBuffer}, or a temp-table
    * {@code BufferField}) name in an XML document.
    *
    * @param   name
    *          The new name of the XML element for this object.
    */
   public void setXmlNodeName(Text name)
   {
      buffer().getParentTable().setXmlNodeName(name);
   }
   
   /**
    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    * The method is overridden only to allow this object to maintain the structure of the chain of
    * BUFFERs (to update the PERMANENT-TABLE pointer if it was pointing to this buffer AND if the
    * {@link HandleChain#delete()} really disconnected it from the chain)
    */
   @Override
   public void delete()
   {
      // save next sibling because super.delete() will wipe it out 
      HandleChain nextResource = this.nextSibling;
      
      super.delete();
      if (this.nextSibling != null)
      {
         // the super.delete() did not removed this object from chain
         return;
      }
      
      // special care for BUFFER resources: update the PERMANENT-TABLE pointer
      if (this == HandleChain.getHead(PERMANENT_TABLE)) 
      {
         HandleChain.setHead(PERMANENT_TABLE, nextResource);
      }
   }
   
   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or DataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(NumberType fieldNumber)
   {
      if (fieldNumber == null || fieldNumber.isUnknown())
      {
         String msg = "Argument for BUFFER-FIELD method for buffer " +
                      buffer().getDmoInfo().legacyTable + " could not be evaluated";
         ErrorManager.recordOrShowError(7349, msg, false, false, false);
         return new handle();
      }
      
      return bufferField(fieldNumber.intValue());
   }
   
   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or DataSet object.
    *
    * @param   fieldName
    *          A CHARACTER string expression representing the name of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(character fieldName)
   {
      if (fieldName.isUnknown())
      {
         String msg = "Argument for BUFFER-FIELD method for buffer " +
                      buffer().getDmoInfo().legacyTable + " could not be evaluated";
         ErrorManager.recordOrShowError(7349, msg, false, false, false);
         
         // TODO: if this message needs to be displayed, the code must be moved in other place
         // String errMsg2 =
         //    "Lead attributes in a chained-attribute expression (a:b:c) must be type " +
         //    "HANDLE or a user-defined type and valid (not UNKNOWN)";
         // ErrorManager.recordOrShowError(10068, errMsg2, false, false, false);
         return new handle();
      }
      else
      {
         return bufferField(fieldName.getValue());
      }
   }
   
   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or DataSet object.
    *
    * @param   fieldName
    *          A CHARACTER string expression representing the name of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(BaseDataType fieldName)
   {
      if (Text.class.isAssignableFrom(fieldName.getClass()))
      {
         return bufferField(fieldName.toStringMessage());
      }
      
      if (NumberType.class.isAssignableFrom(fieldName.getClass()))
      {
         return bufferField((NumberType) fieldName);
      }
      
      if (BinaryData.class.isAssignableFrom(fieldName.getClass()))
      {
         BinaryData bd = (BinaryData) fieldName;
         character nm = bd.getString(0);
         
         return bufferField(nm);
      }
      
      // otherwise, it uses the 'hex representation' of the value to compute a read a string.
      // we don't emulate this at this time.
      UnimplementedFeature.missing("BUFFER-POLY(incompatible-BDT) not implemented!");
      
      String msg = "BUFFER-FIELD " + fieldName + " was not found in buffer " +
                   buffer().getDmoInfo().legacyTable;
      ErrorManager.recordOrShowError(7351, msg, false, false, false);
      
      // return unknown handle
      return new handle();
   }
   
   /**
    * Implementation of the READ-XML method. Reads XML data from the specified data source into
    * this object.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Default field type mappings</li>
    * <li>LOOSE schema verification mode</li>
    * </ul>
    * 
    * @param   source
    *          Data source.
    * @param   readMode
    *          Mode in which data is read into buffer. Must evaluate to {@code APPEND}, {@code
    *          EMPTY}, {@code MERGE}, or {@code REPLACE}.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to use the
    *          {@code xsi:noNamespaceSchemaLocation} value set in the XML data, if any.
    * @param   overrideDefaultMapping
    *          {@code True} to override the default string-to-binary data type mappings when
    *          creating a temp-table schema from an XML schema.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical readXml(SourceData source,
                          character readMode,
                          character schemaLocation,
                          logical overrideDefaultMapping)
   {
      return readXml(source, readMode, schemaLocation, overrideDefaultMapping, new character());
   }
   
   /**
    * Implementation of the READ-XML method. Reads XML data from the specified data source into
    * this object.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>LOOSE schema verification mode</li>
    * </ul>
    * 
    * @param   source
    *          Data source.
    * @param   readMode
    *          Mode in which data is read into buffer. Must evaluate to {@code APPEND}, {@code
    *          EMPTY}, {@code MERGE}, or {@code REPLACE}.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to use the
    *          {@code xsi:noNamespaceSchemaLocation} value set in the XML data, if any.
    * @param   overrideDefaultMapping
    *          {@code True} to override the default string-to-binary data type mappings when
    *          creating a temp-table schema from an XML schema.
    * @param   fieldTypeMapping
    *          Comma-delimited list of field name and data type name pairs to use to override
    *          data type mappings of specific fields. Specify unknown value to use default
    *          mappings.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical readXml(SourceData source,
                          character readMode,
                          character schemaLocation,
                          logical overrideDefaultMapping,
                          character fieldTypeMapping)
   {
      return readXml(source,
                     readMode,
                     schemaLocation,
                     overrideDefaultMapping,
                     fieldTypeMapping,
                     new character("LOOSE"));
   }
   
   /**
    * Implementation of the READ-XML method. Reads XML data from the specified data source into
    * this object.
    * 
    * @param   source
    *          Data source.
    * @param   readMode
    *          Mode in which data is read into buffer. Must evaluate to {@code APPEND}, {@code
    *          EMPTY}, {@code MERGE}, or {@code REPLACE}.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to use the
    *          {@code xsi:noNamespaceSchemaLocation} value set in the XML data, if any.
    * @param   overrideDefaultMapping
    *          {@code True} to override the default string-to-binary data type mappings when
    *          creating a temp-table schema from an XML schema.
    * @param   fieldTypeMapping
    *          Comma-delimited list of field name and data type name pairs to use to override
    *          data type mappings of specific fields. Specify unknown value to use default
    *          mappings.
    * @param   verifySchemaMode
    *          Schema verification mode. Must evaluate to {@code IGNORE}, {@code LOOSE}, or
    *          {@code STRICT}. Default is {@code LOOSE}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical readXml(SourceData source,
                          character readMode,
                          character schemaLocation,
                          logical overrideDefaultMapping,
                          character fieldTypeMapping,
                          character verifySchemaMode)
   {
      RecordBuffer buffer = buffer();
      
      // sanity
      if (!buffer.isTemporary())
      {
         ErrorManager.recordOrShowError(13158);
         // XML Methods only allowed on TEMP-TABLE buffers.
         return new logical(false);
      }
      
      return TempTableBuilder.asTempTable((Temporary) this).readXml(
            source,
            readMode,
            schemaLocation,
            overrideDefaultMapping,
            fieldTypeMapping,
            verifySchemaMode);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Unformatted text</li>
    * <li>UTF-8 encoding</li>
    * <li>xsi:noNamespaceSchemaLocation=?</li>
    * <li>No XML schema written</li>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target)
   {
      return writeXml(target, null, null, null, null, null, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>UTF-8 encoding</li>
    * <li>xsi:noNamespaceSchemaLocation=?</li>
    * <li>No XML schema written</li>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target, logical formatted)
   {
      return writeXml(target, formatted, null, null, null, null, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>xsi:noNamespaceSchemaLocation=?</li>
    * <li>No XML schema written</li>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target, logical formatted, character encoding)
   {
      return writeXml(target, formatted, encoding, null, null, null, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>No XML schema written</li>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to set the
    *          {@code xsi:noNamespaceSchemaLocation} to empty string or unknown value.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target, logical formatted, character encoding, character schemaLocation)
   {
      return writeXml(target, formatted, encoding, schemaLocation, null, null, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to set the
    *          {@code xsi:noNamespaceSchemaLocation} to empty string or unknown value.
    * @param   writeXmlSchema
    *          {@code True} to write XML schema information, else {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target,
                           logical formatted,
                           character encoding,
                           character schemaLocation,
                           logical writeXmlSchema)
   {
      return writeXml(target, formatted, encoding, schemaLocation, writeXmlSchema, null, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>No before-image data written</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to set the
    *          {@code xsi:noNamespaceSchemaLocation} to empty string or unknown value.
    * @param   writeXmlSchema
    *          {@code True} to write XML schema information, else {@code false}.
    * @param   minXmlSchema
    *          {@code True} to write minimal XML schema information, else {@code false}. Must be
    *          {@code false} if {@code writeXmlSchema} is {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target,
                           logical formatted,
                           character encoding,
                           character schemaLocation,
                           logical writeXmlSchema,
                           logical minXmlSchema)
   {
      return writeXml(target, formatted, encoding, schemaLocation, writeXmlSchema, minXmlSchema, null, null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to set the
    *          {@code xsi:noNamespaceSchemaLocation} to empty string or unknown value.
    * @param   writeXmlSchema
    *          {@code True} to write XML schema information, else {@code false}.
    * @param   minXmlSchema
    *          {@code True} to write minimal XML schema information, else {@code false}. Must be
    *          {@code false} if {@code writeXmlSchema} is {@code false}.
    * @param   writeBeforeImage
    *          {@code True} to write before-image data and error information, else {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target,
                           logical formatted,
                           character encoding,
                           character schemaLocation,
                           logical writeXmlSchema,
                           logical minXmlSchema,
                           logical writeBeforeImage)
   {
      return writeXml(target,
                      formatted,
                      encoding,
                      schemaLocation,
                      writeXmlSchema,
                      minXmlSchema,
                      writeBeforeImage,
                      null);
   }
   
   /**
    * Implementation of the WRITE-XML method. Writes XML data from this object into the specified
    * data target.
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the XML to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   schemaLocation
    *          Name of an external XML schema file, or empty string or unknown value to set the
    *          {@code xsi:noNamespaceSchemaLocation} to empty string or unknown value.
    * @param   writeXmlSchema
    *          {@code True} to write XML schema information, else {@code false}.
    * @param   minXmlSchema
    *          {@code True} to write minimal XML schema information, else {@code false}. Must be
    *          {@code false} if {@code writeXmlSchema} is {@code false}.
    * @param   beforeImage
    *          {@code True} to write before-image data and error information, else {@code false}.
    * @param   noInitialVals
    *          {@code True} to omit data for fields whose values match their initial values.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   public logical writeXml(TargetData target,
                           logical formatted,
                           character encoding,
                           character schemaLocation,
                           logical writeXmlSchema,
                           logical minXmlSchema,
                           logical beforeImage,
                           logical noInitialVals)
   {
      RecordBuffer buffer = buffer();
      
      // sanity
      if (!buffer.isTemporary())
      {
         ErrorManager.recordOrShowError(13158);
         // XML Methods only allowed on TEMP-TABLE buffers.
         return new logical(false);
      }
      
      XmlExport export = new XmlExport(
            target,
            formatted != null && !formatted.isUnknown() && formatted.getValue(),
            encoding == null ? null : encoding.toJavaType());
      
      return new logical(export.writeTempTable(
            (TemporaryBuffer) buffer,
            schemaLocation == null ? null : schemaLocation.toJavaType(),
            writeXmlSchema != null && !writeXmlSchema.isUnknown() && writeXmlSchema.getValue(),
            minXmlSchema != null && !minXmlSchema.isUnknown() && minXmlSchema.getValue(),
            beforeImage != null && !beforeImage.isUnknown() && beforeImage.getValue(),
            noInitialVals != null && !noInitialVals.isUnknown() && noInitialVals.getValue()));
   }
   
   /**
    * Writes an XML representation of the definition of this {@code TempTable}.
    *
    * @param   target
    *          Data target.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical writeXmlSchema(TargetData target)
   {
      return writeXmlSchema(target, null, null, null, null);
   }
   
   /**
    * Writes an XML representation of the definition of this {@code TempTable}.
    *
    * @param   target
    *          Data target.
    * @param   formatted
    *          Use {@code true} for a more human-readable document.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical writeXmlSchema(TargetData target, logical formatted)
   {
      return writeXmlSchema(target, formatted, null, null, null);
   }
   
   /**
    * Writes an XML representation of the definition of this {@code TempTable}.
    *
    * @param   target
    *          Data target.
    * @param   formatted
    *          Use {@code true} for a more human-readable document.
    * @param   encoding
    *          The character encoding. The default is UTF-8.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical writeXmlSchema(TargetData target, logical formatted, Text encoding)
   {
      return writeXmlSchema(target, formatted, encoding, null, null);
   }
   
   /**
    * Writes an XML representation of the definition of this {@code TempTable}.
    *
    * @param   target
    *          Data target.
    * @param   formatted
    *          Use {@code true} for a more human-readable document.
    * @param   encoding
    *          The character encoding. The default is UTF-8.
    * @param   minXmlSchema
    *          If {@code true}, the minimum amount of schema for the object will be written.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical writeXmlSchema(TargetData target, logical formatted, Text encoding, logical minXmlSchema)
   {
      return writeXmlSchema(target, formatted, encoding, minXmlSchema, null); 
   }
   
   /**
    * Writes an XML representation of the definition of this {@code TempTable}.
    *
    * @param   target
    *          Data target.
    * @param   formatted
    *          Use {@code true} for a more human-readable document.
    * @param   encoding
    *          The character encoding. The default is UTF-8.
    * @param   minXmlSchema
    *          If {@code true}, the minimum amount of schema for the object will be written.
    * @param   omitInitVal
    *          Use {@code true} to omit the initial values.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical writeXmlSchema(TargetData target,
                                 logical formatted,
                                 Text encoding,
                                 logical minXmlSchema,
                                 logical omitInitVal)
   {
      RecordBuffer buffer = buffer();
      
      // sanity
      if (!buffer.isTemporary())
      {
         ErrorManager.recordOrShowError(13158);
         // XML Methods only allowed on TEMP-TABLE buffers.
         return new logical(false);
      }
      
      XmlExport export = new XmlExport(
            target,
            formatted != null && !formatted.isUnknown() && formatted.getValue(),
            encoding == null ? null : encoding.toJavaType());
      
      return new logical(export.writeTableSchema(
            (TemporaryBuffer) buffer,
            minXmlSchema != null && !minXmlSchema.isUnknown() && minXmlSchema.getValue(), 
            omitInitVal != null && !omitInitVal.isUnknown() && omitInitVal.getValue()));
   }
   
   /**
    * Definition of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   targetFormat
    *          The format ("xml" or "json").
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the data for making it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitialValues
    *          {@code True} to omit data for fields whose values match their initial values.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical serializeRow(character  targetFormat,
                               TargetData target,
                               logical    formatted,
                               character  encoding,
                               logical    omitInitialValues)
   {
      return new logical(serializeRowImpl(targetFormat.toJavaType(),
                                          target,
                                          formatted.toJavaType(),
                                          encoding.toJavaType(),
                                          omitInitialValues.toJavaType(),
                                          null));
   }
   
   /**
    * Definition of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   format
    *          The format ("xml" or "json").
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the data for making it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical serializeRow(character format, TargetData target, logical formatted, character encoding)
   {
      return new logical(serializeRowImpl(
            format.toJavaType(), target, formatted.toJavaType(), encoding.toJavaType(), null, null));
   }
   
   /**
    * Definition of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   format
    *          The format ("xml" or "json").
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the data for making it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical serializeRow(character format, TargetData target, logical formatted)
   {
      return new logical(serializeRowImpl(
            format.toJavaType(), target, formatted.toJavaType(), null, null, null));
   }
   
   /**
    * Definition of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   format
    *          The format ("xml" or "json").
    * @param   target
    *          Data target.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical serializeRow(character format, TargetData target)
   {
      return new logical(serializeRowImpl(format.toJavaType(), target, null, null, null, null));
   }
   
   /**
    * Definition of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   targetFormat
    *          The format ("xml" or "json").
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the data for making it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitialValues
    *          {@code True} to omit data for fields whose values match their initial values.
    * @param   omitOuterObject
    *          {@code True} to omit data for the outermost object from the output.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical serializeRow(character  targetFormat,
                               TargetData target,
                               logical    formatted,
                               character  encoding,
                               logical    omitInitialValues,
                               logical    omitOuterObject)
   {
      return new logical(serializeRowImpl(targetFormat.toJavaType(),
                                          target,
                                          formatted.toJavaType(),
                                          encoding.toJavaType(),
                                          omitInitialValues.toJavaType(),
                                          omitOuterObject.toJavaType()));
   }
   
   /**
    * Reads XML Schema from an XML document and uses that schema to either create a schema for a
    * {@code TempTable} or {@code TempTable} or verify existing schema in a {@code TempTable},
    * {@code TempTable} or temp-table {@code Buffer} object.
    *
    * @param   source
    *          Data source.
    * @param   override
    *          If {@code true} the default mapping between XML Schema string and binary data types
    *          and ABL data types when creating an ABL temp-table schema from an XML Schema will
    *          be overwritten.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical readXmlSchema(SourceData source, logical override)
   {
      return readXmlSchema(source, override, null, null);
   }
   
   /**
    * Reads XML Schema from an XML document and uses that schema to either create a schema for a
    * {@code TempTable} or {@code TempTable} or verify existing schema in a {@code TempTable},
    * {@code TempTable} or temp-table {@code Buffer} object.
    *
    * @param   source
    *          Data source.
    * @param   override
    *          If {@code true} the default mapping between XML Schema string and binary data types
    *          and ABL data types when creating an ABL temp-table schema from an XML Schema will
    *          be overwritten.
    * @param   mapping
    *          A comma-delimited list of field name, data type pairs. This allows to specify the
    *          data type for a specific field.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical readXmlSchema(SourceData source, logical override, Text mapping)
   {
      return readXmlSchema(source, override, mapping, null);
   }
   
   /**
    * Reads XML Schema from an XML document and uses that schema to either create a schema for a
    * {@code TempTable} or {@code TempTable} or verify existing schema in a {@code TempTable},
    * {@code TempTable} or temp-table {@code Buffer} object.
    *
    * @param   source
    *          Data source.
    * @param   override
    *          If {@code true} the default mapping between XML Schema string and binary data types
    *          and ABL data types when creating an ABL temp-table schema from an XML Schema will
    *          be overwritten.
    * @param   mapping
    *          A comma-delimited list of field name, data type pairs. This allows to specify the
    *          data type for a specific field.
    * @param   verifyMode
    *          Specifies how the XML will be verified against the schema. Valid values: "LOOSE" or
    *          "STRICT".
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical readXmlSchema(SourceData source, logical override, Text mapping, Text verifyMode)
   {
      return TempTableBuilder.asTempTable((Temporary) this).readXmlSchema(source,
                                                                          override,
                                                                          mapping,
                                                                          verifyMode);
   }
   
   /**
    * Implementation of the READ-JSON method. Reads JSON data from the specified data source into
    * this object.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>{@code MERGE} read mode</li>
    * </ul>
    * 
    * @param   source
    *          Data source.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical readJson(SourceData source)
   {
      return readJson(source, new character("MERGE"));
   }
   
   /**
    * Implementation of the READ-JSON method. Reads JSON data from the specified data source into
    * this object.
    * 
    * @param   source
    *          Data source.
    * @param   readMode
    *          Mode in which data is read into buffer. Must evaluate to {@code APPEND}, {@code
    *          EMPTY}, {@code MERGE}, or {@code REPLACE}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical readJson(SourceData source, character readMode)
   {
      RecordBuffer buffer = buffer();
      
      // sanity
      if (!buffer.isTemporary())
      {
         ErrorManager.recordOrShowError(15378);
         // JSON methods only allowed on TEMP-TABLE buffers.
         return new logical(false);
      }
      
      // normalize parameters
      JsonImport jsonImport = new JsonImport(source, readMode.toJavaType());
      boolean success;
      
      try
      {
         // perform import
         success = jsonImport.readTable((TemporaryBuffer) buffer, RecordBuffer::loadRecord);
      }
      catch (PersistenceException exc)
      {
         // TODO: log
         success = false;
      }
      
      return new logical(success);
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Unformatted text</li>
    * <li>UTF-8 encoding</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * <li>Do not omit data for the outermost object, unless {@code target} represents a {@code
    *     JsonArray} object.
    * <li>No before-image data written</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target)
   {
      return writeJson(target,
                       new logical(false),
                       new character(),
                       new logical(false),
                       new logical(false),
                       new logical(false));
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>UTF-8 encoding</li>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * <li>Do not omit data for the outermost object, unless {@code target} represents a {@code
    *     JsonArray} object.
    * <li>No before-image data written</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the JSON to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target, logical formatted)
   {
      return writeJson(target, 
                       formatted,
                       new character(),
                       new logical(false),
                       new logical(false),
                       new logical(false));
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Do not omit temp-table fields which contain their initial values</li>
    * <li>Do not omit data for the outermost object, unless {@code target} represents a {@code
    *     JsonArray} object.
    * <li>No before-image data written</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the JSON to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target, logical formatted, character encoding)
   {
      return writeJson(target,
                       formatted,
                       encoding,
                       new logical(false),
                       new logical(false),
                       new logical(false));
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>Do not omit data for the outermost object, unless {@code target} represents a {@code
    *     JsonArray} object.
    * <li>No before-image data written</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the JSON to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitialValues
    *          {@code True} to omit data for fields whose values match their initial values.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target,
                            logical formatted,
                            character encoding,
                            logical omitInitialValues)
   {
      return writeJson(target,
                       formatted,
                       encoding,
                       omitInitialValues,
                       new logical(false),
                       new logical(false));
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * <p>
    * Defaults applied:
    * <ul>
    * <li>No before-image data written</li>
    * </ul>
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the JSON to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitialValues
    *          {@code True} to omit data for fields whose values match their initial values.
    * @param   omitOuterObject
    *          {@code True} to omit data for the outermost object from the output.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target,
                            logical formatted,
                            character encoding,
                            logical omitInitialValues,
                            logical omitOuterObject)
   {
      return writeJson(target,
                       formatted,
                       encoding,
                       omitInitialValues,
                       omitOuterObject,
                       new logical(false));
   }
   
   /**
    * Implementation of the WRITE-JSON method. Writes JSON data from this object into the
    * specified data target.
    * 
    * @param   target
    *          Data target.
    * @param   formatted
    *          {@code True} to format the JSON to make it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitialValues
    *          {@code True} to omit data for fields whose values match their initial values.
    * @param   omitOuterObject
    *          {@code True} to omit data for the outermost object from the output.
    * @param   writeBeforeImage
    *          {@code True} to write before-image data and error information, else {@code false}.
    * 
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   @Override
   public logical writeJson(TargetData target,
                            logical formatted,
                            character encoding,
                            logical omitInitialValues,
                            logical omitOuterObject,
                            logical writeBeforeImage)
   {
      RecordBuffer buffer = buffer();
      
      // sanity
      if (!buffer.isTemporary())
      {
         ErrorManager.recordOrShowError(15378);
         // JSON methods only allowed on TEMP-TABLE buffers.
         return new logical(false);
      }
      if (writeBeforeImage != null      && 
          !writeBeforeImage.isUnknown() &&
          writeBeforeImage.getValue())
      {
         ErrorManager.recordOrShowError(17366);
         // write-before-image not allowed for TEMP-TABLE or BUFFER object.
         ErrorManager.recordOrShowError(4065, "WRITE-JSON", "BUFFER widget");
         // **The <attribute> attribute on the <widget id> has invalid arguments.
         return new logical(false);
      }
      
      com.fasterxml.jackson.core.JsonEncoding jsonEncoding = JsonExport.parseEncoding(encoding);
      if (jsonEncoding == null)
      {
         ErrorManager.recordOrShowError(15342, encoding.toStringMessage(), "WRITE-JSON");
         // Invalid encoding '<encoding>' argument for <method>.
         return new logical(false);
      }
      
      boolean success;
      try
      {
         JsonExport export = new JsonExport();
         success = export.exportTempBuffer((TemporaryBuffer) buffer,
                                           target,
                                           formatted,
                                           jsonEncoding,
                                           omitInitialValues,
                                           omitOuterObject);
      }
      catch (PersistenceException exc)
      {
         // TODO: log
         success = false;
      }
      
      return new logical(success);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other)
   {
      return copyTempTable(other, false, false, false, null);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  {@code true}  if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other, logical append)
   {
      return copyTempTable(other, append, new logical(false), new logical(false), new character());
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  {@code true} if operation is successful
    */
   @Override
   public logical copyTempTable(handle other, logical append, logical replace)
   {
      return copyTempTable(other, append, replace, new logical(false), new character());
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other, logical append, logical replace, logical looseCopy)
   {
      return copyTempTable(other, append, replace, looseCopy, new character());
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  {@code true}  if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other, boolean append)
   {
      return copyTempTable(other, append, false, false, null);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other, boolean append, boolean replace)
   {
      return copyTempTable(other, append, replace, false, null);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other, boolean append, boolean replace, boolean looseCopy)
   {
      return copyTempTable(other, append, replace, looseCopy, null);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other,
                                logical append,
                                logical replace,
                                logical looseCopy,
                                character prefix)
   {
      TempTable tt = TempTableBuilder.asTempTable((Temporary) this);
      return tt.copyTempTable(other, append, replace, looseCopy, prefix);
   }
   
   /**
    * Copies the other TEMP-TABLE. Other must be a TEMP-TABLE handle. By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical copyTempTable(handle other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String prefix)
   {
      TempTable tt = TempTableBuilder.asTempTable((Temporary) this);
      return tt.copyTempTable(other,
                              new logical(append),
                              new logical(replace),
                              new logical(looseCopy),
                              new character(prefix));
   }
   
   /**
    * Obtain the last associated query (to which this buffer was added or set).
    *
    * @return   The associated query or {@code unknown} if none exists.
    */
   @Override
   public handle getQueryAsHandle()
   {
      // bit of maintenance
      if (lastQuery != null && !lastQuery._isValid())
      {
         lastQuery = null;
      }
      
      return lastQuery == null ? new handle() : new handle(lastQuery.getResource());
   }
   
   /**
    * Set the query associated with the given resource. This is illegal when called directly by
    * P4GL application and will raise 4052 condition.
    *
    * @param   qry
    *          The query to associate.
    */
   @Override
   public void setQueryAsHandle(handle qry)
   {
      handle.invalidAttribute("QUERY", true, LegacyResource.BUFFER);
   }
   
   /**
    * Set the query associated with the given resource. This is illegal when called directly by
    * P4GL application and will raise 4052 condition.
    *
    * @param   qry
    *          The query to associate.
    */
   @Override
   public void setQueryAsHandle(QueryWrapper qry)
   {
      handle.invalidAttribute("QUERY", true, LegacyResource.BUFFER);
   }
   
   /**
    * Gets the maximum number of {@code DataSet} temp-table rows to retrieve in each {@code FILL}
    * operation. The default value is zero (which retrieves all rows that satisfy the associated
    * query).
    *
    * @return  the maximum number of DataSet temp-table rows to retrieve in each {@code FILL}
    *          operation.
    */
   @Override
   public integer batchSize()
   {
      return new integer(batchSize); 
   }
   
   /**
    * Sets the maximum number of {@code DataSet} temp-table rows to retrieve in each {@code FILL}
    * operation.
    *
    * @param   size
    *          The new value for the {@code BATCH-SIZE} attribute.
    */
   @Override
   public void batchSize(long size)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(12361);
         // BATCH-SIZE may only be set for dataset temp-tables.
         return;
      }
      batchSize = (int) size;
   }
   
   /**
    * Sets the maximum number of {@code DataSet} temp-table rows to retrieve in each {@code FILL}
    * operation.
    *
    * @param   size
    *          The new value for the {@code BATCH-SIZE} attribute.
    */
   @Override
   public void batchSize(NumberType size)
   {
      if (size == null || size.isUnknown())
      {
         unableToAssignUnknown("BATCH-SIZE", "BUFFER widget");
         return;
      }
      
      batchSize(size.longValue());
   }
   
   /**
    * Gets the {@code ROWID} of the row in the after-image table that is the current version of
    * the row in the before-image table currently associated with this buffer. This row can
    * be a new or modified row.
    *
    * @return  the {@code ROWID} of the row in the after-image table that is the current version
    *          of the row in the before-image table
    */
   @Override
   public rowid afterRowid()
   {
      RecordBuffer recBuf = buffer();
      if (isAfterBuffer() || !recBuf.isAvailable())
      {
         // AFTER-ROWID is only available for BEFORE-BUFFERS of DATASETs
         return new rowid();
      }
      
      return recBuf.peerRowid();
   }
   
   /**
    * Gets the {@code ROWID} of the row in the before-image table that is the current version of
    * the row in the before-image table currently associated with this buffer. This row can
    * be a new or modified row.
    *
    * @return  the {@code ROWID} of the row in the before-image table that is the current version
    *          of the row in the before-image table
    */
   @Override
   public rowid beforeRowid()
   {
      RecordBuffer recBuf = buffer();
      if (isBeforeBuffer() || !recBuf.isAvailable())
      {
         // BEFORE-ROWID is only available for AFTER-BUFFERS of DATASETs
         return new rowid();
      }
      
      return recBuf.peerRowid();
   }
   
   /**
    * Obtain the current change state of the {@code DataSet} temp-table row associated with this
    * buffer object.
    *
    * @return  the current change state of the {@code DataSet} temp-table row associated with this
    *          buffer object.
    */
   @Override
   public integer rowState()
   {
      RecordBuffer recBuf = buffer();
      if (!recBuf.isAvailable() || !recBuf.isTemporary() || beforeBufferType == BeforeType.SIMPLE)
      {
         return new integer();
      }
      
      // even if there is no dataset for this table, the rowState can be read.
      return recBuf.rowState();
   }
   
   /**
    * Accepts changes to the data in one row of this {@code DataSet} temp-table.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical acceptRowChanges()
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11933, doGetName());
         // ACCEPT-ROW-CHANGES must be run on the BEFORE table of buffer <name>.
         return new logical(false);
      }
      
      if (!_available())
      {
         ErrorManager.recordOrShowError(11934, doGetName());
         // ACCEPT-ROW-CHANGES buffer <name> not available.
         return new logical(false);
      }
      
      int crtState = buffer().rowState().intValue();
      if (crtState == Buffer.ROW_MODIFIED || crtState == Buffer.ROW_CREATED)
      {
         // find and update the after-image/record
         BufferImpl after = (BufferImpl) afterBuffer().unwrapBuffer();
         after.findByRowID(this.afterRowid());
         RecordBuffer afterRecBuf = after.buffer();
         afterRecBuf.peerRowid(new rowid());
         afterRecBuf.rowState(Buffer.ROW_UNMODIFIED);
      }
      // otherwise (DELETE - after image is already deleted)
      
      // then drop the before-image data:
      setUpBeforeBuffer(true);
      deleteRecord();
      setUpBeforeBuffer(false);
      
      return new logical(true); 
   }
   
   /**
    * Rejects changes to the data in one row of this {@code DataSet} temp-table.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical rejectRowChanges()
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11939, doGetName());
         // REJECT-ROW-CHANGES must be run on the BEFORE table of buffer <name>.
         return new logical(false);
      }
      
      if (!_available())
      {
         ErrorManager.recordOrShowError(11936, doGetName());
         // REJECT-ROW-CHANGES buffer <name> not available.
         return new logical(false);
      }
      
      // find and update the after-image/record
      BufferImpl after = (BufferImpl) afterBuffer().unwrapBuffer();
      TemporaryBuffer afterImage = (TemporaryBuffer) after.buffer();
      RecordBuffer recBuf = buffer();
      switch (recBuf.rowState().toJavaIntegerType())
      {
         case Buffer.ROW_CREATED:
            after.findByRowID(this.afterRowid());
            after.deleteRecord();
            break;
         case Buffer.ROW_DELETED:
            after.create();
            BaseRecord.copy(afterImage.getCurrentRecord(), recBuf.getCurrentRecord(), false);
            after.release();
            break;
         case Buffer.ROW_MODIFIED:
            after.findByRowID(this.afterRowid());
            BaseRecord.copy(afterImage.getCurrentRecord(), recBuf.getCurrentRecord(), false);
            afterImage.peerRowid(new rowid());
            afterImage.rowState(Buffer.ROW_UNMODIFIED);
            after.release();
            break;
      }
      
      // then drop the before-image data:
      setUpBeforeBuffer(true);
      deleteRecord();
      setUpBeforeBuffer(false);
   
      return new logical(true);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data
    * source.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges()
   {
      return saveRowChangesImpl(0, null, false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffIndex
    *          The index of the buffer in the data source.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(NumberType buffIndex)
   {
      int idx = buffIndex.isUnknown() ? 1 : buffIndex.intValue();
      return saveRowChangesImpl(idx - 1, null, false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffName
    *          The name of the buffer in the data source.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(Text buffName)
   {
      return saveRowChangesImpl(buffName.toJavaType(), null, false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffIndex
    *          The index of the buffer in the data source.
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(NumberType buffIndex, Text skipList)
   {
      int idx = buffIndex.isUnknown() ? 1 : buffIndex.intValue();
      return saveRowChangesImpl(idx - 1, skipList.toJavaType(), false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffName
    *          The name of the buffer in the data source.
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(Text buffName, Text skipList)
   {
      return saveRowChangesImpl(buffName.toJavaType(), skipList.toJavaType(), false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    * <p>
    * <strong>Note:</strong> this form is used when unknown literal {@code ?} is used to specify
    *         the buffer. In either case, numeric or character, the meaning is the same.
    *
    * @param   buffInfo
    *          Unknown value. 
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(unknown buffInfo, Text skipList)
   {
      return saveRowChangesImpl(0, skipList.toJavaType(), false);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffIndex
    *          The index of the buffer in the data source.
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    * @param   noLobs
    *          Use {@code true} to specify the BLOB and CLOB fields should be ignored in the
    *          save operation.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(NumberType buffIndex, Text skipList, logical noLobs)
   {
      boolean skipLobs = !noLobs.isUnknown() && noLobs.getValue();
      int idx = buffIndex.isUnknown() ? 1 : buffIndex.intValue();
      return saveRowChangesImpl(idx - 1, skipList.toJavaType(), skipLobs);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    * <p>
    * <strong>Note:</strong> this form is used when unknown literal {@code ?} is used to specify
    *         the buffer. In either case, numeric or character, the meaning is the same.
    *
    * @param   buffInfo
    *          Unknown value. 
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(unknown buffInfo, Text skipList, logical noLobs)
   {
      boolean skipLobs = !noLobs.isUnknown() && noLobs.getValue();
      return saveRowChangesImpl(0, skipList.toJavaType(), skipLobs);
   }
   
   /**
    * Saves changes from one row of the {@code DataSet} temp-table to the associated data source.
    *
    * @param   buffName
    *          The name of the buffer in the data source.
    * @param   skipList
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    * @param   noLobs
    *          Use {@code true} to specify the BLOB and CLOB fields should be ignored in the
    *          save operation.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical saveRowChanges(Text buffName, Text skipList, logical noLobs)
   {
      boolean skipLobs = !noLobs.isUnknown() && noLobs.getValue();
      return saveRowChangesImpl(buffName.toJavaType(), skipList.toJavaType(), skipLobs);
   }
   
   /**
    * Obtain the mode in which the {@link DataSet#fill()} method fills a {@code DataSet} member
    * buffer. The default mode is {@code MERGE}.
    *
    * @return  the fill mode for this buffer.
    */
   @Override
   public character fillMode()
   {
      if (dataSet == null)
      {
         // FILL-MODE is unknown for 'free' buffers
         return new character();
      }
      
      return new character(fillMode); 
   }
   
   /**
    * Configures the mode in which the {@link DataSet#fill()} method fills a {@code DataSet}
    * member buffer. The allowed modes are: {@code APPEND}, {@code EMPTY}, {@code MERGE},
    * {@code NO-FILL}, and {@code REPLACE}.
    *
    * @param   mode
    *          The new fill mode for this buffer.
    */
   @Override
   public void fillMode(String mode)
   {
      if (mode == null)
      {
         unableToAssignUnknown("FILL-MODE", "BUFFER widget");
         return;
      }
      
      if (dataSet == null)
      {
         ErrorManager.recordOrShowError(11959);
         // FILL-MODE is only valid for a dataset buffer.
         ErrorManager.recordOrShowError(4058, "FILL-MODE", "BUFFER widget", mode);
         // **Attribute <attribute> for the <widget id> has an invalid value of <string>.
         return;
      }
      
      String upperMode = mode.toUpperCase();
      switch (upperMode)
      {
         case FM_APPEND:
         case FM_EMPTY:
         case FM_MERGE:
         case FM_NOFILL:
         case FM_REPLACE:
            // one of these is expected
            break;
         default:
            // but this is not. Complain about it and use default, MERGE
            ErrorManager.recordOrShowError(11960);
            // FILL-MODE must be one of NO-FILL, MERGE, EMPTY, etc. 
            ErrorManager.recordOrShowError(4058, "FILL-MODE", "BUFFER widget", mode);
            // **Attribute <attribute> for the <widget id> has an invalid value of <string>.
            upperMode = FM_MERGE;
      }
      
      this.fillMode = upperMode;
   }
   
   /**
    * Configures the mode in which the {@link DataSet#fill()} method fills a {@code DataSet}
    * member buffer. The allowed modes are: {@code APPEND}, {@code EMPTY}, {@code MERGE},
    * {@code NO-FILL}, and {@code REPLACE}.
    *
    * @param   mode
    *          the new fill mode for this buffer.
    */
   @Override
   public void fillMode(Text mode)
   {
      fillMode(mode == null ? null : mode.toJavaType());
   }
   
   /**
    * Attaches a {@code DataSource} object to this temp-table {@code Buffer} in a {@code DataSet}
    * object.
    *
    * @param   hds
    *          The handle to the {@code DataSource} to be attached.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical attachDataSource(handle hds)
   {
      return attachDataSource(hds, null, null, null);
   }
   
   /**
    * Attaches a {@code DataSource} object to this temp-table {@code Buffer} in a {@code DataSet}
    * object.
    *
    * @param   hds
    *          The handle to the {@code DataSource} to be attached.
    * @param   pairList
    *          A comma-delimited list of field name pairs that specify a mapping between different
    *          field names in the {@code DataSource} buffer and the {@code DataSet} temp-table
    *          buffer.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical attachDataSource(handle hds, character pairList)
   {
      return attachDataSource(hds, pairList, null, null);
   }
   
   /**
    * Attaches a {@code DataSource} object to this temp-table {@code Buffer} in a {@code DataSet}
    * object.
    *
    * @param   hds
    *          The handle to the {@code DataSource} to be attached.
    * @param   pairList
    *          A comma-delimited list of field name pairs that specify a mapping between different
    *          field names in the {@code DataSource} buffer and the {@code DataSset} temp-table
    *          buffer.
    * @param   exceptFields
    *          A comma-separated list of fields in the {@code DataSet} object buffer that will not
    *          be populated with data from the data source.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical attachDataSource(handle hds, character pairList, character exceptFields)
   {
      return attachDataSource(hds, pairList, exceptFields, null);
   }
   
   /**
    * Attaches a {@code DataSource} object to this temp-table {@code Buffer} in a {@code DataSet}
    * object. This method provides much of the internal definition for the {@code FILL}. Among
    * other things, this method creates the {@code FILL} query.
    * <p>
    * This Java method corresponds to ABL {@code ATTACH-DATA-SOURCE} method with all 4 parameters.
    *
    * @param   hds
    *          The handle to the {@code DataSource} to be attached.
    * @param   pairList
    *          A comma-delimited list of field name pairs that specify a mapping between different
    *          field names in the {@code DataSource} buffer and the {@code DataSet} temp-table buffer.
    * @param   exceptFields
    *          A comma-separated list of fields in the {@code DataSet} object buffer that will not
    *          be populated with data from the data source.
    * @param   includeFields
    *          A comma-separated list of fields to include in the {@code DataSet} object buffer,
    *          as an ALTERNATIVE to specifying fields to exclude in except-fields.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical attachDataSource(handle hds,
                                   character pairList,
                                   character exceptFields,
                                   character includeFields)
   {
      // validate parameters
      if (hds == null || !hds._isValid() || !hds.isType(LegacyResource.DATA_SOURCE))
      {
         ErrorManager.recordOrShowError(11957);
         // ATTACH-DATA-SOURCE must have a valid data-source handle as the first argument.
         return new logical(false);
      }
      
      DataSource newDSrc = (DataSource) hds.unwrapDataSource();
      if (newDSrc._getNumSourceBuffers() == 0)
      {
         ErrorManager.recordOrShowError(11958);
         // ATTACH-DATA-SOURCE may not be called until source buffers have been assigned to the DATA-SOURCE.
         return new logical(false);
      }
      
      if (exceptFields != null && !exceptFields.isUnknown() && !exceptFields.getValue().isEmpty() &&
          includeFields != null && !includeFields.isUnknown() && !includeFields.getValue().isEmpty())
      {
         ErrorManager.recordOrShowError(12779);
         // Cannot have both exclude list and include list.
         return new logical(false);
      }
      
      Set<String> incFields = null;
      Set<String> excFields = null;
      if (exceptFields != null && !exceptFields.isUnknown() && !exceptFields.getValue().isEmpty())
      {
         excFields = new HashSet<>();
         // note: TAB, CR, LF are treated like normal character, only SPACE and , are separators
         StringTokenizer st = new StringTokenizer(exceptFields.toJavaType(), " ,", false);
         while (st.hasMoreTokens())
         {
            String tok = st.nextToken();
            int i = tok.indexOf('.');
            if (i != -1)
            {
               // .f1 is OK, <not-a-table>.f1 is also OK, f1 is not
               excFields.add(tok.substring(i + 1).toLowerCase());
            }
            else
            {
               excFields.add(tok.toLowerCase());
            }
         }
      }
      else if (includeFields != null && !includeFields.isUnknown() && !includeFields.getValue().isEmpty())
      {
         incFields = new HashSet<>();
         // note: TAB, CR, LF are treated like normal character, only SPACE and , are separators
         StringTokenizer st = new StringTokenizer(includeFields.toJavaType(), " ,", false);
         while (st.hasMoreTokens())
         {
            String tok = st.nextToken();
            int i = tok.indexOf('.');
            if (i != -1)
            {
               // .f1 is OK, <not-a-table>.f1 is also OK, f1 is not
               incFields.add(tok.substring(i + 1).toLowerCase());
            }
            else
            {
               incFields.add(tok.toLowerCase());
            }
         }
      }
      
      // silently return true if same data source is assigned
      if (newDSrc == dataSource)
      {
         return new logical(true);
      }
      
      // this must be a DataSet buffer
      if (dataSet == null)
      {
         ErrorManager.recordOrShowError(11952);
         // ATTACH-DATA-SOURCE is only valid for buffers that are members of a dataset.
         return new logical(false);
      }
      
      // analyze field mapping between this buffer and those from new DataSource
      String pairListStr = null;
      
      Map<CaseInsensitiveString, DataSource.FieldReference> matchingNameMap = new LinkedHashMap<>();
      Map<CaseInsensitiveString, DataSource.FieldReference> mappedNamesMap = new LinkedHashMap<>();
      
      // map all name-matching fields from this buffer to the fields in data-source buffers
      // if the ABL programmer wishes another association, this mapping will ve overwritten when
      // processing [pairList], below.
      List<TableMapper.LegacyFieldInfo> myFields = TableMapper.getLegacyOrderedList(this, true);
      for (TableMapper.LegacyFieldInfo myField : myFields)
      {
         String myFieldName = myField.getLegacyName();
         BufferImpl ref = null;
         for (int k = 1; k <= newDSrc._getNumSourceBuffers(); k++)
         {
            BufferImpl buffer = (BufferImpl) newDSrc.getSourceBuffer(k).unwrapBuffer();
            if (buffer.hasField(myFieldName))
            {
               // if multiple buffers have the same field, then the last buffer having this field will be used
               ref = buffer;
            }
         }
         if (ref != null)
         {
            boolean dropIt = incFields != null && !incFields.contains(myFieldName.toLowerCase()) ||
                             excFields != null && excFields.contains(myFieldName.toLowerCase());
            
            if (!dropIt)
            {
               matchingNameMap.put(new CaseInsensitiveString(myFieldName), 
                                   new DataSource.FieldReference(ref, myFieldName));
            }
         }
      }
      
      // now process the mapping the ABL programmer instructed:
      if (pairList != null && !pairList.isUnknown())
      {
         // validate the list of field pairs
         pairListStr = pairList.getValue();
         String[] splits = pairListStr.isEmpty() ? new String[0] : pairListStr.split(",");
         if (splits.length % 2 == 1) 
         {
            ErrorManager.recordOrShowError(9035);
            // Pairs list had too many pairs or odd number of entries.
            return new logical(false);
         }
         
         // identify the fields for each table. Note that the pair may be swapped!
         for (int i = 0; i < splits.length; i += 2)
         {
            String fld1 = splits[i];
            String fld2 = splits[i + 1];
            
            if (fld1.toLowerCase().startsWith("rowid("))
            {
               fld1 = fld1.substring(fld1.indexOf('(') + 1, fld1.indexOf(')')) + "." + Session.PK;
               splits[i] = fld1;
            }
            if (fld2.toLowerCase().startsWith("rowid("))
            {
               fld2 = fld2.substring(fld2.indexOf('(') + 1, fld2.indexOf(')')) + "." + Session.PK;
               splits[i + 1] = fld2;
            }
            String fld1NoEx = fld1.indexOf('[') == -1 ? fld1 : fld1.substring(0, fld1.indexOf('['));
            String fld2NoEx = fld2.indexOf('[') == -1 ? fld2 : fld2.substring(0, fld2.indexOf('['));
            
            if (!hasFieldEx(fld1NoEx))
            {
               if (!hasFieldEx(fld2NoEx))
               {
                  ErrorManager.recordOrShowError(9034, fld1);
                  // Pair with <member> failed to have a member in each table.
                  return new logical(false);
               }
               
               // swap [fld1] and [fld2] so that [fld1] is in current buffer
               String tmp = fld1;
               fld1 = fld2;
               fld2 = tmp;
               
               tmp = fld1NoEx;
               fld1NoEx = fld2NoEx;
               fld2NoEx = tmp;
            }
            
            boolean found = false;
            BufferImpl buffer = null;
            for (int j = 1; j <= newDSrc._getNumSourceBuffers(); j++)
            {
               buffer = (BufferImpl) newDSrc.getSourceBuffer(j).unwrapBuffer();
               if (buffer.hasFieldEx(fld2NoEx))
               {
                  found = true;
                  break;
               }
            }
            if (!found)
            {
               // the [fld2] could not be found in any of the buffers of DataSource
               ErrorManager.recordOrShowError(9034, splits[i]);
               // Pair with <member> failed to have a member in each table.
               return new logical(false);
            }
            
            // normalize [fld1] 
            int k = fld1.indexOf('.');
            if (k > 0)
            {
               fld1 = fld1.substring(k + 1);
            }
            mappedNamesMap.put(new CaseInsensitiveString(fld1),
                               new DataSource.FieldReference(buffer, fld2));
         }
      }
      
      // cache the DATA-SOURCE-COMPLETE-MAP value
      dataSrcMapping = new LinkedHashMap<>();
      StringBuilder sb = new StringBuilder();
      String myName = name.toStringMessage();
      
      // first the 'synonym' fields
      for (Map.Entry<CaseInsensitiveString, DataSource.FieldReference> pair : matchingNameMap.entrySet())
      {
         CaseInsensitiveString myFieldCiName = pair.getKey();
         DataSource.FieldReference dataSrcField = pair.getValue();
         if (mappedNamesMap.containsKey(myFieldCiName))
         {
            continue;
         }
         
         dataSrcMapping.put(myFieldCiName, dataSrcField);
         if (sb.length() != 0)
         {
            sb.append(',');
         }
         sb.append(myName).append('.').append(myFieldCiName.getValue()).append(',')
           .append(dataSrcField.getBuffer().name().toStringMessage()).append('.')
           .append(dataSrcField.getName());
      }
      
      // then the user-defined mapping, in the order they were defined in mapping list
      Set<Map.Entry<CaseInsensitiveString, DataSource.FieldReference>> entries = mappedNamesMap.entrySet();
      for (Map.Entry<CaseInsensitiveString, DataSource.FieldReference> entry : entries)
      {
         CaseInsensitiveString ciName = entry.getKey();
         DataSource.FieldReference dataSrcField = entry.getValue();
         dataSrcMapping.put(ciName, dataSrcField);
         if (sb.length() != 0)
         {
            sb.append(',');
         }
         
         sb.append(myName).append('.').append(ciName.getValue()).append(',');
         if (Session.PK.equals(dataSrcField.getName()))
         {
            sb.append("rowid(").append(dataSrcField.getBuffer().name().toStringMessage()).append(")");
         }
         else
         {
            sb.append(dataSrcField.getBuffer().name().toStringMessage()).append('.')
              .append(dataSrcField.getName());
            
            if (dataSrcField.getIndex() > 0)
            {
               sb.append("[").append(dataSrcField.getIndex()).append("]");
            }
         }
      }
      dataSrcMappingStr = sb.toString().toLowerCase();
      
      // all fine, create the links between objects:
      this.dataSource = newDSrc;
      this.attachedPairlist = pairListStr;
      dataSource.attach(this, dataSrcMapping);
      
      return new logical(true);
   }
   
   /**
    * Detaches the {@code DataSource} from the temp-table buffer in a {@code DataSet} object.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical detachDataSource()
   {
      // TODO: check if something is vetoing the detach process
      if (dataSource != null)
      {
         dataSource.attach(null, null);
         dataSource = null;
      }
      
      attachedPairlist = null;
      dataSrcMappingStr = null;
      dataSrcMapping = null;
      return new logical(true);
   }
   
   /**
    * Get the handle to the {@code DataSource} object currently attached to the {@code DataSet}
    * object buffer.
    *
    * @return  The value of {@code DATA-SOURCE} as described above.
    */
   @Override
   public handle dataSource()
   {
      return new handle(dataSource);
   }
   
   /**
    * Obtain the handle to the default buffer of the after-image table that corresponds to
    * the buffer of the before-image table currently associated with this buffer handle.
    *
    * @return  The value of {@code AFTER-BUFFER} attribute as described above.
    */
   @Override
   public handle afterBuffer()
   {
      if (!isBeforeBuffer())
      {
         // permanent and simple temp-tables do not have BEFORE/AFTER buffers
         return new handle();
      }
      
      return new handle(peerBuffer);
   }
   
   /**
    * Obtain the {@code rowid} of the row in the original before-image table that corresponds to
    * the row in the change table currently associated with this buffer handle.
    * <p>
    * Implementation note: 
    *    FWD uses as a storage backing for both {@code ORIGIN-ROWID} and {@code DATA-SOURCE-ROWID}
    *    the same field ({@link TempRecord#_originRowid}) exclusively.
    *    <ul>
    *       <li>when the record is a normal dataset temp-table record the meaning is 
    *            {@code DATA-SOURCE-ROWID}, for both BEFORE and AFTER image. In the case the
    *            {@code DataSource} must be also attached or an ABL error is raised. </li> 
    *       <li>when we use it as a CHANGE dataset temp-table, the meaning of the field is
    *             {@code ORIGIN-ROWID}.</li>
    *   </ul>
    *
    * @return  The value of {@code ORIGIN-ROWID} attribute as described above.
    * 
    * @see     #dataSourceRowid() 
    */
   @Override
   public rowid originRowid()
   {
      RecordBuffer recBuf = buffer();
      
      if (!recBuf.isTemporary() || !recBuf.isAvailable())
      {
         return new rowid();
      }
      TempTable tableHandle = (TempTable) tableHandle().getResource();
      if (!tableHandle.getOriginHandle()._isValid())
      {
         return new rowid();
      }
      
      if (!recBuf.isTemporary() || !recBuf.isAvailable())
      {
         return new rowid();
      }
      
      TempRecord record = (TempRecord) recBuf.getCurrentRecord();
      return new rowid(record._originRowid());
   }
   
   /**
    * Obtain the handle to the default buffer of the before-image table that corresponds to
    * the buffer of the after-image table currently associated with this buffer handle.
    *
    * @return  The value of {@code BEFORE-BUFFER} attribute as described above.
    */
   @Override
   public handle beforeBuffer()
   {
      if (!isAfterBuffer())
      {
         // permanent and simple temp-tables do not have BEFORE/AFTER buffers
         return new handle();
      }
      
      return new handle(peerBuffer);
   }
   
   /**
    * The list of field name pairs for all fields  (in comm-delimited format) in a {@code DataSet}
    * temp-table buffer that are mapped to corresponding fields in an attached data-source object.
    *
    * @return  The value of {@code DATA-SOURCE-COMPLETE-MAP} attribute.
    */
   @Override
   public character dataSourceCompleteMap()
   {
      if (dataSet == null)
      {
         ErrorManager.recordOrShowError(12780);
         // DATA-SOURCE-COMPLETE-MAP only legal for dataset tables.
         return new character();
      }
      
      if (dataSource == null)
      {
         ErrorManager.recordOrShowError(12781);
         // DATA-SOURCE-COMPLETE-MAP requires prior ATTACH-DATA-SOURCE method to be run.
         return new character();
      }
      
      return new character(dataSrcMappingStr);
   }
   
   /**
    * Obtains the {@code DataSource} row corresponding to this buffer object.
    * <p>
    * Implementation note: 
    *    FWD uses as a storage backing for both {@code ORIGIN-ROWID} and {@code DATA-SOURCE-ROWID}
    *    the same field ({@link TempRecord#_originRowid}) exclusively.
    *    <ul>
    *       <li>when the record is a normal dataset temp-table record the meaning is 
    *            {@code DATA-SOURCE-ROWID}, for both BEFORE and AFTER image. In the case the
    *            {@code DataSource} must be also attached or an ABL error is raised. </li> 
    *       <li>when we use it as a CHANGE dataset temp-table, the meaning of the field is
    *            {@code ORIGIN-ROWID}.</li>
    *   </ul>
    *
    * @return  The current value of {@code DATA-SOURCE-ROWID} ABL attribute.
    * 
    * @see  #originRowid()
    */
   @Override
   public rowid dataSourceRowid()
   {
      if (dataSet == null || !dataSet.valid())
      {
         if (!isBeforeBuffer() || peerBuffer.dataSet == null || !peerBuffer.dataSet.valid())
         {
            ErrorManager.recordOrShowError(14275, doGetName());
            // Dataset record for <name> required for DATA-SOURCE-ROWID
            return new rowid();
         }
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return new rowid();
      }
      
      if (dataSource == null || !dataSource.valid())
      {
         if (!isBeforeBuffer() || peerBuffer.dataSource == null || !peerBuffer.dataSource.valid())
         {
            ErrorManager.recordOrShowError(14276, doGetName());
            // Dataset record for <name> required for DATA-SOURCE-ROWID
            return new rowid();
         }
      }
      
      TempRecord record = (TempRecord) buffer.getCurrentRecord();
      
      return new rowid(record._datasourceRowid());
   }
   
   /**
    * Obtains the {@code DataSource} row corresponding to this buffer object.
    *
    * @param   joinLevel
    *          The level of join, if the data-source is a join.
    *
    * @return  The current value of {@code DATA-SOURCE-ROWID} ABL attribute.
    */
   @Override
   public rowid dataSourceRowid(int64 joinLevel)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID
         return new rowid();
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return new rowid();
      }
      
      UnimplementedFeature.missing("BUFFER:DATA-SOURCE-ROWID(n) not implemented");
      return new rowid();
   }
   
   /**
    * Obtains the {@code DataSource} row corresponding to this buffer object.
    *
    * @param   buffName
    *          The name of one of the data-source buffers.
    *
    * @return  The current value of {@code DATA-SOURCE-ROWID} ABL attribute.
    */
   @Override
   public rowid dataSourceRowid(character buffName)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID
         return new rowid();
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return new rowid();
      }
      
      UnimplementedFeature.missing("BUFFER:DATA-SOURCE-ROWID(name) not implemented");
      return new rowid();
   }
   
   /**
    * Configures the {@code DataSource} row corresponding to this buffer object.
    *
    * @param   val
    *          The new value of {@code DATA-SOURCE-ROWID} ABL attribute.
    */
   @Override
   public void dataSourceRowid(rowid val)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID
         return;
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return;
      }
      
      if (isBeforeBuffer())
      {
         peerBuffer.dataSourceRowid(val);
         return;
      }
      
      setDatasourceRowidImpl(val.getValue());
      
      // TODO: set the _datasourceRowid on the peer record (AFTER / BEFORE)
   }
   
   /**
    * Configures the {@code DataSource} row corresponding to this buffer object.
    *
    * @param   joinLevel
    *          The level of join, if the data-source is a join.
    * @param   val
    *          The new value of {@code DATA-SOURCE-ROWID} ABL attribute.
    */
   @Override
   public void dataSourceRowid(int64 joinLevel, rowid val)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID
         return;
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return;
      }
      
      if (isBeforeBuffer())
      {
         peerBuffer.dataSourceRowid(joinLevel, val);
         return;
      }
      
      UnimplementedFeature.missing("BUFFER:DATA-SOURCE-ROWID(n)= not implemented");
   }
   
   /**
    * Configures the {@code DataSource} row corresponding to this buffer object.
    *
    * @param   buffName
    *          The name of one of the data-source buffers.
    * @param   val
    *          The new value of {@code DATA-SOURCE-ROWID} ABL attribute.
    */
   @Override
   public void dataSourceRowid(character buffName, rowid val)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID
         return;
      }
      
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() || !buffer.isAvailable())
      {
         return;
      }
      
      if (isBeforeBuffer())
      {
         peerBuffer.dataSourceRowid(buffName, val);
         return;
      }
      
      UnimplementedFeature.missing("BUFFER:DATA-SOURCE-ROWID(name)= not implemented");
   }
   
   /**
    * Get the {@code DataSet} object of which the buffer is a member or {@code null} otherwise.
    *
    * @return  the {@code DataSet} object of which the buffer is a member.
    */
   public DataSet _dataSet()
   {
      return isBeforeBuffer() ? peerBuffer.ref().dataSet : dataSet;
   }
   
   /**
    * Get the handle for the {@code DataSet} object of which the buffer is a member.
    *
    * @return  handle for the {@code DataSet} object of which the buffer is a member.
    */
   @Override
   public handle dataSet()
   {
      return new handle(_dataSet());
   }
   
   /**
    * Counts the ACTIVE relations for which the buffer is the parent. A buffer may be a parent in
    * multiple relations, but a child in only one.
    *
    * @return  The value of {@code NUM-CHILD-RELATIONS} attribute as described above.
    */
   @Override
   public integer numChildRelations()
   {
      if (dataSet == null)
      {
         // quick out, this is no dataset buffer
         return new integer(0);
      }
      
      List<DataRelation> rels = dataSet.getRelations(this, true, false, true);
      return new integer(rels.size());
   }
   
   /**
    * Obtain the handle to the ACTIVE parent data-relation object for the parent of this buffer.
    * Since a buffer may be a child in only one ACTIVE relation, there can be maximum only one
    * parent for any buffer at any time. 
    *
    * @return  The value of {@code PARENT-RELATION} attribute as described above.
    */
   @Override
   public handle parentRelation()
   {
      if (dataSet == null)
      {
         // quick out, this is no dataset buffer
         return new handle();
      }
      
      List<DataRelation> rels = dataSet.getRelations(this, false, true, true);
      if (rels.isEmpty())
      {
         return new handle();
      }
      
      for (DataRelation rel : rels)
      {
         if (rel._isActive())
         {
            return new handle(rel);
         }
      }
      
      return new handle();
   }
   
   /**
    * Creates a blank before-table record for each buffer record, with a {@code ROW-STATE} of
    * {@code ROW-CREATED} for a table with only new data.
    *
    * @return  {@code true} on success.
    */
   @Override
   public logical markNew()
   {
      if (!isAfterBuffer())
      {
         ErrorManager.recordOrShowError(14274);
         // MARK-ROW-STATE and MARK-NEW may only be used on a dataset member buffer with an existing BEFORE-TABLE.
         return new logical(false);
      }
      
      return new logical(((TemporaryBuffer) buffer()).markNew(
            (BufferImpl) beforeBuffer().unwrapBuffer()));
   }
   
   /**
    * Manually set the {@code ROW-STATE} attribute for the current record/row.
    *
    * @param   newState
    *          The new state the record is marked with. Legal values: {@code ROW_CREATED},
    *          {@code ROW_MODIFIED}, {@code ROW_DELETED}, and {@code ROW_UNMODIFIED}.
    *
    * @return {@code true} on success.
    */
   @Override
   public logical markRowState(int64 newState)
   {
      int jState = (newState == null || newState.isUnknown()) ?
            Buffer.ROW_UNMODIFIED :
            newState.intValue();
      return markRowState(jState, null);
   }
   
   /**
    * Manually set the {@code ROW-STATE} attribute for the current record/row.
    *
    * @param   newState
    *          The new state the record is marked with. Legal values: {@code ROW_CREATED},
    *          {@code ROW_MODIFIED}, {@code ROW_DELETED}, and {@code ROW_UNMODIFIED}.
    *
    * @return {@code true} on success.
    */
   @Override
   public logical markRowState(long newState)
   {
      return markRowState(newState, null);
   }
   
   /**
    * Manually set the {@code ROW-STATE} attribute for the current record/row.
    *
    * @param   newState
    *          The new state the record is marked with. Legal values: {@code ROW_CREATED},
    *          {@code ROW_MODIFIED}, {@code ROW_DELETED}, and {@code ROW_UNMODIFIED}.
    * @param   before
    *          A handle to the before-image source buffer.
    *
    * @return {@code true} on success.
    */
   @Override
   public logical markRowState(int64 newState, handle before)
   {
      int jState = (newState == null || newState.isUnknown()) ?
            Buffer.ROW_UNMODIFIED :
            newState.intValue();
      return markRowState(jState, before);
   }
   
   /**
    * Manually set the {@code ROW-STATE} attribute for the current record/row.
    *
    * @param   newState
    *          The new state the record is marked with. Legal values: {@code ROW_CREATED},
    *          {@code ROW_MODIFIED}, {@code ROW_DELETED}, and {@code ROW_UNMODIFIED}.
    * @param   compareTo
    *          A handle to the before-image source buffer - used to compare to current data.
    *
    * @return {@code true} on success.
    */
   @Override
   public logical markRowState(long newState, handle compareTo)
   {
      if (!isAfterBuffer())
      {
         ErrorManager.recordOrShowError(14274);
         // MARK-ROW-STATE and MARK-NEW may only be used on a dataset member buffer with an existing BEFORE-TABLE.
         return new logical(false);
      }
      
      if (newState == Buffer.ROW_UNMODIFIED)
      {
         return new logical(true); // nothing to process
      }
      else if (newState != Buffer.ROW_CREATED  &&
               newState != Buffer.ROW_MODIFIED &&
               newState != Buffer.ROW_DELETED)
      {
         ErrorManager.recordOrShowError(14288);
         // Invalid ROW-STATE given to MARK-ROW-STATE.
         return new logical(false);
      }
      
      if (!_available())
      {
         ErrorManager.recordOrShowError(14275, doGetName());
         // Dataset record for <name> required for DATA-SOURCE-ROWID.
         return new logical(false);
      }
      
      RecordBuffer compRB = null;
      BufferImpl compBuffer = null;
      if (compareTo != null && !compareTo.isUnknown())
      {
         if (!compareTo.isType(LegacyResource.BUFFER))
         {
            ErrorManager.recordOrShowError(14289);
            // Invalid handle or handle type for MARK-ROW-STATE
            return new logical(false);
         }
         
         compBuffer = (BufferImpl) compareTo.unwrapBuffer();
         if (!compBuffer._available())
         {
            ErrorManager.recordOrShowError(0, "MARK-ROW-STATE called with an empty BEFORE-BUFFER");
            // Progress OpenEdge crashes.
            return new logical(false);
         }
         if (compBuffer.buffer().isTemporary())
         {
            compRB = compBuffer.buffer();
         }
      }
      
      integer rowState = rowState();
      if (!rowState.isUnknown() && rowState.intValue() != Buffer.ROW_UNMODIFIED)
      {
         return new logical(true); // already has before image/record
      }
      
      BufferImpl b4Buffer = (BufferImpl) beforeBuffer().unwrapBuffer();
      RecordBuffer b4RecBuf = b4Buffer.buffer();
      RecordBuffer recBuf = buffer();
      b4Buffer.setUpBeforeBuffer(true);
      switch ((int) newState)
      {
         case ROW_CREATED:
            b4Buffer.create();
            recBuf.rowState((int) newState);
            recBuf.peerRowid(b4RecBuf.rowID());
            b4RecBuf.peerRowid(recBuf.rowID());
            b4RecBuf.rowState((int) newState);
            b4Buffer.release();
            break;
         case ROW_MODIFIED:
            b4Buffer.create();
            if (compRB != null)
            {
               RecordBuffer.copyCommonFields(b4RecBuf.getCurrentRecord(), compRB.getCurrentRecord());
            }
            else
            {
               RecordBuffer.copyCommonFields(b4RecBuf.getCurrentRecord(), recBuf.getCurrentRecord());
            }
            recBuf.rowState((int) newState);
            recBuf.peerRowid(b4RecBuf.rowID());
            b4RecBuf.peerRowid(recBuf.rowID());
            b4RecBuf.rowState((int) newState);
            b4Buffer.release();
            break;
         case ROW_DELETED:
            // this buffer is not updated
            b4Buffer.create();
            if (compRB != null)
            {
               RecordBuffer.copyCommonFields(b4RecBuf.getCurrentRecord(), compRB.getCurrentRecord());
            }
            else 
            {
               RecordBuffer.copyCommonFields(b4RecBuf.getCurrentRecord(), recBuf.getCurrentRecord());
            }
            b4RecBuf.peerRowid(new rowid());
            b4RecBuf.rowState((int) newState);
            b4Buffer.release();
            break;
      }
      b4Buffer.setUpBeforeBuffer(false);
      
      return new logical(true);
   }
   
   /**
    * Obtain the handle to a {@code DataRelation} object for which the buffer is the parent.
    *
    * @param   index
    *          The index of the data-relation object (1-based).
    *
    * @return  the handle to a {@code DataRelation} object for which the buffer is the parent.
    */
   @Override
   public handle childRelation(long index)
   {
      if (dataSet == null || index < 1)
      {
         // quick out, this is no dataset buffer
         return new handle();
      }
      
      List<DataRelation> rels = dataSet.getRelations(this, true, false, true);
      if (index > rels.size())
      {
         return new handle();
      }
      
      return new handle(rels.get((int) (index - 1)));
   }
   
   /**
    * Obtain the handle to a {@code DataRelation} object for which the buffer is the parent.
    *
    * @param   index
    *          The index of the data-relation object (1-based).
    *
    * @return  the handle to a {@code DataRelation} object for which the buffer is the parent.
    */
   @Override
   public handle childRelation(int64 index)
   {
      if (index == null || index.isUnknown())
      {
         return new handle();
      }
      
      return childRelation(index.longValue());
   }
   
   /**
    * Obtain a comma-separated list of field name pairs for fields in a {@code DataSet}
    * temp-table buffer that are mapped to corresponding fields in an attached {@code DataSource}
    * object. This list includes only the field name pairs you specified with the most recently
    * attached data-source object, in the order they were specified.
    *
    *
    * @return  the list of attach fields, as described above.
    */
   @Override
   public character attachedPairlist()
   {
      return new character(attachedPairlist);
   }
   
   /**
    * Returns the state of the REJECTED attribute
    *
    * @return Current state of the REJECTED attribute.
    */
   @Override
   public logical rejected()
   {
      RecordBuffer recBuf = buffer();
      if (!recBuf.isAvailable() || !recBuf.isTemporary())
      {
         return new logical(false);
      }
      
      DataSet ds = _dataSet();
      if (ds == null)
      {
         return new logical(false);
      }
      
      // NOTE: internally, the P4GL database the __error-flag__ is an integer storing multiple flags.
      //       We need to extract the right bit (ERROR_REJECTED) and convert it to a logical value.
      Integer flags = recBuf.errorFlags();
      return new logical(flags != null && (flags & Buffer.ERROR_REJECTED) != 0);
   }
   
   /**
    * Setter for REJECTED attribute.
    *
    * @param value The new value of REJECTED attribute.
    */
   @Override
   public void rejected(boolean value)
   {
      RecordBuffer recBuf = buffer();
      
      if (!_available())
      {
         // temp-tables case
         ErrorManager.recordOrShowError(11916, doGetName());
         // <name> record is not available to set ERROR in.
         ErrorManager.recordOrShowError(3131, "REJECTED", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
      }
      
      if (!recBuf.isTemporary())
      {
         // permanent table case
         ErrorManager.recordOrShowError(11917, doGetName());
         // Cannot set ERROR if there is no BEFORE-TABLE for <name>.
         ErrorManager.recordOrShowError(3131, "REJECTED", "", LegacyResource.BUFFER);
         // Unable to set attribute <attribute name> in widget <widget name> of type <widget-type>.
         return;
      }
      
      if (!isBeforeBuffer())
      {
         if (!isAfterBuffer() || // SIMPLE BUFFER
             (recBuf.peerRowid() == null || recBuf.peerRowid().isUnknown())) // no AFTER-IMAGE
         {
            ErrorManager.recordOrShowError(11918, doGetName(), "");
            // Cannot set ERROR if there is no BEFORE-TABLE row for <name>.
            return;
         }
      }
      
      // NOTE: internally, the P4GL database the __error-flag__ is an integer storing multiple flags.
      //       We need to update the right bit (ERROR_ERROR) to [value].
      recBuf.updateErrorFlags(Buffer.ERROR_REJECTED, value);
   }
   
   /**
    * Setter for REJECTED attribute.
    *
    * @param value The new value of REJECTED attribute.
    */
   @Override
   public void rejected(logical value)
   {
      if (value == null || value.isUnknown())
      {
         unableToAssignUnknown("REJECTED", "BUFFER widget");
         return;
      }
      
      rejected(value.getValue());
   }
   
   /**
    * Fills a {@code Buffer} object, recursively, based on its defined data sources.
    *
    * @return {@code true} if operation is successful.
    */
   @Override
   public logical fill()
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(11945);
         // BUFFER:FILL is only valid for buffers that are members of a dataset.
         return new logical(false);
      }
      
      lastBatch = false;
      
      if (FM_NOFILL.equalsIgnoreCase(fillMode))
      {
         // nothing to do
         return new logical(true);
      }
      
      handle self = new handle(this);
      
      CallbackData bed = callbacks.get(BEFORE_FILL_EVENT);
      if (bed != null)
      {
         boolean ok = dataSet.invokeCallback(self, bed);
         if (!ok)
         {
            dataSet.error(true);
            return new logical(false);
         }
         // TODO: check: if (!ok) return false?
      }
      else if (dataSource == null || !dataSource.valid())
      {
         ErrorManager.recordOrShowError(11874, doGetName());
         // FILL requires an attached data-source for buffer %1, or an active before-fill callback procedure.
         dataSet.error(true);
         return new logical(false);
      }
      
      if (TempTableBuilder.asTempTable((Temporary) this).isTrackingChanges().getValue())
      {
         ErrorManager.recordOrShowError(11863, doGetName());
         // FILL is not valid for table <name> when TRACKING-CHANGES is on.
         dataSet.error(true);
         return new logical(false);
      }
      
      logical fillOk = new logical(true);
      
      if (dataSource != null && dataSource.valid())
      {
         // check DATA-SOURCE sanity and decide whether we need to create a new query
         int srcBufCount = dataSource._getNumSourceBuffers();
         boolean suppliedQuery = false;
         if (srcBufCount > 1)
         {
            if (!dataSource.getQueryAsHandle()._isValid())
            {
               ErrorManager.recordOrShowError(11980);
               // Must assign a user query to a data-source that is a join before filling.
               ErrorManager.recordOrShowError(11875, doGetName());
               // FILL of buffer <name> requires a DATA-SOURCE QUERY to be prepared and assigned for this type of buffer.
               dataSet.error(true);
               return new logical(false);
            }
            
            suppliedQuery = true;
         }
         if (dataSource.getQueryAsHandle()._isValid())
         {
            suppliedQuery = true;
         }
         
         // we can start FILLing the table now 
         if (FM_EMPTY.equalsIgnoreCase(fillMode))
         {
            deleteAll();
         }
         
         // cache buffers for quick access. Make sure they are initialized.
         BufferImpl[] srcBuffs = new BufferImpl[srcBufCount];
         for (int k = 0; k < srcBufCount; k++)
         {
            srcBuffs[k] = (BufferImpl) dataSource.getSourceBuffer(k + 1).unwrapBuffer();
            srcBuffs[k].buffer().initialize();
         }
         
         RecordBuffer recBuf = buffer();
         recBuf.initialize();
         
         // flush all of them (including myself)
         boolean inTx = false;
         try
         {
            inTx = recBuf.persistenceContext.beginTransaction(null);
            recBuf.flush();
            
            for (BufferImpl srcBuff : srcBuffs)
            {
               srcBuff.buffer().flush();
            }
         }
         catch (ValidationException | PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         finally
         {
            if (inTx)
            {
               try
               {
                  recBuf.persistenceContext.commit();
               }
               catch (PersistenceException exc)
               {
                  ErrorManager.recordOrThrowError(exc);
               }
            }
         }
         
         // create a query that eventually uses the fill-where-string and get the data
         handle hQuery = null;
         P2JQuery fillQuery = null;
         if (suppliedQuery)
         {
            hQuery = dataSource.getQueryAsHandle();
            fillQuery = (P2JQuery) hQuery.getResource();
            if (fillQuery.prepareString().isUnknown())
            {
               ErrorManager.recordOrShowError(11876, doGetName());
               // FILL of buffer <name> with user-supplied query requires that the query be QUERY-PREPARED. 
               dataSet.error(true);
               return new logical(false);
            }
         }
         else // srcBufCount == 1
         {
            hQuery = new handle();
            QueryWrapper.createQuery(hQuery);
            fillQuery = hQuery.unwrapQuery();
            StringBuilder sb = new StringBuilder("FOR EACH ");
            sb.append(srcBuffs[0].name().toStringMessage());
            fillQuery.addBuffer(srcBuffs[0]);
            
            // add the buffer(s) in parent relation to this 
            List<DataRelation> parentRelations = dataSet.getRelations(this, false, true, true);
            if (!parentRelations.isEmpty())
            {
               List<Buffer> substs = new ArrayList<>(parentRelations.size());
               for (DataRelation rel : parentRelations)
               {
                  substs.add(rel.getParentBuffer().unwrapBuffer());
               }
               ((QueryWrapper) fillQuery).setSubstitutionBuffers(substs);
            }
            
            String whereFillStr = dataSource.getFillWhereString().toJavaType();
            if (whereFillStr != null)
            {
               sb.append(" ");
               sb.append(whereFillStr);
            }
            sb.append(" NO-LOCK");
            
            if (!fillQuery.prepare(sb.toString()).getValue())
            {
               // TODO: is it possible to have an error here?
               dataSet.error(true);
               return new logical(false);
            }
         }
         
         final P2JQuery finalFillQuery = fillQuery;
         ErrorManager.ErrorHelper errHlp = recBuf.bufferManager.getTxHelper().errHlp;
         boolean wmOn = errHlp.isWarningMode();
         if (!wmOn)
         {
            errHlp.setWarningMode(true);
         }
         
         try
         {
            finalFillQuery.queryOpen();
            finalFillQuery.first();
            Long[] rowids = dataSource.getRestartRowids();
            if (rowids[0] != null)
            {
               finalFillQuery.repositionByID(rowids);
               
               // if [repositionByID] returns [false] the error message was already displayed 
               finalFillQuery.getNext();
               // [repositionByID] positioned the cursor just before the [rowids], we need to
               // advance a step to load the records in buffers 
            }
            
            boolean isAppend = FM_APPEND.equals(fillMode);
            boolean isMerge = FM_MERGE.equals(fillMode);
            boolean isReplace = FM_REPLACE.equals(fillMode);
            
            if (isReplace)
            {
               P2JIndex primaryIndex = recBuf.getDmoInfo().getPrimaryIndex(true);
               if (primaryIndex == null || !primaryIndex.isUnique())
               {
                  ErrorManager.recordOrShowError(11885);
                  // A unique primary index is required in the target table, each field of which is mapped
                  // to some source field, in order to do a replace-mode or parent-mode COPY/GET/MERGE/FILL
                  // type of operation on a dataset table.
                  dataSet.error(true);
                  return new logical(false);
               }
            }
            
            List<DataRelation> childRelations = dataSet.getRelations(this, true, false, true);
            
            // iterate the result and populate this buffer
            integer batchRemaining = new integer(batchSize > 0 ? batchSize : -1);
            LogicalOp test = () -> new logical(!finalFillQuery._isOffEnd() && batchRemaining.intValue() != 0);
            String lbl = "fill-buffer-" + name.toStringMessage();
            logical hasMore = new logical();
            BlockManager.repeatToWhile(TransactionType.SUB, lbl, null, test, new Block((Body) () ->
            {
               if (!fillOk.getValue())
               {
                  BlockManager.leave();
               }
               
               if (!dataSet.invokeCallback(self, callbacks.get(BEFORE_ROW_FILL_EVENT)))
               {
                  fillOk.assign(false);
                  BlockManager.leave();
               }
               
               this.create();
               
               Runnable update = () -> RecordBuffer.batch(() -> {
                  Set<Map.Entry<CaseInsensitiveString, DataSource.FieldReference>> entries = dataSrcMapping.entrySet();
                  for (Map.Entry<CaseInsensitiveString, DataSource.FieldReference> entry : entries)
                  {
                     CaseInsensitiveString myField = entry.getKey();
                     DataSource.FieldReference srcField = entry.getValue();
                     String crtField = myField.getValue();
                     
                     BufferImpl srcBuffer = srcField.getBuffer();
                     if (srcField.getName().equals(Session.PK))
                     {
                        dereference(crtField, srcBuffer.buffer().rowID());
                     }
                     else
                     {
                        assign(myField.getValue().toLowerCase(), srcField, srcBuffer);
                     }
                  }
                  
                  // update the DATASOURCE-ROWID attribute
                  if (srcBufCount == 1)
                  {
                     TempRecord newRec = (TempRecord) recBuf.getCurrentRecord();
                     newRec._datasourceRowid(srcBuffs[0].buffer().rowID().getValue());
                     
                     // NOTES:
                     // - the ORIGIN-ROWID can be queried for any buffer of the DataSource, if there are
                     //   more than one, see ORIGIN-ROWID(name).
                     // - in a recursive query, all the levels ROWID should also be stored to support the
                     //   ORIGIN-ROWID(n).
                     // - OTOH, in P4GL, not always the ORIGIN-ROWID is found even in single buffer
                     //   DataSource case.
                  }
                  
                  if (isMerge || isReplace)
                  {
                     boolean validated = false;
                     try
                     {
                        validated = recBuf.validate(false);
                     }
                     catch (ValidationException e)
                     {
                        // ignore, no 'expected' exception will be thrown
                     }
                     
                     // was there a unique conflict with this record?
                     if (!validated)
                     {
                        if (isMerge)
                        {
                           // merge mode: ignore it, just skip to next record
                           TransactionManager.disableLoopProtection(null);
                           hasMore.assign(finalFillQuery.getNext());
                           // we need to release the newly created record as if undone. We cannot call
                           // BlockManager.undoNext(lbl) here as it will affect the buffer(s) from the query
                           // driving the fill operation.
                           recBuf.release(true, false); // I.e. drop() - this form avoids a buffer() lookup
                           BlockManager.next(lbl);
                        }
                        else
                        {
                           // replace mode: eliminate competition, then try again on clean ground
                           try
                           {
                              OrmUtils.dropUniqueIndexConflicts(recBuf,
                                                                recBuf.getPersistence(),
                                                                recBuf.getMultiplexID());
                           }
                           catch (PersistenceException e)
                           {
                              throw new ErrorConditionException(
                                    "FILL: Failed to make room for new record in REPLACE mode", e);
                           }
                           // now there should not be any collisions and the new record should be flushed
                        }
                     }
                  }
               });
               
               if (isMerge || isReplace)
               {
                  recBuf.txHelper.errHlp.silent(update);
                  recBuf.txHelper.errHlp.setPending(false); // clear the possible 'ERROR' flag
               }
               else
               {
                  // APPEND or EMPTY mode
                  try
                  {
                     update.run();
                  }
                  catch (ErrorConditionException ex)
                  {
                     fillOk.assign(false);
                     ErrorManager.recordOrThrowError(11878, "-1", doGetName());
                  }
               }
               
               if (batchRemaining.intValue() > 0)
               {
                  batchRemaining.decrement();
               }
               
               if (!dataSet.invokeCallback(self, callbacks.get(AFTER_ROW_FILL_EVENT)))
               {
                  this.release();
                  fillOk.assign(false);
                  BlockManager.leave();
               }
               
               // call nested fill on child relation buffers
               for (DataRelation rel : childRelations)
               {
                  // if [reposition] attribute is set, the data-relation is ignored.
                  Buffer childBuffer = rel.getChildBuffer().unwrapBuffer();
                  if (rel._isReposition())
                  {
                     childBuffer.lastBatch(true);
                     continue;
                  }
                  
                  fillOk.assign(childBuffer.fill().getValue());
                  if (!fillOk.getValue()) 
                  {
                     break; // the for
                  }
               }
               
               this.release();
               
               if (!fillOk.getValue())
               {
                  BlockManager.leave(); // repeatToWhile
               }
               
               hasMore.assign(finalFillQuery.getNext());
            }));
            
            if (batchSize > 0 && batchRemaining.intValue() != 0 && childRelations.isEmpty())
            {
               // the batch ended before [batchSize] records were processed
               lastBatch = true;
            } 
            else if (!hasMore.booleanValue())
            {
               // the last invocation to fill-query:get-next() returned false so there are no more batches
               lastBatch = true;
            }
         }
         finally
         {
            dataSource.updateFillRowidInfo(fillOk.getValue());
            finalFillQuery.queryClose();
            // HandleOps.delete(hQuery);
            this.release();
            if (!wmOn)
            {
               errHlp.setWarningMode(false);
            }
         }
      }
      
      boolean ok = dataSet.invokeCallback(self, callbacks.get(AFTER_FILL_EVENT));
      ok = ok && fillOk.booleanValue();
      
      // NOTE: strange thing, I could not find any proof that the ERROR attribute of [this]
      //       BUFFER or the TEMP-TABLE is also altered by a FILL failure
      dataSet.error(!ok);
      return new logical(ok);
   }
   
   /**
    * Assigns the value of {@code srcBuffer.df} to field {@code f} of this buffer. Takes care of the extents.
    * 
    * @param   myField
    *          The destination field of this buffer.
    * @param   df
    *          The source field.
    * @param   srcBuffer
    *          The source buffer.
    */
   private void assign(String myField, DataSource.FieldReference df, BufferImpl srcBuffer)
   {
      int dstIndex = 0;
      String crtField = myField;
      int i = myField.indexOf('[');
      if (i != -1)
      {
         dstIndex = Integer.parseInt(myField.substring(i+1, myField.length() - 1));
         crtField = myField.substring(0, i);
      }
      DmoMeta dmoInfo = buffer().getDmoInfo();
      Property dstProp = dmoInfo.byLegacyName(crtField);
      
      if (dstProp == null)
      {
         return; // something is wrong... invalid [crtField] name. Raise OE condition?
      }
   
      if (dstProp.extent == 0)
      {
         if (df.getIndex() == 0)
         {
            dereference(crtField, srcBuffer.dereference(df.getName()));
         }
         else
         {
            dereference(crtField, srcBuffer.dereference(df.getIndex() /* -1? */ , df.getName()));
         }
      }
      else
      {
         if (dstIndex != 0)
         {
            if (df.getIndex() == 0)
            {
               dereference(dstIndex, crtField, srcBuffer.dereference(df.getName()));
            }
            else
            {
               dereference(dstIndex /* -1? */, crtField, srcBuffer.dereference(df.getIndex() /* -1? */ , df.getName()));
            }
         }
         else
         {
            // bulk copy: (assume) both source and destination are extent fields
            for (int k = 0; k < dstProp.extent; k++)
            {
               BufferField srcField = (BufferField) srcBuffer.bufferField(df.getName()).getResource();
               BufferField destField = (BufferField) bufferField(crtField).getResource();
               
               destField.changeValue(k + 1, srcField.value(k + 1));
            }
         }
      }
   }
   
   /**
    * Applies a callback procedure, which allows execution of a defined event without duplicating
    * the event procedure definition.
    *
    * @param   eventName
    *          The event whose callback will be called.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical applyCallback(String eventName)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(11949);
         // APPLY-CALLBACK is only for buffers that are members of datasets.
         return new logical(false);
      }
      
      if (!isCallbackName(eventName))
      {
         return new logical(false);
      }
      
      return new logical(invokeCallback(eventName.toUpperCase()));
   }
   
   /**
    * Retrieves the name of the internal procedure associated with the ABL callback for the
    * specified event.
    *
    * @param   eventName
    *          The name of the event.
    *
    * @return the name of the internal procedure associated with specified event.
    */
   @Override
   public character getCallbackProcName(String eventName)
   {
      if (!isCallbackName(eventName))
      {
         return new character();
      }
      
      CallbackData cbd = callbacks.get(eventName.toUpperCase());
      if (cbd == null)
      {
         return new character();
      }
      
      return new character(cbd.getTarget());
   }
   
   /**
    * Retrieves the handle of the procedure that contains the internal procedure associated
    * with the ABL callback for the specified event
    *
    * @param   eventName
    *          The name of the event.
    *
    * @return  a handle to the procedure that contains the callback procedure.
    */
   @Override
   public handle getCallbackProcContext(String eventName)
   {
      if (!isCallbackName(eventName))
      {
         return new handle();
      }
      
      CallbackData cbd = callbacks.get(eventName.toUpperCase());
      if (cbd == null || cbd.getProcedure() == null)
      {
         return new handle();
      }
      
      return new handle(cbd.getProcedure());
   }
   
   /**
    * Associates an internal procedure with an ABL callback event.
    *
    * @param   eventName
    *          The eventName of the event.
    * @param   callback
    *          The eventName of the internal procedure associated with the callback event.
    * @param   context
    *          A handle to a procedure that contains the internal procedure specified by
    *          {@code callback}. If not specified, {@code THIS-PROCEDURE} is used as the procedure
    *          context.
    *
    * @return  {@code true} when the method ends with success.
    */
   @Override
   public logical setCallbackProcedure(String eventName, String callback, handle context)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(11951);
         // SET-CALLBACK-PROCEDURE is only valid for buffers that are members of a dataset.
         return new logical(false);
      }
      
      if (!isCallbackName(eventName))
      {
         return new logical(false);
      }
      
      if (callback == null)
      {
         return new logical(true);
      }
      
      if (callback.isEmpty())
      {
         callbacks.remove(eventName.toUpperCase());
         return new logical(true);
      }
      
      if (context == null || context.isUnknown())
      {
         // defaulting to THIS-PROCEDURE
         context = ProcedureManager.thisProcedure();
      }
      else
      {
         if (!context._isValid() || !(context.getResource() instanceof PersistentProcedure))
         {
            ErrorManager.recordOrShowError(13273);
            // Third argument to SET-CALLBACK must be a valid procedure handle or object reference.
            return new logical(false);
         }
      }
      
      InvokeConfig callSite = getCallSite(eventName);
      callbacks.put(eventName.toUpperCase(), new CallbackData(callback, context, null, callSite));
      
      return new logical(true);
   }
   
   /**
    * Configures a callback. Creates an association between a method within a class instance or
    * an internal procedure within a persistent procedure, with an ABL callback event.
    *
    * @param   callbackName
    *          The name of a callback.
    * @param   routineName
    *          The name of a method or an internal procedure to be associated.
    * @param   context
    *          The context in which the callback will be executed.
    *
    * @return  {@code true} if operation is successful and {@code false} if the parameters are invalid.
    */
   @Override
   public logical setCallback(String callbackName, String routineName, object<?> context)
   {
      if (dataSet == null || !dataSet.valid())
      {
         ErrorManager.recordOrShowError(11951);
         // SET-CALLBACK-PROCEDURE is only valid for buffers that are members of a dataset.
         return new logical(false);
      }
      
      if (!isCallbackName(callbackName))
      {
         return new logical(false);
      }
      
      if (routineName == null)
      {
         return new logical(true);
      }
      
      if (routineName.isEmpty())
      {
         callbacks.remove(callbackName.toUpperCase());
         return new logical(true);
      }
      
      CallbackData cbd;
      if (context == null || context.isUnknown())
      {
         // defaulting to THIS-OBJECT or to THIS-PROCEDURE, depending on the context
         if (ObjectOps.thisObject().isUnknown())
         {
            InvokeConfig callSite = getCallSite(callbackName);
            cbd = new CallbackData(routineName, ProcedureManager.thisProcedure(), null, callSite);
         }
         else
         {
            cbd = new CallbackData(routineName, null, ObjectOps.thisObject(), null);
         }
      }
      else
      {
         cbd = new CallbackData(routineName, null, context, null);
      }
      
      callbacks.put(callbackName.toUpperCase(), cbd);
      
      return new logical(true);
   }
   
   /**
    * Loads the empty target {@code TempTable}) object with changed rows from original (source)
    * object. This can be either the BEFORE or the AFTER buffer. If the BEFORE/AFTER temp-tables
    * are already populated, they are emptied before copying the changes.
    *
    * @param   original
    *          Handle to original object {@code DataSet} buffer (BEFORE or AFTER buffer) that
    *          contains the changed rows to load into the target object.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical getChanges(handle original)
   {
      if (original == null || original.isUnknown() || !original.isType(LegacyResource.BUFFER))
      {
         ErrorManager.recordOrShowError(11947);
         // bufferhdl:GET-CHANGES must have a valid buffer handle to the original table, from which changes are coming, as the first argument.
         return new logical(false);
      }
      
      // check the argument first: it must be an AFTER or BEFORE buffer:
      BufferImpl srcBefore = (BufferImpl) original.unwrapBuffer();
      BufferImpl srcAfter = null;
      handle xHandle = srcBefore.beforeBuffer();
      if (xHandle.isUnknown())
      {
         xHandle = srcBefore.afterBuffer();
         if (xHandle.isUnknown())
         {
            ErrorManager.recordOrShowError(11925, srcBefore.doGetName());
            // GET/MERGE-CHANGES require a BEFORE-TABLE for table <name>.
            return new logical(false);
         }
         else
         {
            srcAfter = (BufferImpl) xHandle.unwrapBuffer(); 
         }
      }
      else
      {
         // oops, [xHandle] is the BEFORE buffer and [srcBefore] is the AFTER buffer, fixing:
         srcAfter = srcBefore;
         srcBefore = (BufferImpl) xHandle.unwrapBuffer();
      }
      
      // now check the current buffer: it also must be an AFTER or BEFORE buffer:
      BufferImpl dstBefore = null;
      BufferImpl dstAfter = null;
      xHandle = beforeBuffer();
      if (xHandle.isUnknown())
      {
         xHandle = afterBuffer();
         if (xHandle.isUnknown())
         {
            ErrorManager.recordOrShowError(11925, doGetName());
            // GET/MERGE-CHANGES require a BEFORE-TABLE for table <name>.
            return new logical(false);
         }
         else 
         {
            dstBefore = this;
            dstAfter = (BufferImpl) xHandle.unwrapBuffer();
         }
      }
      else
      {
         dstBefore = (BufferImpl) xHandle.unwrapBuffer();
         dstAfter = this;
      }
      
      // @this moment we have the dst/src-Before/After initialized. Be sure dsts are empty first:
      dstBefore.setUpBeforeBuffer(true);
      AbstractTempTable dstTT = (AbstractTempTable) dstAfter.tableHandle().getResource();
      logical trackingChanges = dstTT.isTrackingChanges();
      dstTT.setTrackingChanges(false);
      try
      {
         if (!dstBefore.deleteAll().getValue() || !dstAfter.deleteAll().getValue())
         {
            return new logical(false);
         }
      }
      finally
      {
         // restore back
         if (!trackingChanges.isUnknown() && trackingChanges.booleanValue())
         {
            dstTT.setTrackingChanges(true);
         }
         dstBefore.setUpBeforeBuffer(false);
      }
      
      // ... and flush any content the source buffers may contain:
      srcBefore.validate();
      srcAfter.validate();
      
      boolean ok = TemporaryBuffer.copyChanges((TemporaryBuffer) srcBefore.buffer(),
                                               (TemporaryBuffer) srcAfter.buffer(),
                                               (TemporaryBuffer) dstBefore.buffer(),
                                               (TemporaryBuffer) dstAfter.buffer());
      
      // in 4GL, "there is no guarantee as to the contents of the buffers involved, when the method returns
      // to add predictability, FWD makes sure they are empty
      srcBefore.release();
      srcAfter.release();
      
      if (ok)
      {
         TempTable origin = TempTableBuilder.asTempTable((Temporary) srcAfter);
         ((AbstractTempTable) TempTableBuilder.asTempTable((Temporary) dstBefore)).setOriginTable(origin);
         ((AbstractTempTable) TempTableBuilder.asTempTable((Temporary) dstAfter)).setOriginTable(origin);
      }
      return new logical(ok);
   }
   
   /**
    * Loads the empty target ({@code DataSet} or {@code TempTable}) object with changed rows from
    * original (source) object.
    *
    * @param   original
    *          Handle to original object ({@code DataSet} or {@code TempTable}) that contains the
    *          changed rows to load into the target object.
    * @param   parentMode
    *          When {@code TRUE} the target gets the changed parent rows of each changed child row.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical getChanges(handle original, logical parentMode)
   {
      return getChanges(original, parentMode.getValue());
   }
   
   /**
    * Loads the empty target ({@code DataSet} or {@code TempTable}) object with changed rows from
    * original (source) object.
    *
    * @param   original
    *          Handle to original object ({@code DataSet} or {@code TempTable}) that contains the
    *          changed rows to load into the target object.
    * @param   parentMode
    *          When {@code TRUE} the target gets the changed parent rows of each changed child row.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical getChanges(handle original, boolean parentMode)
   {
      ErrorManager.recordOrThrowError(12778);
      // GET-PARENT mode not supported for buffer GET-CHANGES.
      return new logical(false);
   }
   
   /**
    * Merges single changed row of a single temp-table in a source {@code DataSet} object loaded
    * with the {@code GET-CHANGES()} method into the corresponding rows of a single temp-table
    * into the original (target) {@code DataSet} buffer.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeRowChanges()
   {
      return mergeRowChanges(null, false);
   }
   
   /**
    * Merges single changed row of a single temp-table in a source {@code DataSet} object loaded
    * with the {@code GET-CHANGES()} method into the corresponding rows of a single temp-table
    * into the original (target) {@code DataSet} buffer.
    *
    * @param   original
    *          A handle to the original (target) buffer object of a single target {@code DataSet}
    *          temp-table to merge with the changed rows from the source object.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeRowChanges(handle original)
   {
      return mergeRowChanges(original, false);
   }
   
   /**
    * Merges single changed row of a single temp-table in a source {@code DataSet} object loaded
    * with the {@code GET-CHANGES()} method into the corresponding rows of a single temp-table
    * into the original (target) {@code DataSet} buffer.
    *
    * @param   original
    *          A handle to the original (target) buffer object of a single target {@code DataSet}
    *          temp-table to merge with the changed rows from the source object.
    * @param   allCopyMode
    *          Use {@code TRUE} to indicates that all rows of the source after-image table must be
    *          merged.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeRowChanges(handle original, boolean allCopyMode)
   {
      if (!_available())
      {
         ErrorManager.recordOrShowError(11937, doGetName());
         // MERGE-ROW-CHANGES buffer <name> not available.
         return new logical(false);
      }
      
      BufferImpl afterBufferChanges = null;
      if (!isBeforeBuffer())
      {
         if (!isAfterBuffer())
         {
            ErrorManager.recordOrShowError(11925, doGetName());
            // GET/MERGE-CHANGES require a BEFORE-TABLE for table <name>.
            return new logical(false);
         }
         else
         {
            afterBufferChanges = this;
         }
      }
      else
      {
         afterBufferChanges = (BufferImpl) afterBuffer().unwrapBuffer();
      }
      
      BufferImpl origBuffer;
      TempTable originTable;
      if (original == null)
      {
         TempTable parentTable = this.buffer().tableHandle().unwrapTempTable();
         handle originTableHandle = parentTable.getOriginHandle();
         if (originTableHandle.isUnknown() || !originTableHandle._isValid())
         {
            ErrorManager.recordOrShowError(11944, afterBufferChanges.doGetName());
            // MERGE-ROW-CHANGES finds ORIGIN-HANDLE in table <name> is no longer valid.
            return new logical(false);
         }
         originTable = originTableHandle.unwrapTempTable();
         origBuffer = (BufferImpl) originTable.defaultBufferHandle().unwrapBuffer(); 
      }
      else
      {
         if (!original._isValid() || !original.isType(LegacyResource.BUFFER))
         {
            ErrorManager.recordOrShowError(11943);
            // First argument to MERGE-ROW-CHANGES must be a valid BUFFER handle.
            return new logical(false);
         }
         origBuffer = (BufferImpl) original.unwrapBuffer();
         originTable = origBuffer.tableHandle().unwrapTempTable();
      }
      
      if (originTable.isTrackingChanges().getValue())
      {
         ErrorManager.recordOrShowError(11922, originTable.name().toStringMessage());
         // MERGE-CHANGES may not be run on table <name> with TRACKING-CHANGES on.
         return new logical(false);
      }
      
      return new logical(mergeRowChangesImpl(this.ref(), origBuffer.ref()));
   }
   
   /**
    * Merges single changed row of a single temp-table in a source {@code DataSet} object loaded
    * with the {@code GET-CHANGES()} method into the corresponding rows of a single temp-table
    * into the original (target) {@code DataSet} buffer.
    *
    * @param   original
    *          A handle to the original (target) buffer object of a single target {@code DataSet}
    *          temp-table to merge with the changed rows from the source object.
    * @param   allCopyMode
    *          Use {@code TRUE} to indicates that all rows of the source after-image table must be
    *          merged.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeRowChanges(handle original, logical allCopyMode)
   {
      return mergeRowChanges(original, !allCopyMode.isUnknown() && allCopyMode.getValue());
   }
   
   /**
    * Merges the changed rows of all temp-tables or a single temp-table in a source
    * {@code DataSet} object loaded with the {@code GET-CHANGES()} method into the corresponding
    * rows of all temp-tables or a single temp-table (respectively) in the original (target)
    * {@code DataSet} object.
    *
    * @param   original
    *          A handle to the original (target) {@code DataSet} object or the buffer object of a
    *          single target {@code DataSet} temp-table to merge with the changed rows from the
    *          source object.
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeChanges(handle original)
   {
      return mergeChanges(original, false);
   }
   
   /**
    * Merges the changed rows of all temp-tables or a single temp-table in a source
    * {@code DataSet} object loaded with the {@code GET-CHANGES()} method into the corresponding
    * rows of all temp-tables or a single temp-table (respectively) in the original (target)
    * {@code DataSet} object.
    *
    * @param   original
    *          A handle to the original (target) {@code DataSet} object or the buffer object of a
    *          single target {@code DataSet} temp-table to merge with the changed rows from the
    *          source object.
    * @param   allCopyMode
    *          Use {@code TRUE} to indicates that all rows of the source after-image table must be
    *          merged.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeChanges(handle original, logical allCopyMode)
   {
      return mergeChanges(original, !allCopyMode.isUnknown() && allCopyMode.getValue());
   }
   
   /**
    * Merges the changed rows of all temp-tables or a single temp-table in a source {@code DataSet} object
    * loaded with the {@code GET-CHANGES()} method into the corresponding rows of all temp-tables or a single
    * temp-table (respectively) in the original (target) {@code DataSet} object.
    *
    * @param   original
    *          A handle to the original (target) {@code DataSet} object or the buffer object of a single
    *          target {@code DataSet} temp-table to merge with the changed rows from the source object.
    * @param   allCopyMode
    *          Use {@code TRUE} to indicates that all rows of the source after-image table must be merged.
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical mergeChanges(handle original, boolean allCopyMode)
   {
      // the parameter must be a valid dataset handle:
      if (original == null || original.isUnknown() || !original.isType(LegacyResource.BUFFER))
      {
         ErrorManager.recordOrShowError(11953);
         // First argument to MERGE-CHANGES must be a valid BUFFER handle.
         return new logical(false);
      }
      
      BufferImpl origBuf = (BufferImpl) original.getResource();
      origBuf = origBuf.ref();
      TempTable origTT = (TempTable) tableHandle().unwrapTempTable().getOriginHandle().getResource();
      if (origTT != origBuf.tableHandle().unwrapTempTable())
      {
         ErrorManager.recordOrShowError(11923, origBuf.doGetName(), doGetName());
         // MERGE-CHANGES run on original table <name> that does not match the ORIGIN-HANDLE in the GET-CHANGES table <name>.
         return new logical(false);
      }
      
      if (origTT.isTrackingChanges().getValue())
      {
         ErrorManager.recordOrShowError(11922, origTT.name().toStringMessage());
         // MERGE-CHANGES may not be run on table <name> with TRACKING-CHANGES on.
         return new logical(false);
      }
      
      // make sure we use the AFTER-TABLE buffer
      if (origBuf.isBeforeBuffer())
      {
         origBuf = origBuf.peerBuffer;
      }
      
      if (isAfterBuffer())
      {
         return new logical(mergeChangesImpl(origBuf, allCopyMode));
      }
      else if (isBeforeBuffer())
      {
         return new logical(peerBuffer.mergeChangesImpl(origBuf, allCopyMode));
      }
      
      ErrorManager.recordOrShowError(11925, doGetName());
      // GET/MERGE-CHANGES require a BEFORE-TABLE for table <name>. 
      return new logical(true);
   }
   
   /**
    * Accepts changes to the data in target object (a temp-table or all temp-tables in a {@code DataSet}).
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical acceptChanges()
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11887, doGetName());
         // ACCEPT-CHANGES run on table <name> that has no BEFORE-TABLE.
         return new logical(false);
      }
      
      RecordBuffer beforeBuf = buffer();
      Persistence persistence = beforeBuf.getPersistence();
      BufferImpl after = (BufferImpl) afterBuffer().unwrapBuffer();
      RecordBuffer afterBuf = after.buffer();
      
      // invalidate index-based caches because of low-level operations:
      afterBuf.invalidateFFCache(0);
      beforeBuf.invalidateFFCache(0);
      
      // execute a simple query to update the [_rowState], [_originRowid] and [_peerRowid] fields to default
      // values in the AFTER-BUFFER. The [_dataSourceRowid] remains unmodified. 
      StringBuilder buf = new StringBuilder();
      buf.append("update ").append(afterBuf.getTable()).append(" set ")
         .append(Buffer.ROW_STATE_FIELD).append("=").append(Buffer.ROW_UNMODIFIED).append(", ")
         .append(Buffer.ORIGIN_ROWID_FIELD).append("=NULL, ")
         .append(Buffer.AFTER_ROWID_FIELD).append("=NULL")
         .append(" where ")
         .append(TemporaryBuffer.MULTIPLEX_FIELD_NAME).append(" = ").append(afterBuf.getMultiplexID());
      
      try
      {
         boolean noUndo = !beforeBuf.isUndoable();
         int cnt = persistence.executeSQL(buf.toString(), null, noUndo);
         afterBuf.pruneSessionCache();
         if (log.isLoggable(Level.FINE))
         {
            log.log(Level.FINE, "ACCEPT-CHANGES accepted " + cnt + " changes from " + beforeBuf.getDMOName());
         }
      }
      catch (PersistenceException e)
      {
         return new logical(false);
      }
      
      // then drop the before-image data for this table:
      try
      {
         beforeBuf.delete(TemporaryBuffer.MULTIPLEX_FIELD_NAME + "=?", beforeBuf.getMultiplexID());
      }
      catch (PersistenceException e)
      {
         if (log.isLoggable(Level.WARNING))
         {
            log.log(Level.WARNING, 
                    "ACCEPT-CHANGES: Failed to drop before images of " + afterBuf.getDMOName());
         }
      }
      
      beforeBuf.pruneSessionCache();
      
      ((AbstractTempTable) afterBuf.tableHandle().getResource()).setOriginTable(null);
      ((AbstractTempTable) beforeBuf.tableHandle().getResource()).setOriginTable(null);
      
      return new logical(true);
   }
   
   /**
    * Rejects changes to the data in target object (a temp-table or all temp-tables in a
    * {@code DataSet}).
    *
    * @return  {@code true} if operation is successful.
    */
   @Override
   public logical rejectChanges()
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11919, doGetName());
         // REJECT-CHANGES run on table <name> that has no BEFORE-TABLE.
         return new logical(false);
      }
      
      TemporaryBuffer buffer = (TemporaryBuffer) buffer();
      try
      {
         buffer.flush();
      }
      catch (ValidationException e)
      {
         // normally this should not happen because before tables do not have unique indexes
         // TODO: but if it still happens? which error we we throw?
         return new logical(false);
      }
      return new logical(buffer.rejectChanges((BufferImpl) afterBuffer().unwrapBuffer()));
   }
   
   /**
    * Tests whether the table from this buffer has a legacy field named {@code legFieldName}.
    * 
    * @param   legFieldName
    *          The legacy field name to be tested.
    *
    * @return  {@code true} if the buffer has a legacy field as specified.
    */
   protected boolean hasField(String legFieldName)
   {
      return buffer().getDmoInfo().byLegacyName(legFieldName) != null;
   }
   
   /**
    * Tests whether the table from this buffer has a legacy field named {@code legFieldName}.
    * Unlike the {@code #hasField} method, this method will validate the field names that also
    * contain the buffer name as qualification.
    * 
    * @param   legFieldName
    *          The field name to be tested.
    *
    * @return  {@code true} if the buffer has a legacy field as specified.
    */
   protected boolean hasFieldEx(String legFieldName)
   {
      // normalize name (redundant with DmoMeta.byLegacyName below, but needed for pre-processing)
      legFieldName = TextOps.rightTrimLower(legFieldName);
      
      // check whether this is a scoped name
      if (legFieldName.contains("."))
      {
         String prefix = name.toStringMessage().toLowerCase() + ".";
         if (legFieldName.startsWith(prefix))
         {
            legFieldName = legFieldName.substring(prefix.length());
         }
         else
         {
            return false; // not this buffer
         }
      }
      
      if (legFieldName.equals(Session.PK))
      {
         return true;
      }
      
      return buffer().getDmoInfo().byLegacyName(legFieldName) != null;
   }
   
   /**
    * Actual setting of the query. This is called indirectly, by FWD from the query when this
    * buffer is set / added.
    * 
    * @param   qry
    *          The query to associate.
    */
   protected void setQuery(handle qry)
   {
      if (qry == null || qry.isUnknown() || !qry._isValid())
      {
         this.lastQuery = null;
         return;
      }
      
      this.lastQuery = new handle(qry.getResource());
   }
   
   /**
    * Sets the current 'parent' dataset. A buffer can be in at most one dataset at a specific
    * time. Trying to add it to another one will generate error 11860. This is checked in
    * dataset add... procedures.
    * 
    * @param   ds
    *          The new parent {@code DataSet}.
    */
   protected void setDataSet(DataSet ds)
   {
      if (ds != null && this.dataSet != null && this.dataSet.valid())
      {
         System.err.println("New dataset configured but one is already active.");
         // return internal error ?
      }
      
      detachDataSource();
      dataSet = (ds == null) ? null : ds.ref();
      lastBatch = false;
   }
   
   /**
    * Flags the bracketing internal management of BEFORE-BUFFER. ABL programmer's direct
    * actions (CREATE/DELETE/ALTER buffer) are outside this bracket and will lead to errors.
    * The flag is forwarded to {@code RecordBuffer}.
    * 
    * @param   on
    *          {@code true} to activate bracket, {@code false} to end it and activate errors.
    */
   public void setUpBeforeBuffer(boolean on)
   {
      this.setUpBeforeBuffer = on;
   }
   
   /**
    * Checks whether we are out of BEFORE-TABLE management mode. In this mode, changes to the
    * record in buffer, including CREATE/DELETE <strong>WILL</strong> throw errors.
    * 
    * @return  {@code false} when {@link #setUpBeforeBuffer} is set.
    */
   public boolean hasBeforeBufferConstraints()
   {
      return !setUpBeforeBuffer;
   }
   
   /**
    * Tests whether this is a {@code BEFORE-BUFFER} of a dataset.
    * 
    * @return  {@code true} if this is a {@code BEFORE-BUFFER} of a dataset.
    */
   public boolean isBeforeBuffer()
   {
      if (beforeBufferType == BeforeType.UNKNOWN)
      {
         computeBufferType();
      }
      
      return beforeBufferType == BeforeType.BEFORE;
   }
   
   /**
    * Tests whether this is a {@code AFTER-BUFFER} of a dataset.
    *
    * @return  {@code true} if this is a {@code AFTER-BUFFER} of a dataset.
    */
   public boolean isAfterBuffer()
   {
      if (beforeBufferType == BeforeType.UNKNOWN)
      {
         computeBufferType();
      }
      
      return beforeBufferType == BeforeType.AFTER;
   }
   
   /**
    * Constructs and returns a string containing the P4GL schema definition of the table. The result is not
    * cached so this method is to be used only for debug.
    *
    * @return  a string containing the P4GL legacy schema definition of the table.
    */
   public String tableDefinition()
   {
      return buffer().getDmoInfo().getTableDefinition();
   }
   
   /**
    * Computes the buffer type and if it is an {@code AFTER} or a {@code BEFORE} the
    * peer (complementary) buffer is also stored in {@code peerBuffer}.
    */
   void computeBufferType()
   {
      if (!(this instanceof Temporary))
      {
         peerBuffer = null;
         beforeBufferType = BeforeType.SIMPLE;
         return;
      }
      
      TempTable tempTable = TempTableBuilder.asTempTable((Temporary) this);
      
      handle ttBefore = tempTable.getBeforeTable();
      if (!ttBefore.isUnknown())
      {
         peerBuffer = (BufferImpl) ttBefore.unwrapTempTable().defaultBufferHandle().getResource();
         peerBuffer = peerBuffer.ref();
         beforeBufferType = BeforeType.AFTER;
         return;
      }
      
      handle ttAfter = tempTable.getAfterTable();
      if (!ttAfter.isUnknown())
      {
         peerBuffer = (BufferImpl) ttAfter.unwrapTempTable().defaultBufferHandle().getResource();
         peerBuffer = peerBuffer.ref();
         beforeBufferType = BeforeType.BEFORE;
         return;
      }
      
      peerBuffer = null;
      beforeBufferType = BeforeType.SIMPLE;
   }
   
   /**
    * Invokes a defined callback if one exist.
    *
    * @param   event
    *          Event name.
    *
    * @return  {@code false} if the callback raised an error. If the callback does not exist or ended with
    *          success {@code true} is returned.
    */
   boolean invokeCallback(String event)
   {
      if (dataSet == null)
      {
         return true;
      }
      
      CallbackData cbd = callbacks.get(event);
      if (cbd == null)
      {
         return true;
      }
      
      return dataSet.invokeCallback(new handle(this), cbd);
   }
   
   /**
    * Set NAME attribute without any checks.
    *
    * @param name
    *        Name to set.
    */
   protected void doSetName(String name)
   {
      this.name.assign(name);
   }
   
   /**
    * Get the {@code NAME} attribute as java string.
    *
    * @return  the {@code NAME} attribute as a java string. {@code null} if {@code NAME} is {@code unknown}.
    */
   public String doGetName()
   {
      return this.name.toJavaType();
   }
   
   /**
    * Makes a backup of the current name and it replaces it with another one.
    * This is usually used by an argument proxy to temporarily replace the name.
    * The override is successful only if the name wasn't overridden beforehand.
    * 
    * @param     newName
    *            The new name which should be temporarily used by this Buffer.
    *            
    * @return    true if the overriding was successful, false otherwise.
    */
   protected boolean overrideName(String newName)
   {
      if (oldName != null)
      {
         return false;
      }
      
      oldName = new character(name);
      name.assign(newName);
      return true;
   }
   
   /**
    * Restores the name by replacing the current name with the one stored in the backup.
    */
   protected void restoreName()
   {
      name.assign(oldName);
      oldName = null;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute. Since the session Buffer list
    * is a simple linked list, the returned value is {@code true}. The chain is one-directional,
    * starting at SESSION:FIRST-BUFFER.
    * <p> 
    * Only the {@link #dynamic} buffers are stored in the chain. The method will return false for
    * static buffers. 
    * 
    * @return   {@code true} only for dynamic buffers.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return buffer() != null && _dynamic();
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute. Since the session Buffer list
    * is only a simple linked list, the returned value is {@code false}. The buffers chain is
    * one-directional, starting at SESSION:FIRST-BUFFER. There is no LAST-BUFFER attribute
    * associated with the SESSION handle.
    * 
    * @return   Always {@code false}.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Interlinks the newly created HandleChain in the both chains. The back and forth links are
    * created with respective elements.
    * The heads of the chains are updated: if this is the first of its kind, it becomes the first
    * element of the chain. It is always stored as the last resource in his chain.
    * Even if this kind of resource does not support NEXT or PREV sibling, both heads of the
    * chain are kept for a fast access in future maintenance.
    * <p>
    * BUFFERs use special order: TEMP-TABLES (static and dynamic in the order they are added)
    * first, then PERMANENT buffers at the end of the list.
    */
   @Override
   protected void interlink()
   {
      if (!hasNextSibling() && !hasPrevSibling())
      {
         // quick out this resource is not chain-able
         return;
      }
      
      String type = type();
      boolean permanentTable = this.tableHandle().isUnknown();
      
      BufferImpl firstPermanent = (BufferImpl) HandleChain.getHead(PERMANENT_TABLE);
      if (permanentTable)
      {
         // add buffer at the end, like other resources, but don't forget to update the 
         // first-permanent-buffer pointer:
         if (firstPermanent == null)
         {
            // there are no others, this is the first permanent buffer
            HandleChain.setHead(PERMANENT_TABLE, this);
         }
         // then we continue processing like any other resource, 
         // appending this resource at the end of the list 
      }
      else
      {
         // we must insert this TEMP buffer just before the FIRST-PERMANENT pointer
         if (firstPermanent != null)
         {
            if (firstPermanent.prevSibling == null)
            {
               // there are no TEMP-TABLE buffers? then install in front
               HandleChain.setHead(type, this);
               firstPermanent.prevSibling = this;
               this.nextSibling = firstPermanent;
               this.prevSibling = null;
            }
            else
            {
               // there are both TEMP and PERMANENT TABLE buffers, insert in between
               this.nextSibling = firstPermanent;
               this.prevSibling = firstPermanent.prevSibling;
               ((BufferImpl)this.nextSibling).prevSibling = this;
               ((BufferImpl)this.prevSibling).nextSibling = this;
            }
            return; // all processing done here
         }
         // else there are no other permanent buffers, continue processing like any other
         // resource, appending this resource at the end of the list 
      }
      
      super.interlink();
   }
   
   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!_dynamic() && !implicitDeletion())
      {
         // static buffers can' be deleted, but no error message is shown
         return false;
      }
      
      handle ht = tableHandle();
      RecordBuffer recBuf = buffer();
      boolean staticTT = recBuf.isTemporary() && !recBuf.isDynamic();
      if (ht._isValid() && staticTT)
      {
         TempTable tt = ht.unwrapTempTable();

         // if this is the default buffer for a GLOBAL SHARED TEMP-TABLE, then do not delete
         if (recBuf.isGlobal() && tt.defaultBufferHandle().get() == this)
         {
            return false;
         }
         
         // if there are still other buffers bounded to it, then do not delete.
         TemporaryBuffer tbuf = (TemporaryBuffer) recBuf;
         if (tbuf.getMasterBuffer() == tbuf && tbuf.hasExplicitBuffers())
         {
            return false;
         }
      }
      
      deleting = true;
      recBuf.finished();
      recBuf.resourceDeleted();
      
      SharedVariableManager.cleanupBuffer(recBuf.getDMOProxy());
      
      deleted = true;
      deleting = false;
      
      if (staticTT)
      {
         TemporaryBuffer tbuf = (TemporaryBuffer) recBuf;
         TemporaryBuffer mbuf = (TemporaryBuffer) tbuf.getMasterBuffer();
         if (!mbuf.isGlobal() && mbuf != tbuf && !mbuf.hasExplicitBuffers())
         {
            handle htt = mbuf.tableHandle();
            if (htt._isValid() && !((TempTable) htt.getResource()).instantiatingProcedure()._isValid())
            {
               // if the master table is referenced only from other programs and its program has been deleted,
               // and this is the last child buffer referencing it, then delete it, too.
               StaticTempTable stt = (StaticTempTable) htt.getResource();
               stt.setForceDelete(true);
               
               ProcedureManager.deleteResource(null, stt);
            }
         }
      }
      
      if (fields != null)
      {
         for (int i = 0; i < fields.length; i++)
         {
            if (fields[i] != null)
            {
               fields[i].forceDelete();
            }
         }
      }
      
      return true;
   }
   
   /**
    * Check whether an extent field index is within the valid range {@code 0 &le; index &lt; max}.
    * Raise an error if not, unless errors are silenced.
    * <p>
    * This method is meant to be called by DMO proxy objects which extend this class, from
    * within their indexed setter and getter methods which accept an {@code int} value for the
    * element index.
    * 
    * @param   index
    *          Zero-based extent field index to check.
    * @param   max
    *          Maximum number of elements in the extent field.
    * 
    * @return  {@code true} if {@code index} falls within the valid range,
    *          {@code false} if outside the range and errors are silenced.
    * 
    * @throws  ErrorConditionException
    *          if {@code index} is outside the valid range and errors are not silenced.
    */
   protected boolean checkExtentRange(int index, int max)
   {
      if (index < 0 || index >= max)
      {
         outOfBoundsError(index);
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Record or throw an "array subscript out of range" error condition for the given subscript.
    * 
    * @param   index
    *          Zero-based subscript value.
    */
   protected void outOfBoundsError(int index)
   {
      String msg = "Array subscript " + (index + 1) + " is out of range";
      ErrorManager.recordOrThrowError(26, msg);
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @param   validated
    *          <code>true</code> if there was validation and it passed. <code>false</code> if
    *          there was no validation.
    *
    * @return  {@code true} if the delete was possible.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   protected logical deleteRecord(boolean validated)
   {
      RecordBuffer recBuf = buffer();
      
      if (!recBuf.isTemporary())
      {
         if (!verifyTransaction(recBuf))
         {
            return new logical(false);
         }
         
         if (!validated)
         {
            Class<? extends DataModelObject> iface = recBuf.getDMOInterface();
            if (TableMapper.hasValidation(iface))
            {
               String msg = "Table " + recBuf.getDmoInfo().legacyTable + " is defined with schema delete " +
                            "validation and may only be deleted by the DELETE statement";
               ErrorManager.recordOrShowError(7347, msg, false, false);
               
               return new logical(false);
            }
         }
      }
      
      if (dataSet != null) // implies buffer.isTemporary()
      {
         // checks whether tracking is active
         if (recBuf.getParentTable().isTrackingChanges().toJavaType())
         {
            if (!invokeCallback(ROW_DELETE_EVENT))
            {
               returnError(); // veto on record delete
            }
         }
      }
      
      try
      {
         recBuf.delete();
      }
      catch (PersistenceException exc)
      {
         return new logical(false);
      }
      
      return new logical(true);
   }
   
   /**
    * Test whether a string is a valid CALLBACK name for {@code DataSet}s. This method is
    * responsible of raising the 11950 error if the name specified is unknown or nor not a name
    * of a known callback for {@code DataSet}.
    *
    * @param   someName
    *          The name to test.
    *
    * @return  {@code true} only if {@code someName} is equals to "BEFORE-FILL", "AFTER-FILL",
    *          "BEFORE-ROW-FILL", "AFTER-ROW-FILL", "ROW-CREATE", "ROW-DELETE", "ROW-UPDATE",
    *          "FIND-FAILED", or "SYNCHRONIZE", in ignore-case mode.
    */
   private static boolean isCallbackName(String someName)
   {
      if (someName == null ||
         !BEFORE_FILL_EVENT.equalsIgnoreCase(someName)     &&
         !AFTER_FILL_EVENT.equalsIgnoreCase(someName)      &&
         !BEFORE_ROW_FILL_EVENT.equalsIgnoreCase(someName) &&
         !AFTER_ROW_FILL_EVENT.equalsIgnoreCase(someName)  &&
         !ROW_CREATE_EVENT.equalsIgnoreCase(someName)      &&
         !ROW_DELETE_EVENT.equalsIgnoreCase(someName)      &&
         !ROW_UPDATE_EVENT.equalsIgnoreCase(someName)      &&
         !FIND_FAILED_EVENT.equalsIgnoreCase(someName)     &&
         !SYNCHRONIZE_EVENT.equalsIgnoreCase(someName))
      {
         ErrorManager.recordOrShowError(11950, someName != null ? someName : "");
         // Argument  to CALLBACK method must be a valid buffer or dataset event name.
         return false;
      }
      
      return true;
   }
   
   /**
    * Get the {@link InvokeConfig callsite} for the specified event.
    * 
    * @param    eventName
    *           The event name.
    */
   private static InvokeConfig getCallSite(String eventName)
   {
      switch (eventName.toUpperCase())
      {
         case BEFORE_FILL_EVENT:
            return BEFORE_FILL_CALL_SITE;
         case AFTER_FILL_EVENT:
            return AFTER_FILL_CALL_SITE;
         case BEFORE_ROW_FILL_EVENT:
            return BEFORE_ROW_FILL_CALL_SITE;
         case AFTER_ROW_FILL_EVENT:
            return AFTER_ROW_FILL_CALL_SITE;
         case ROW_CREATE_EVENT:
            return ROW_CREATE_CALL_SITE;
         case ROW_DELETE_EVENT:
            return ROW_DELETE_CALL_SITE;
         case ROW_UPDATE_EVENT:
            return ROW_UPDATE_CALL_SITE;
         case FIND_FAILED_EVENT:
            return FIND_FAILED_CALL_SITE;
         case SYNCHRONIZE_EVENT:
            return SYNCHRONIZE_CALL_SITE;
      }
      
      return null;
   }
   
   /**
    * Return {@code LockType} instance for the given internal lock and nowait with corresponding
    * error processing.
    *
    * @param   lock
    *          Integer value.
    * @param   nowait
    *          Integer value.
    *
    * @return  {@code LockType} instance.
    *
    * @throws  ErrorConditionException
    *          If locking operation fails.
    */
   private static LockType fromLockNowait(long lock, long nowait)
   throws ErrorConditionException
   {
      LockType lockType = null;
      try
      {
         lockType = LockType.fromLockNowait(lock, nowait);
      }
      catch (LockTypeException e)
      {
         String err;
         int num;
         if (e.isNowaitCorrect())
         {
            err = "FIND-BY-ROWID, FIND-FIRST etc., second argument must be " +
                  "NO-LOCK, SHARE-LOCK or EXCLUSIVE-LOCK";
            num = 7344;
         }
         else
         {
            err = "FIND method argument must be NO-WAIT, or 0 or ?";
            num = 7343;
         }
         ErrorManager.recordOrThrowError(num, err);
      }
      return lockType;
   }

   /**
    * Implementation of the SERIALIZE-ROW method. Writes current row from this buffer into the specified data 
    * target, using the specified format.
    *
    * @param   format
    *          The format ("xml" or "json").
    * @param   targetData
    *          Data target.
    * @param   formatted
    *          {@code True} to format the data for making it more readable; {@code false} to write it
    *          as unformatted text. Unknown value is treated as {@code false}.
    * @param   encoding
    *          Name of the character encoding to use.
    * @param   omitInitial
    *          {@code True} to omit data for fields whose values match their initial values.
    * @param   omitOuter
    *          {@code True} to omit data for the outermost object from the output.
    *
    * @return  {@code True} if the operation completed successfully, else {@code false}.
    */
   private boolean serializeRowImpl(String format,
                                    TargetData targetData,
                                    Boolean formatted,
                                    String encoding,
                                    Boolean omitInitial,
                                    Boolean omitOuter)
   {
      if (!buffer().isTemporary())
      {
         ErrorManager.recordOrShowError(13158);
         // XML Methods only allowed on TEMP-TABLE buffers.
         return false;
      }
      
      // the buffer must not be empty
      if (!_available())
      {
         ErrorManager.recordOrShowError(16017);
         // No record in buffer for SERIALIZE-ROW.
         return false;
      }
      
      // check the specified target format
      if (format == null)
      {
         ErrorManager.recordOrShowError(16019, "unknown", "");
         // Invalid target-format '<format>' for SERIALIZE-ROW.
         return false;
      }
      boolean isXML = "XML".equalsIgnoreCase(format);
      boolean isJSON = "JSON".equalsIgnoreCase(format);
      if (!isJSON && !isXML)
      {
         ErrorManager.recordOrShowError(16019, format, "");
         // Invalid target-format '<format>' for SERIALIZE-ROW.
         return false;
      }
      
      if (formatted == null)
      {
         formatted = false;
      }
      if (omitInitial == null)
      {
         omitInitial = false;
      }
      if (omitOuter == null)
      {
         omitOuter = false;
      }
      
      if (isXML)
      {
         if (omitOuter)
         {
            ErrorManager.recordOrShowError(16020);
            // omit-outer-object argument not allowed in SERIALIZE-ROW for XML
            ErrorManager.recordOrShowError(4065, "SERIALIZE-ROW", "BUFFER widget");
            // **The <attribute> attribute on the <widget id> has invalid arguments.
            return false;
         }
         XmlExport export = new XmlExport(targetData, formatted, encoding);
         return export.writeRecord(this, omitInitial);
      }
      else 
      {
         // is JSON
         com.fasterxml.jackson.core.JsonEncoding jsonEnc = JsonExport.parseEncoding(new character(encoding));
         if (jsonEnc == null)
         {
            ErrorManager.recordOrShowError(15342, encoding, "WRITE-JSON");
            // Invalid encoding '<encoding>' argument for <method>.
            return false;
         }
         
         JsonExport export = new JsonExport();
         return export.exportRecord(this, targetData, formatted, jsonEnc, omitInitial, omitOuter);
      }
   }
   
   /**
    * Getter of NUM-FIELDS attribute.  Only normal fields are counted, the hidden fields are not
    * taken into consideration.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   int _numFields()
   {
      return buffer().getNumFields();
   }
   
   /**
    * Execute a dynamic FIRST, LAST or UNIQUE find query. All the public overloaded methods
    * (findFirst, findLast and findUnique) will call this private implementation using the 
    * appropriate parameters. Internally this method uses the DynamicQueryHelper.parseFindQuery
    * static parser to return a FindQuery object using the passed predicate and locking mode.
    * After the FindQuery is obtained, the appropriate query is executed (FIRST, LAST or UNIQUE)
    * and the result is returned to caller method.
    * <p> 
    * If the method fails, it does not raise an error but displays a message. To suppress it,
    * the method should be called with NO-ERROR clause.
    *
    * @param   predicate
    *          Predicate expression for find that evaluates to the following syntax:
    *             [ WHERE [ logical-expression ] ] [ USE-INDEX index-name ]
    *          <ul>
    *             <li>Once evaluated, predicate-expression can contain only constants and
    *                   unabbreviated references to fields from the buffer.
    *             <li>The predicate-expression itself can be built using a concatenation of
    *                   character expressions.
    *          </ul>
    * @param   mode
    *          Search mode, one of FIRST, LAST or UNIQUE.
    * @param   lock
    *          Lock type. May additionally have no-wait flag set.
    * @param   substBuffers
    *          A list of substitution buffers.
    *
    * @return  {@code true} if operation succeeds, else {@code false}.
    */
   logical _find(String predicate, FindMode mode, LockType lock, Buffer... substBuffers)
   {
      final String qName = "find_" + mode + "_" + lock;
      P2JQuery query = DynamicQueryHelper.parseFindQuery(
                                 this, predicate, lock, qName, Lists.newArrayList(substBuffers));
      if (query == null)
      {
         String msg = "Invalid WHERE clause in FIND method for buffer " + buffer().getLegacyName() + " (10041)";
         ErrorManager.recordOrShowError(
               new int[] { 10041 },
               new String[] { msg },
               false, false, false, true, false);  // no DOT at the end of error message!
         return new logical(false);
      }
      
      boolean success = false;
      try
      {
         switch (mode)
         {
            case FIRST:
               query.first(lock);
               break;
            
            case LAST:
               query.last(lock);
               break;
            
            case UNIQUE:
               query.unique(lock);
               break;
         }
         if (!ErrorManager.isPendingError())
         {
            success = _available();
         }
      }
      catch (ErrorConditionException e)
      {
         // the error condition already has been handled by the ErrorManager, so we just return failure;
         // unexpected, unchecked exceptions are allowed to unwind the stack
         success = false;
      }
      
      return new logical(success);
   }
   
   /**
    * internal implementation of FIND-CURRENT() method.
    *
    * Changes the lock type for the record currently in the buffer. Refreshes the record if it was
    * changed by other client. Resets the CURRENT-CHANGED attribute accordingly.
    *
    * @param   lockType
    *          Exetnded lock type (includes the NO-WAIT status).
    *
    * @return  {@code true} if FIND-CURRENT() succeeds (the lock type is changed), else
    *          {@code false}.
    */
   private logical _current(LockType lockType)
   {
      RecordBuffer recBuf = buffer();
      
      // make sure to ignore errors caused by LockUnavailableException 
      try
      {
         recBuf.txHelper.errHlp.ignore(() ->
         {
            recBuf.armCurrentChanged();
            recBuf.reload(lockType, false);
         });
      }
      
      finally
      {
         recBuf.updateCurrentChanged();
      }
      
      // NOTE: the reference manual states that if lock operation fails, a message is displayed.
      //       However, this is not how the behaviour was observed. Actually there is no error
      //       whatsoever. Instead, the buffer is discarded after a failure to acquire the lock.
      //       The error was not observable, in fact at all, in all my testcases.
      return new logical(recBuf.isAvailable());
   }
   
   /**
    * Implementation of the merge changes. Always to be called for the AFTER-BUFFER. It will also
    * process the changes for the BEFORE BUFFER (the DELETE changes only), too.
    *
    * @param   porigBuf
    *          The original AFTER buffer.
    * @param   allCopyMode
    *          Use {@code true} to indicates that all rows of the source after-image table must be
    *          merged.
    *
    * @return  {@code true} on success. When {@code false} is returned the error code was already
    *          configured in {@code ErrorManager}.
    */
   private boolean mergeChangesImpl(BufferImpl porigBuf, boolean allCopyMode)
   {
      if (!isAfterBuffer() || !porigBuf.isAfterBuffer())
      {
         ErrorManager.recordOrShowError(11925, doGetName());
         // GET/MERGE-CHANGES require a BEFORE-TABLE for table <name>.
         return false;
      }
      
      logical ok = new logical(true);
      TemporaryBuffer afterBuffer = (TemporaryBuffer) buffer();
      TemporaryBuffer beforeBuffer = (TemporaryBuffer) peerBuffer.buffer();
      Class<? extends Record> origAfterImplClass = porigBuf.buffer().getDmoInfo().getImplementationClass();
      Class<? extends Record> origBeforeImplClass = ((BufferImpl) porigBuf.beforeBuffer().getResource())
            .buffer().getDmoInfo().getImplementationClass();
      Class<? extends Record> afterImplClass = afterBuffer.getDmoInfo().getImplementationClass();
      Session session = afterBuffer.getPersistence().getSession();
      beforeBuffer.readAllRows(
         (dmo) -> {
            if (!ok.booleanValue())
            {
               return;
            }
            
            // prepare the original before image
            TempRecord origBefRecord = null;
            try
            {
               origBefRecord = (TempRecord) session.get(origBeforeImplClass, dmo._originRowid());
            }
            catch (PersistenceException | ConditionException e)
            {
               // origBefRecord stays null
            }
            if (origBefRecord == null)
            {
               // this should not happen. The origin-rowid of the dmo should not be unknown! If it was not 
               // found then the changes in the original table were already accepted (?)
               
               ErrorManager.recordOrShowError(11928, doGetName(), porigBuf.doGetName());
               // INVALID origin-rowid of the change table <name> of <name>.
               ErrorManager.recordOrShowError(11926);
               // Found unexpected error while merging changes.
               
               ok.assign(false);
               return;
            }
            
            Integer errFlags = dmo._errorFlags();
            if (errFlags != null && errFlags != 0)
            {
               BufferImpl porigBefBuf = porigBuf.peerBuffer;
               porigBefBuf.buffer().loadRecord(origBefRecord);
               
               // if the record's any error/rejected markers were set, drop it from merge operation
               ok.assign(porigBefBuf.rejectRowChanges());
            }
            else
            {
               // prepare the original after image and put it in the [porigBuf] buffer
               TempRecord origRecord = null;
               try
               {
                  origRecord = (TempRecord) session.get(origAfterImplClass, origBefRecord._peerRowid());
               }
               catch (PersistenceException | ConditionException e)
               {
                  // origRecord stays null
               }
               porigBuf.buffer().loadRecord(origRecord);
               
               // prepare the after image/record in current buffer 
               TempRecord afterRecord;
               try
               {
                  afterRecord = (TempRecord) session.get(afterImplClass, dmo._peerRowid());
               }
               catch (PersistenceException | ConditionException e)
               {
                  afterRecord = null;
               }
               afterBuffer.loadRecord(afterRecord);
               
               // run the merge-row-changes code for prepared records 
               ok.assign(mergeRowChangesImpl(this.ref(), porigBuf.ref()));
            }
         });
      
      return ok.booleanValue();
   }
   
   /**
    * Implementation of the row merge changes. Always to be called for the AFTER-BUFFER.
    *
    * @param   changesBuf
    *          The changes AFTER buffer.
    * @param   origAfterBuf
    *          The original AFTER buffer.
    *
    * @return  {@code true} on success. When {@code false} is returned the error code was already
    *          configured in {@code ErrorManager}.
    */
   private static boolean mergeRowChangesImpl(BufferImpl changesBuf, BufferImpl origAfterBuf)
   {
      TempRecord dmo = (TempRecord) changesBuf.buffer().getCurrentRecord();
      if (dmo == null)
      {
         log.log(Level.SEVERE, "mergeRowChangesImpl: null TempRecord in changesBuf");
         return true;
      }
      
      Long originB4Rowid = dmo._originRowid();
      BufferImpl origB4Buffer = origAfterBuf.peerBuffer;
      
      switch (changesBuf.rowState().intValue())
      {
         case ROW_DELETED:
            if (dmo._rowState() == Buffer.ROW_DELETED &&
                origB4Buffer.findByRowID(new rowid(originB4Rowid)).getValue())
            {
               if (origB4Buffer.rowState().intValue() != Buffer.ROW_DELETED)
               {
                  ErrorManager.recordOrShowError(11928, changesBuf.doGetName(), origAfterBuf.doGetName());
                  // INVALID origin-rowid of the change table <name> of <name>.
                  ErrorManager.recordOrShowError(11926);
                  // Found unexpected error while merging changes.
                  return false;
               }
               // just remove the BEFORE-IMAGE, the AFTER-IMAGE is long gone 
               origB4Buffer.setUpBeforeBuffer(true);
               origB4Buffer.deleteRecord();
               origB4Buffer.setUpBeforeBuffer(false);
            }
            break;
            
         case ROW_MODIFIED:
         case ROW_CREATED:
            if (originB4Rowid == null)
            {
               ErrorManager.recordOrShowError(11923, origAfterBuf.doGetName(), changesBuf.doGetName());
               // MERGE-CHANGES run on original table <name> that does not match the ORIGIN-HANDLE in the GET-CHANGES table <name>.
               return false;
            }
            else if (!origB4Buffer.findByRowID(new rowid(originB4Rowid)).getValue())
            {
               ErrorManager.recordOrShowError(11928, changesBuf.doGetName(), origAfterBuf.doGetName());
               // INVALID origin-rowid of the change table <name> of <name>.
               ErrorManager.recordOrShowError(11926);
               // Found unexpected error while merging changes.
               return false;
            }
            
            Long afterRowid = origB4Buffer.afterRowid().getValue();
            if (afterRowid == null)
            {
               ErrorManager.recordOrShowError(11926);
               // Found unexpected error while merging changes.
               return false;
               // exits lambda, continue with next record.
               // The error message will be displayed multiple times, one for each record.
            }
            else if (!origAfterBuf.findByRowID(new rowid(afterRowid)).getValue())
            {
               ErrorManager.recordOrShowError(11923, origAfterBuf.doGetName(), changesBuf.doGetName());
               // MERGE-CHANGES run on original table <name> that does not match the ORIGIN-HANDLE in the GET-CHANGES table <name>.
               return false;
               // exits lambda, continue with next record.
               // The error message will be displayed multiple times, one for each record.
            }
            
            RecordBuffer origAfterRecBuf = origAfterBuf.buffer();
            TempRecord dstRec = (TempRecord) origAfterRecBuf.getCurrentRecord();
            if (dstRec._rowState() == null || dstRec._rowState() == Buffer.ROW_UNMODIFIED)
            {
               ErrorManager.recordOrShowError(11928, changesBuf.doGetName(), origAfterBuf.doGetName());
               // INVALID origin-rowid of the change table <name> of <name>.
               ErrorManager.recordOrShowError(11926);
               // Found unexpected error while merging changes.
               return false;
            }
            Record snapshot = dstRec.snapshot();
            
            // accepts the changes by copying the after-image table rows to the corresponding
            // after-image table rows in the original DataSet temp-table.
            BaseRecord.copy(dstRec, dmo, false);
            
            Long oldBefore = dstRec._peerRowid();
            if (origAfterBuf.isAfterBuffer() && oldBefore != null)
            {
               // removes the before-image table row (if it has one).
               if (origB4Buffer.findByRowID(new rowid(oldBefore)).getValue())
               {
                  origB4Buffer.setUpBeforeBuffer(true);
                  origB4Buffer.deleteRecord();
                  origB4Buffer.setUpBeforeBuffer(false);
               }
            }
            
            // also:
            //    sets the BEFORE-ROWID attribute of the row in the after-image table to the Unknown value (?)
            //    sets the ROW-STATE of the row in the after-image table to ROW-UNMODIFIED (0)
            //    resets ERROR/REJECTED attributes and ERROR-STRING to unknown value.
            dstRec._peerRowid(null);
            dstRec._errorFlags(null);
            dstRec._errorString(null);
            dstRec._peerRowid(null);
            dstRec._rowState(Buffer.ROW_UNMODIFIED);
            
            try
            {
               Map<String, List<Integer>> props = new HashMap<>();
               props.put(Buffer.ROW_STATE_FIELD, null);
               props.put(Buffer.AFTER_ROWID_FIELD, null);
               props.put(Buffer.ERROR_FLAG_FIELD, null);
               props.put(Buffer.ERROR_STRING_FIELD, null);
               origAfterRecBuf.getChangeBroker().stateChanged(
                     RecordChangeEvent.Type.UPDATE, origAfterRecBuf, dstRec, snapshot, props);
            }
            catch (PersistenceException e)
            {
               throw new RuntimeException(e);
            }
      }
      
      // great, return with success
      return true;
   }
   
   /**
    * Sets the {@code _originRowid} internal field and register the change with the {@code ChangeBroker} to
    * make sure the change is persisted.
    * 
    * @param   val
    *          The new value for {@code _originRowid} field. May be {@code null} for an unknown.
    */
   private void setOriginRowidImpl(Long val)
   {
      RecordBuffer recBuf = buffer();
      TempRecord dmo = (TempRecord) recBuf.getCurrentRecord();
      TempRecord snapshot = (TempRecord) dmo.deepCopy();
      dmo._originRowid(val);
      try
      {
         Map<String, List<Integer>> props = new HashMap<>();
         props.put(Buffer.ORIGIN_ROWID_FIELD, null);
         recBuf.getChangeBroker().stateChanged(RecordChangeEvent.Type.UPDATE, recBuf, dmo, snapshot, props);
      }
      catch (PersistenceException e)
      {
         e.printStackTrace();
      }
   }
      
   /**
    * Sets the {@code _dataSourceRowid} internal field and register the change with the {@code ChangeBroker}
    * to make sure the change is persisted.
    * 
    * @param   val
    *          The new value for {@code _dataSourceRowid} field. May be {@code null} for an unknown.
    */
   private void setDatasourceRowidImpl(Long val)
   {
      RecordBuffer recBuf = buffer();
      TempRecord dmo = (TempRecord) recBuf.getCurrentRecord();
      TempRecord snapshot = (TempRecord) dmo.deepCopy();
      dmo._datasourceRowid(val);
      try
      {
         Map<String, List<Integer>> props = new HashMap<>();
         props.put(Buffer.DATA_SOURCE_ROWID_FIELD, null);
         recBuf.getChangeBroker().stateChanged(RecordChangeEvent.Type.UPDATE, recBuf, dmo, snapshot, props);
      }
      catch (PersistenceException e)
      {
         e.printStackTrace();
      }
   }
   
   /**
    * Partial implementation of {@code SAVE-ROW-CHANGES} method. Validates some parameters and
    * calls the final worker.
    *
    * @param   bufNo
    *          The 0-base buffer index in the attached {@code DataSource}. 
    * @param   skip
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    * @param   noLobs
    *          Use {@code true} to specify the BLOB and CLOB fields should be ignored in the save
    *          operation.
    *
    * @return  {@code true} if operation is successful.
    */
   private logical saveRowChangesImpl(int bufNo, String skip, boolean noLobs)
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11938, doGetName());
         // SAVE-ROW-CHANGES must be run on the BEFORE table of buffer <name>.
         return new logical(false);
      }
      
      BufferImpl after = (BufferImpl) afterBuffer().getResource();
      DataSource dsrc = after.ref().dataSource;
      if (dsrc == null || !dsrc.valid())
      {
         ErrorManager.recordOrShowError(11942, doGetName(), after.doGetName());
         // SAVE-ROW-CHANGES for BEFORE buffer <name>, requires that the corresponding AFTER dataset member buffer <name> be attached to a DATA-SOURCE.
         return new logical(false);
      }
      
      return saveRowChangesImpl2(dsrc, after, bufNo, skip, noLobs);
   }
   
   /**
    * Implementation of {@code SAVE-ROW-CHANGES} method. Validates some parameters and
    * calls the final worker.
    *
    * @param   bufName
    *          The buffer name in the attached {@code DataSource}. 
    * @param   skip
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    * @param   noLobs
    *          Use {@code true} to specify the BLOB and CLOB fields should be ignored in the save
    *          operation.
    *
    * @return  {@code true} if operation is successful.
    */
   private logical saveRowChangesImpl(String bufName, String skip, boolean noLobs)
   {
      if (!isBeforeBuffer())
      {
         ErrorManager.recordOrShowError(11938, doGetName());
         // SAVE-ROW-CHANGES must be run on the BEFORE table of buffer <name>.
         return new logical(false);
      }
      
      BufferImpl after = (BufferImpl) afterBuffer().getResource();
      DataSource dsrc = after.ref().dataSource;
      if (dsrc == null || !dsrc.valid())
      {
         ErrorManager.recordOrShowError(11942, doGetName(), after.doGetName());
         // SAVE-ROW-CHANGES for BEFORE buffer <name>, requires that the corresponding AFTER dataset member buffer <name> be attached to a DATA-SOURCE.
         return new logical(false);
      }
      
      return saveRowChangesImpl2(dsrc, after, dsrc.getSourceBufferIndex(bufName), skip, noLobs);
   }
   
   /**
    * Implementation of {@code SAVE-ROW-CHANGES} method.
    *
    * @param   dSrc
    *          The {@code DataSource} that will be updated.
    * @param   after
    *          The AFTER-BUFFER to which the {@code DataSource} is attached.
    * @param   bufNo
    *          The 0-base buffer index in the attached {@code DataSource}.
    * @param   skip
    *          A list of field names that should not be assigned after a new row is created
    *          (comma-separated).
    * @param   noLobs
    *          Use {@code true} to specify the BLOB and CLOB fields should be ignored in the save
    *          operation.
    *
    * @return  {@code true} if operation is successful.
    */
   private logical saveRowChangesImpl2(DataSource dSrc,
                                       BufferImpl after,
                                       int bufNo,
                                       String skip,
                                       boolean noLobs)
   {
      if (!_available())
      {
         ErrorManager.recordOrShowError(11935, doGetName());
         // SAVE-ROW-CHANGES buffer <name> not available.
         return new logical(false);
      }
      
      if (bufNo < 0 || bufNo >= dSrc._getNumSourceBuffers())
      {
         ErrorManager.recordOrShowError(11906);
         // Unable to identify data-source buffer for SAVE-ROW-CHANGES.
         return new logical(false);
      }
      
      BufferImpl srcBuff = (BufferImpl) dSrc.getSourceBuffer(bufNo + 1).getResource();
      Set<String> skipFields = new HashSet<>();
      if (skip != null && !skip.isEmpty())
      {
         String[] skipList = skip.split(",");
         for (String tok : skipList)
         {
            if (tok.contains(".") || !srcBuff.hasField(tok))
            {
               ErrorManager.recordOrShowError(11909, tok, srcBuff.doGetName());
               // Unable to find sequence field  <name> in table <name> during SAVE-ROW-CHANGES.
               return new logical(false);
            }
            skipFields.add(tok);
         }
      }
      
      String saveWhere = dSrc.getSaveWhereString(bufNo + 1).toJavaType();
      int rowState = rowState().intValue();
      character allPairs = new character();
      character allSrcFields = new character();
      
      if (rowState != ROW_DELETED)
      {
         StringBuilder sb = new StringBuilder();
         StringBuilder sbAllSrc = new StringBuilder();
         for (Map.Entry<CaseInsensitiveString, DataSource.FieldReference> entry :
                  after.dataSrcMapping.entrySet())
         {
            String srcFld = entry.getValue().getName();
            String dsFld = entry.getKey().getValue();
            
            if (sbAllSrc.length() != 0)
            {
               sbAllSrc.append(",");
            }
            sbAllSrc.append(srcFld);
            
            if (entry.getValue().getBuffer() == srcBuff && !skipFields.contains(srcFld))
            {
               if (sb.length() != 0)
               {
                  sb.append(",");
               }
               sb.append(dsFld).append(",").append(srcFld);
            }
         }
         allPairs.assign(sb.toString());
         allSrcFields.assign(sbAllSrc.toString());
      }
      
      logical retval = new logical();
      doBlock(TransactionType.FULL, "save-row-changes-label", new Block() {
         @Override
         public void body()
         {
            if (rowState == ROW_CREATED)
            {
               srcBuff.create();
               
               after.findByRowID(afterRowid());
               logical ok = TemporaryBuffer.copy(
                     srcBuff.buffer(), after.buffer(), allSrcFields, allPairs, new logical(noLobs));
               
               if (ok.getValue())
               {
                  after.setDatasourceRowidImpl(srcBuff.rowID().getValue());
               }
               
               if (!ok.booleanValue())
               {
                  setErrorOnSave();
               }
               retval.assign(ok);
               return;
            }
            
            // first, detect whether the record loaded in buffer has been changed. If so (changed or modified), 
            // update the DATA-SOURCE-MODIFIED attribute for me, my temp-table, the temp-table of the
            // after-buffer, and dataset. The newly created record will not affect this attribute.
            // NOTE: the attribute/function name suggests that the data source buffer was altered but, from my
            //       tests, changing the source buffers does not affect this attribute. Having an modified or
            //       deleted before-image loaded in before-buffer(this) will set the attribute, instead.
            if (rowState == ROW_MODIFIED || rowState == ROW_DELETED)
            {
               BufferImpl.this.setDataSourceModified(true);
               tableHandle().unwrapDataSourceModifiable().setDataSourceModified(true);
               after.tableHandle().unwrapDataSourceModifiable().setDataSourceModified(true);
               peerBuffer.dataSet.setDataSourceModified(true);
            }
            
            logical ok = new logical();
            ErrorManager.silent(() -> ok.assign(
                  srcBuff._find(saveWhere, FindMode.UNIQUE, LockType.EXCLUSIVE, BufferImpl.this)));
            
            if (ok.isUnknown() || !ok.getValue() || !srcBuff._available())
            {
               if (rowState == ROW_DELETED)
               {
                  // this is a silent acceptance of the fact that the record was deleted from both
                  // locations (source and dataset table)
                  
                  // in any case, update peer: the deleted row is now gone for good
                  setOriginRowidImpl(null);
                  retval.assign(true);
                  return;
               }
               
               ErrorManager.recordOrShowError(11912, srcBuff.doGetName());
               // Unable to find record for db buffer <name> during SAVE-ROW-CHANGES
               setErrorOnSave();
               retval.assign(false);
               return;
            }
            
            if (rowState == ROW_DELETED)
            {
               srcBuff.deleteRecord();
               // update peer: the deleted row is now gone for good
               setOriginRowidImpl(null);
               retval.assign(true);
               return;
            }
            
            if (rowState == ROW_MODIFIED)
            {
               after.findByRowID(afterRowid());
               
               boolean preferDataset = dSrc.getPreferDataset().getValue();
               boolean mergeByField = dSrc.getMergeByField().getValue();
               List<TableMapper.LegacyFieldInfo> myFields = 
                     TableMapper.getLegacyOrderedList(BufferImpl.this, true);
               logical res = new logical();
               
               try
               {
                  if (preferDataset)
                  {
                     if (!mergeByField)
                     {
                        // overwrite and lose any DB changes to entire row 
                        res.assign(TemporaryBuffer.copy(srcBuff.buffer(),
                                                        after.buffer(),
                                                        allSrcFields,
                                                        allPairs,
                                                        new logical(noLobs)));
                     }
                     else
                     {
                        // save DS changes. Non overlapping DB field changes are kept.
                        StringBuilder sb = new StringBuilder();
                        for (TableMapper.LegacyFieldInfo fieldInfo : myFields)
                        {
                           String fldName = fieldInfo.getLegacyName();
                           BaseDataType fldVal = Session.PK.equals(fldName) ? after.rowID()
                                                                            : after.dereference(fldName);
                           if (!fldVal.equals(dereference(fldName)))
                           {
                              DataSource.FieldReference srcField =
                                    after.dataSrcMapping.get(new CaseInsensitiveString(fldName));
                              if (srcField == null)
                              {
                                 continue;
                              }
                              
                              String srcName = srcField.getName();
                              if (sb.length() != 0)
                              {
                                 sb.append(",");
                              }
                              sb.append(fldName).append(",").append(srcName);
                           }
                        }
                        character dsChangedPairs = new character(sb.toString());
                        
                        res.assign(TemporaryBuffer.copy(srcBuff.buffer(),
                                                        after.buffer(),
                                                        allSrcFields,
                                                        dsChangedPairs,
                                                        new logical(noLobs)));
                     }
                  }
                  else
                  {
                     if (!mergeByField)
                     {
                        // if any field was modified in database raise error, otherwise update all fields
                        for (TableMapper.LegacyFieldInfo fieldInfo : myFields)
                        {
                           String fldName = fieldInfo.getLegacyName();
                           DataSource.FieldReference srcField =
                                 after.dataSrcMapping.get(new CaseInsensitiveString(fldName));
                           if (srcField == null)
                           {
                              continue;
                           }
                           
                           String srcName = srcField.getName();
                           boolean wasAlteredInSrc;
                           
                           if (fieldInfo.getExtent() <= 0)
                           {
                              BaseDataType srcVal = Session.PK.equals(srcName) ? srcBuff.rowID()
                                    : srcBuff.dereference(srcName);
                              wasAlteredInSrc = !srcVal.equals(dereference(fldName));
                           }
                           else
                           {
                              wasAlteredInSrc = false;
                              for (int k = 1; k <= fieldInfo.getExtent(); k++)
                              {
                                 if (!srcBuff.dereference(k, srcName).equals(dereference(k, fldName)))
                                 {
                                    wasAlteredInSrc = true;
                                 }
                              }
                           }
                           
                           // before-buffer and source must be the same (otherwise, there was a change
                           // in the database).
                           if (wasAlteredInSrc)
                           {
                              // synchronize before & after buffers with the source
                              TemporaryBuffer.copy(after.buffer(),
                                                   srcBuff.buffer(),
                                                   allSrcFields,
                                                   allPairs,
                                                   new logical(noLobs));
                              setUpBeforeBuffer(true);
                              TemporaryBuffer.copy(buffer(),
                                                   srcBuff.buffer(),
                                                   allSrcFields,
                                                   allPairs,
                                                   new logical(noLobs));
                              setUpBeforeBuffer(false);
                              
                              // set ERROR & DATA-SOURCE-MODIFIED attributes
                              ErrorManager.recordOrShowError(11913, srcBuff.doGetName());
                              // SAVE-ROW-CHANGES found <name> record with conflicting change by another user.
                              setErrorOnSave();
                              res.assign(false);
                              break;
                           }
                        }
                        
                        if (res.isUnknown()) // ie. no 11913 was encountered
                        {
                           res.assign(TemporaryBuffer.copy(srcBuff.buffer(),
                                                           after.buffer(),
                                                           allSrcFields,
                                                           allPairs,
                                                           new logical(noLobs)));
                        }
                     }
                     else
                     {
                        // attempt to save DS changes for each field. If the field was altered in source
                        // buffer raise error and return.
                        StringBuilder sb = new StringBuilder();
                        for (TableMapper.LegacyFieldInfo fieldInfo : myFields)
                        {
                           BaseDataType fldVal;
                           BaseDataType originalVal;
                           String fldName = fieldInfo.getLegacyName();
                           boolean wasChanged;
                           
                           if (Session.PK.equals(fldName))
                           {
                              fldVal = after.rowID();
                              originalVal = rowID();
                              wasChanged = !fldVal.equals(originalVal);
                           }
                           else
                           {
                              if (fieldInfo.getExtent() <= 0)
                              {
                                 fldVal = after.dereference(fldName);
                                 originalVal = dereference(fldName);
                                 wasChanged = !fldVal.equals(originalVal);
                              }
                              else
                              {
                                 originalVal = new unknown(); // ignored, in fact
                                 wasChanged = false;
                                 for (int k = 1; k <= fieldInfo.getExtent(); k++)
                                 {
                                    fldVal = after.dereference(k, fldName);
                                    originalVal = dereference(k, fldName);
                                    
                                    if (!fldVal.equals(originalVal))
                                    {
                                       wasChanged = true;
                                    }
                                 }
                              }
                           }
                           
                           if (wasChanged)
                           {
                              DataSource.FieldReference srcField =
                                    after.dataSrcMapping.get(new CaseInsensitiveString(fldName));
                              if (srcField == null)
                              {
                                 continue;
                              }
                              
                              String srcName = srcField.getName();
                              boolean wasAlteredInSrc;
                              if (Session.PK.equals(srcName))
                              {
                                 wasAlteredInSrc = !originalVal.equals(srcBuff.rowID());
                              }
                              else
                              {
                                 if (fieldInfo.getExtent() <= 0)
                                 {
                                    wasAlteredInSrc = !originalVal.equals(srcBuff.dereference(srcName));
                                 }
                                 else
                                 {
                                    // originalVal is unknown(), must be recomputed
                                    wasAlteredInSrc = false;
                                    for (int k = 1; k <= fieldInfo.getExtent(); k++)
                                    {
                                       fldVal = srcBuff.dereference(k, srcName);
                                       originalVal = dereference(k, srcName);
                                       
                                       if (!fldVal.equals(originalVal))
                                       {
                                          wasAlteredInSrc = true;
                                       }
                                    }
                                 }
                              }
                              
                              if (wasAlteredInSrc)
                              {
                                 // synchronize before & after buffers with the source
                                 TemporaryBuffer.copy(after.buffer(),
                                                      srcBuff.buffer(),
                                                      allSrcFields,
                                                      allPairs,
                                                      new logical(noLobs));
                                 setUpBeforeBuffer(true);
                                 TemporaryBuffer.copy(buffer(),
                                                      srcBuff.buffer(),
                                                      allSrcFields,
                                                      allPairs,
                                                      new logical(noLobs));
                                 setUpBeforeBuffer(false);
                                 
                                 // set ERROR & DATA-SOURCE-MODIFIED attributes
                                 ErrorManager.recordOrShowError(11913, srcBuff.doGetName());
                                 // SAVE-ROW-CHANGES found <name> record with conflicting change by another user.
                                 res.assign(false);
                                 break;
                              }
                              
                              if (sb.length() != 0)
                              {
                                 sb.append(",");
                              }
                              sb.append(fldName).append(",").append(srcName);
                           }
                        }
                        
                        if (res.isUnknown()) // ie. no 11913 was encountered
                        {
                           res.assign(TemporaryBuffer.copy(srcBuff.buffer(),
                                                           after.buffer(),
                                                           allSrcFields,
                                                           new character(sb.toString()),
                                                           new logical(noLobs)));
                        }
                     }
                  }
               }
               catch (ErrorConditionException ex)
               {
                  ErrorManager.recordOrShowError(11910, srcBuff.doGetName());
                  // SAVE-ROW-CHANGES was unable to update the database table <table>.
                  setErrorOnSave();
                  retval.assign(false);
                  return;
               }
               
               srcBuff.release();
               if (!res.booleanValue())
               {
                  setErrorOnSave();
               }
               retval.assign(res);
               return;
            }
         }
      });
      
      if (retval.isUnknown())
      {
         // something has gone wrong, there is no other valid value for [rowState]
         setErrorOnSave();
         retval.assign(false);
      }
      
      return retval;
   }
   
   /**
    * Sets the {@code ERROR} attribute to {@code true} after a failed SAVE-ROW operation.
    * Affected entities are:
    * <ul>
    *    <li>current IMAGE (it is a BEFORE IMAGE)</li>
    *    <li>the AFTER-IMAGE of this BEFORE-IMAGE</li>
    *    <li>the TEMP-TABLE of current BUFFER</li>
    *    <li>the TEMP-TABLE of the AFTER-BUFFER</li>
    *    <li>the DATASET of the BUFFER</li>
    * </ul>
    * 
    * The  is not affected.
    * <p>
    * The attribute is not set to FALSE on a successful SAVE operation, the programmer must do
    * this by hand.
    */
   private void setErrorOnSave()
   {
      if (!isBeforeBuffer())
      {
         return;
      }
      
      error(true);               // current IMAGE (it is a BEFORE IMAGE)
      peerBuffer.error(true);    // the AFTER-IMAGE of this BEFORE-IMAGE
      
      TempTable beforeTable = TempTableBuilder.asTempTable((Temporary) this);
      TempTable afterTable = (TempTable) beforeTable.getAfterTable().getResource();
      
      beforeTable.error(true);   // the TEMP-TABLE of current BUFFER
      afterTable.error(true);    // the TEMP-TABLE of the AFTER-BUFFER
      
      DataSet ds = this.dataSet != null ? this.dataSet : peerBuffer.dataSet;
      if (ds != null && ds.valid())
      {
         ds.error(true);       // the DATASET of the BUFFER
      }
   }
   
   /**
    * Read record data from the ObjectInputStream 
    * 
    * @param   fields
    *          Record fields
    * @param   extents
    *          Record extents
    * @param   ois
    *          Source stream
    * 
    * @return  Values of the record fields or <code>null</code> if failed.
    */
   private Map<String, BaseDataType[]> readValues(List<TableMapper.LegacyFieldInfo> fields,
                                                  Map<String, Integer> extents,
                                                  ObjectInputStream ois)
   {
      Map<String, BaseDataType[]> values = new HashMap<>();
      try
      {
         for (TableMapper.LegacyFieldInfo lfi : fields)
         {
            String prop = lfi.getJavaName();
            Integer extent = extents.get(prop);
            
            Class<? extends BaseDataType> cls = lfi.getDataType();
            BaseDataType value = null;
            BaseDataType[] val;
            
            if (extent == null) 
            {
               // scalar case
               val = new BaseDataType[1];
               // call the default constructor
               val[0] = cls.newInstance();
               
               // deserialize
               val[0].readExternal(ois);
            }
            else
            {
               // array case
               val = new BaseDataType[extent];
               for (int i = 0; i < extent; i++)
               {
                  // call the default constructor
                  val[i] = cls.newInstance();
                  
                  // deserialize
                  val[i].readExternal(ois);
               }
            }
            values.put(prop, val);
         }
         
         return values;
      }
      catch (IOException              |
            SecurityException         |
            IllegalAccessException    |
            ClassNotFoundException    |
            InstantiationException ex)
      {
         // TODO: if something happened during the copy but after the create, do we need to put
         //       the original record back in?
         ErrorManager.recordOrShowError(
               4952,
               "Couldn't extract raw value from source field of RAW-TRANSFER",
               false, false, true
         );
         return null;
      }
   }
}
